#ifndef MPC_AVOIDTRAJGEN_HPP
#define MPC_AVOIDTRAJGEN_HPP

#include "common/inc/Position.hpp"
#include "common/inc/State.hpp"
#include "common/inc/BinTree.hpp"
#include "common/inc/StateTimeTree.hpp"
#include "motion/inc/Path.hpp"
#include "motion/inc/ShiftRecord.hpp"
#include "motion/inc/PCTraj.hpp"
#include "motion/inc/PSTraj.hpp"
#include "system/inc/System.hpp"
#include "environment/inc/DEM.hpp"
#include "environment/inc/BoolObsMap.hpp"
#include "environment/inc/DistObsMap.hpp"
#include "trajGen/inc/TrajGen.hpp"
#include "collisionCheck/inc/CollisionCheck.hpp"
#include "avoidObs/inc/GraphPassageway.hpp"

#include <vector>

#include <boost/thread/thread.hpp>
#include <boost/bind.hpp>

#define DEBUG_ATG 0

#define COEF_SAFETY 1.2

#define ATF_DFLT_MAX_DIST_TGT_REACHED 0.5
#define ATF_DFLT_MAX_DIST_WP_REACHED 2.5
#define ATF_DFLT_MAX_ANGLE_WP_REACHED 0.75*M_PI

namespace mpc
{
  typedef struct MAXDISTPOINT
  {
    int ind;                        /**< index */
    float s;                        /**< curve abscica */
    PositionXY pos;                 /**< position (x,y) */
    float dist;                     /**< minimal distance to obstacles */
    PositionXY near;                /**< position (x,y) of the nearest obstacle */
  } MaxDistPoint;

  typedef struct MAXDISTPATH
  {
    MaxDistPoint mdp_beg;
    MaxDistPoint mdp_end;
  } MaxDistPath;


                                    /**************************************************/
                                    /*************** Class AvoidTrajGen ***************/
                                    /**************************************************/

  /** \class AvoidTrajGen
   *  \brief Trajectory generation for obstacle avoidance while following a reference trajectory
   */
  class AvoidTrajGen
  {
  public : 
    // general options
    static int version;                                       /**< AvoidTrajGen version used (1=list version, 2=graph version) */

  protected : 
    // general inputs
    System* sys_;                                      /**< Robotic system constraints */
    XYThS_Path* path_;                                 /**< Center path of the corridor */
    double width_;                                     /**< Width of the corridor */
    double max_length_;                                /**< Length of the corridor currently treated */
    double fov_;                                       /**< Robot field of view */

    // stage inputs
    DEM* dem_;                                         /**< Data Elevation Map */
    State* loc_;                                       /**< Current location */

//     // stage outputs
    PCTraj* pctraj_;                                   /**< Best control trajectory found */
//     PSTraj* pstraj_;                                   /**< Best state trajectory found */

    STraj* straj_;                                     /**< Best state trajectory found */

    // stage temp data
    double curS_;                                      /**< Current curve abscica reached */
    std::vector<MaxDistPoint> maxDistPoints_;          /**< Vector containing the points maximizing the distance to obstacles found in the reference corridor */
//     std::vector<MaxDistPoint> strangPoints_;           /**< Vector containing the strangling points found in the reference corridor */
    std::list<MaxDistPoint> strangPoints_;             /**< list containing the strangling points found in the reference corridor */
    std::vector<WayPoint> waypoints_;                  /**< Vector containing the waypoints found in the reference corridor */

    XYThS_Config* tempGoal_;                           /**< goal position at the current stage */
    BinNode<MaxDistPath>* treeMDP_;                    /**< Tree used for the strangling points computation */

    PositionXY* nearPathPos_;                          /**< nearest configuration of the path from the current location the current stage */
    std::vector<PositionXY> vecNearPathPos_;           /**< nearest configuration of the path from the current location the current stage */

    DistObsMap* distmap_;                              /**< Data Elevation Map */
    boost::mutex mutexDistMap_;                        /**< mutex used to prevent DistMap access conflicts */

    // stage temp data (2nd version)
    std::vector< std::vector<Passageway>* > vecFreeItv_;   /**< list of the free intervals at each discrete segment along the corridor */
    GraphPassageway graphPsgwy_;                           /**< graph of the available passageways */
    GraphTopoPassageway* graphTopoPsgwy_;                  /**< topological graph of the available passageways */
    GraphPassageway *simplifiedGraphPsgwy_;                /**< simplified graph of the available passageways */
    StateTimeNode* treePsgwy_;                             /**< tree of the available passageways */
 
    // general temp data
    XYThS_Path* border1_;                              /**< First border of the corridor */
    XYThS_Path* border2_;                              /**< Second border of the corridor */

    TrajGen* trajgen_;                                 /**< trajectory generator tool used to find a control trajectory between waypoints */

    // general options
    static double ds_maxDist_points_;                  /**< distance between the max dist points */
    static int minNbItv_trajgen_;                      /**< minimal number of intervals to keep between two strangling points */
    static double coefTime_;                           /**< time coef used to select the waypoints arriving time */

    // mutexes
    boost::mutex mutexDEM_;                            /**< mutex used to prevent DEM access conflicts */
    boost::mutex mutexTraj_;                           /**< mutex used to prevent control and state trajectories access conflicts */

  public : 
    // dem treatment parameters
    static double SLOPEMAP_DIST;                       /**< Maximal distance considered between two cells of the slope map to compute the slope */
    static double MAX_SLOPE;                           /**< Maximal slope that the robot can climb */
    static double MAX_VAR_SLOPE;                       /**< Maximal slope derivative that the robot can climb */
    static double SAFE_DIST;                           /**< Distance that the robot must keep between its center of rotation and the obstacle to be safe */

    // corridor treatment parameters
    static double MAX_LENGTH_PATH_TTT;                 /**< maximal length of the path that will be analized for the trajectory computation */

    static bool POSTER_DEBUG_ATG;                      /**< true if the debug posters must be filled */

    static float maxDistTgtReached_;                  /*< maximal distance from the goal position until which it is considered as reached */
    static float maxDistWpReached_;                   /*< maximal distance from a waypoint until which it is considered as reached */
    static float maxAngleWpReached_;                  /*< maximal angle (in deg.) in the front of a waypoint until which it is considered as reached */

#if DEBUG_ATG
    RGBImage* image;                                   /**< image used to debug the process */
    RGBImage* image2;                                  /**< image used to debug the process */
    RGBImage* image3;                                  /**< image used to debug the process */
    RGBImage* image4;                                  /**< image used to debug the process */
#endif


    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    AvoidTrajGen();

  public : 
    /**
     *  \brief init constructor
     */
    AvoidTrajGen(System* sys, XYThS_Path* path, double max_length, double width, double fov);

    /**
     *  \brief destructor
     */
    virtual ~AvoidTrajGen();

    /*************** Getters and setters **************/

  public : 
    /**
     *  \brief set the distance between max dist points
     */
    static void setDSMaxDistPoints(double ds);

    /**
     *  \brief set the distance between max dist points
     */
    static void setMinNbItvTrajgen(int min_nb_itv);

    /**
     *  \brief set the time coef used to select the waypoints arriving time
     *  \param coef_time : the new time coefficient
     *
     *  coef_time = 1  -> maximal velocity used
     *  coef_time > 1  -> multiply the time needed to reach the goal
     */
    static void setCoefTime(double coef_time);

    /**
     *  \brief return a pointer to the current system
     */
    virtual const System* sys() const = 0;

    /**
     *  \brief return a copy of the list of waypoints found in the reference corridor
     *  \return a copy of the list of waypoints found in the reference corridor
     */
//     std::vector<XYThS_Config> waypoints();
    std::vector<WayPoint> waypoints();

    /**
     *  \brief return a copy of the current distmap
     *  \return a copy of the current distmap
     */
    DistObsMap* distmap();

    /**
     *  \brief return the current position of the robot
     *  \return the current position of the robot
     */
    PositionXYTh curPos() const;

    /**
     *  \brief return the goal position (final position of the followed trajectory)
     *  \return the goal position (final position of the followed trajectory)
     */
    PositionXYTh goalPos() const;

    /**
     *  \brief return the goal configuration (final configuration of the followed trajectory)
     *  \return the goal configuration (final configuration of the followed trajectory)
     */
    XYThS_Config goalCfg() const;

    /********** Corridor borders computation **********/

  protected : 
    /*
     *  \brief compute the corridor border paths
     *  \return 0 if the process succeded, -1 else
     */
    int computeBorderPaths();

    /************* Initialization functons ************/

  protected : 
    /**
     *  \brief Initialize the standard parameters of the trajectory generator
     *  \param dbg : the trajectory generator debugging mode
     */
    virtual void initTrajGen(int dbg=0);


    /************* Trajectory computation *************/

  protected : 
    /*
     *  \brief check for the AvoidTrajGen input parameters consistency
     *  \return 0 if all parameters are valid
     *  \return -2 if an I/O problem occured
     *  \return -3 if the avoid trajgen settings is not valid
     */
    int checkSettings();

    /*
     *  \brief reinit the avoidance trajectory computation process
     */
    void reinit();

    /*
     *  \brief get back indices of the path corresponding to the input curve abscica interval
     *  \param minS : input minimal curve abscica
     *  \param maxS : input maximal curve abscica
     *  \param indStart : output start index
     *  \param indEnd : output end index
     */
    void getIndPathItv_from_SItv(double minS, double maxS, int& indStart, int& indEnd);
    
    /*
     *  \brief the current curve abscica of the path from the robot location
     *  \param pos : the robot location
     *  \param oldS : the previous curve abscica reached
     *  \param maxDS : the maximum distance traveled from the previous recorded curve abscica
     *  \return the current curve abscica
     */
    double computeCurS(const PositionXY& pos, double oldS, double maxDS);

    /*
     *  \brief set a border path as an obstacle on the obstacle map
     *  \param obsmap : the obstacle map
     *  \param border : the border path
     *  \param curS : the current curve abscica
     *  \param maxDS : the distance on which the border path must be computed
     */
    void setBorderOnObsMap(BoolObsMap* obsmap_vsm, XYThS_Path* border1_, double curS, double maxDS);

    /*
     *  \brief compute the obstacle map from the reference corridor and the input DEM
     *  \param dem : the input DEM
     */
    BoolObsMap* buildObsMap(const DEM* dem);
//     BoolObsMap* buildObsMap(const DEM* dem, const DPMPath* path, double width);

    /*
     *  \brief search for the max dist point at curve abscica s
     *  \brief distMap : distance to obstacles map
     *  \param s : the current curve abscica
     *  \param pos : resulting cell position
     *  \param dist : resulting cell minimal distance to obstacles
     *  \param near : resulting cell nearest obstacle position
     *  \return true if a free position has been found, false else
     */
    bool searchMaxDistPoint(const DistObsMap& distMap, double s, PositionXY& pos, float& dist, PositionXY& near);

    /*
     *  \brief search for points maximizing the distance to obstacles in the navigation corridor
     *  \param obsmap : the obstacle map
     */
    void computeMaxDistPoints(const BoolObsMap* obsmap);
//     void computeMaxDistPoints(const DistObsMap* distmap);

    /*
     *  \brief compute the current temporary goal
     */
    void computeTempGoal();

    /*
     *  \brief search for a new strangling point
     *  \param indStart : the starting index of the search
     *  \param indEnd : the ending index of the search
     *  \param strang_point : the resulting strangling point
     */
    void searchForStrangPoint(int indStart, int indEnd, MaxDistPoint& strang_point);

    /*
     *  \brief compute the tree of strangling points recursively
     */
    void computeRecStrangPoints(BinNode<MaxDistPath>* nodeMDP);
    
    /*
     *  \brief compute the list of strangling points from the tree
     */
    void computeListStrangPoints();
    
    /*
     *  \brief add a first waypoint to the list of strangling points
     *  \param ind : index of the first strangling point
     *  \param (x,y) : location of the strangling point
     *  \param s : curve abscica reached at the initial location
     *  \param dist : distance to the nearest obstacle
     *  \param (nearX,nearY) : location of the nearest obstacle
     */
    void addFirstStrangPoint(int ind, double x, double y, double s, double dist, double nearX, double nearY);

    /*
     *  \brief search for strangling points in the navigation corridor (from max dist points)
     *  \param obsmap : the obstacle map
     */
    void computeStrangPoints();

    /*
     *  \brief search for strangling points in the navigation corridor (from max dist points) (2nd version)
     *  \param obsmap : the obstacle map
     */
    void computeStrangPointsV2();

    /*
     *  \brief compute the heading of the current waypoint
     *  \param posWP : position of the current waypoint
     *  \param posPrev : position of the previous waypoint
     *  \param posObs : position of the nearest obstacle
     */
    double computeHeading(const PositionXY& posWP, const PositionXY& posPrev, const PositionXY& posObs);

    /*
     *  \brief search for waypoints in the navigation corridor (from strangling points)
     */
    void computeWaypoints();

    /*
     *  \brief select states from waypoints
     */
    virtual void selectStatesFromWPs() = 0;

    /*
     *  \brief generate a trajectory from the found waypoints
     *  \return 0 if a free trajectory has been found, -1 else
     */
    virtual int computeTrajBtwnStates() = 0;

    /**
     *  \brief compute the obstacle avoidance trajectory (1st version : Waypoint list)
     *  \return 1 if the goal has been reached
     *  \return 0 if the process succeeded,
     *  \return -1 if the avoid trajgen computation failed
     *  \return -2 if an I/O problem occured
     *  \return -3 if the avoid trajgen settings is not valid
     *  \return -4 if a possible collision or ICS may occur
     */
    int computeAvoidTraj();


    /****** Trajectory computation (2nd version) ******/

  protected : 
    /*
     *  \brief reinit the avoidance trajectory computation process
     */
    void reinitV2();

    /**
     *  \brief return true if both passageways borders match
     *  \return true if both passageways borders match
     */
    static bool pasagewaysMatch(const Passageway& p1, const Passageway& p2);
    
    /*
     *  \brief compute the free passageway intervals on a single discrete segment of the corridor
     *  \param distMap : distance to obstacles map
     *  \param s : current curve absica wrt. the followed trajectory
     *  \return a vector containing the current passageway intervals
     */
    std::vector<Passageway>* computeSegmentItvs(const DistObsMap& distMap, float s);

    /*
     *  \brief compute the free passageway intervals on each discrete segment of the corridor
     *  \param distCol : if a collision has been found, contains the free distance in the front of the robot up to the obstacle
     *  \return 0 if the process succeeded, -1 if no passageway has been found
     */
    int computePassagewayItv(float* distCol=NULL);

    /*
     *  \brief compute the passageway graph
     *  \return 0 if the process succeeded, -1 if no passageway has been found
     */
    int computePassagewayGraph();

    /*
     *  \brief compute the topological graph
     */
    void computeTopologicalGraph();
    
    /*
     *  compute the strangling points on each node of the topological graph
     */
    void computeStrangPointsTopoGraph();

    /*
     *  \brief convert the topological graph back to a simple passageway graph
     */
    void convertTopoToSimpleGraph();
    
    /*
     *  remove the shortest and non collision-free edges
     */
    void removeShortAndColEdges();

    /*
     *  \brief simplify the passageway graph into a strangling points graph
     */
    void simplifyPassagewayGraph();

//     /*
//      *  \brief compute the passageway tree corresponding to each possible trajectories
//      */
//     void computePassagewayTree();

    /*
     *  \brief compute the state-time tree corresponding to each possible trajectories
     */
    virtual void computeStateTimeTree() = 0;

    /*
     *  \brief search for the best state-time sequence (minimizing the final time)
     */
    void searchBestBranch();

  public : 
    /**
     *  \brief compute the obstacle avoidance trajectory (2nd version : Waypoint graph)
     *  \return 0 if the process succeeded, -1 else
     */
    int computeAvoidTrajV2();


    /***************** Collision check ****************/

  protected : 
    /*
     *  \brief check for collision safety
     *  \param map : the obstacle map
     *  \param pstraj : the input state trajectory
     *  \return true if the trajectory is safe up to its end
     */
    bool checkSafety(const BoolObsMap& map, const STraj& pstraj);

    /*
     *  \brief return the time needed by the robot to stop
     *  \return the time needed by the robot to stop
     */
    virtual double timeToStop() const = 0;

    /*
     *  \brief compute the minimal distance traveled by the robot from its current to stop
     *  \return the minimal distance traveled by the robot from its current to stop
     */
    virtual double distToStop() const = 0;


    /****************** I/O functions *****************/

  public : 
//     /**
//      *  \brief compute the obstacle avoidance trajectory
//      *  \param state : the current state (position and speed) of the robot
//      *  \return 0 if the process succeeded, -1 else
//      */
//     int computeAvoidTraj(const State& s);

    /*
     *  \brief update the localization and speed of the robot
     *  \param s : the current localization and speed
     */
    void updateState(const State& s);

    /**
     *  \brief update the Data Elevation Map
     *  \param dem : the input DEM
     *  \return 0 if the process succeed, -1 else
     */
    int updateDEM(const DEM* dem);

//     /**
//      *  \brief get back the resulting state trajectory
//      *  \return the resulting state trajectory
//      */
//     virtual PSTraj* getSTraj() = 0;
    
    /**
     *  \brief get back the resulting control trajectory
     *  \return the resulting control trajectory
     */
    virtual PCTraj* getPCTraj() = 0;

    /**
     *  \brief get back the resulting state trajectory
     *  \return the resulting state trajectory
     */
    virtual STraj* getSTraj() = 0;

    /**
     *  \brief get back the resulting shift trajectory
     *  \return a new instance of the resulting shift trajectory
     */
    virtual ShiftRecord* getShiftRecord() = 0;

    /**
     *  \brief return true if the goal has been reached
     *  \return true if the goal has been reached
     */
    virtual bool goalReached() const;

    /************ Debug / Display functions ***********/

  public : 
    /*
     *  \brief display the list of max dist points
     */
    void displayMaxDistPoints() const;

    /*
     *  \brief display the list of strangling points
     */
    void displayStrangPoints() const;

    /*
     *  \brief display the list of waypoints
     */
    void displayWayPoints() const;

#if DEBUG_ATG
    /*
     *  \brief draw the state trajectory found on the image
     *  \param image : a pointer to the image to modify
     */
    virtual void drawPSTrajOnImage(RGBImage* image) = 0;

    /*
     *  \brief draw a path on the image
     *  \param path : the path to draw
     *  \param img : a pointer to the image to modify
     *  \param col : the color of the path
     */
    void drawXYThS_PathOnImage(const XYThS_Path& path, RGBImage* img, const RGBColor& col);

    /*
     *  \brief draw the passageway intervals found on the image
     *  \param img : a pointer to the image to modify
     *  \param col : the color of the intervals
     */
    void drawPassagewayItvOnImage(RGBImage* img, const RGBColor& col);

    /*
     *  \brief draw the passageway graph computed on the image
     *  \param graph : the passageway graph to display
     *  \param img : a pointer to the image to modify
     *  \param col : the color of the graph
     */
    void drawPassagewayGraphOnImage(GraphPassageway& graph, RGBImage* img, const RGBColor& col);

    /*
     *  \brief draw the topological passageway graph computed on the image
     *  \param img : a pointer to the image to modify
     *  \param colVertices : the color of the vertices of the graph
     *  \param colEdges : the color of the edges of the graph
     */
    void drawTopoPwyGraphOnImage(RGBImage* img, const RGBColor& colVertices, const RGBColor& colEdges);
#endif
  };

};

#endif // MPC_AVOIDTRAJGEN_HPP
