#ifndef MPC_DD2AVOIDTRAJGEN_HPP
#define MPC_DD2AVOIDTRAJGEN_HPP

#include "avoidObs/inc/AvoidTrajGen.hpp"
#include "trajGen/inc/ManaTrajGen.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /************** Class DD2AvoidTrajGen *************/
                                    /**************************************************/

  /** \class DD2AvoidTrajGen
   *  \brief Trajectory generation for obstacle avoidance while following a reference trajectory for a differential drive system
   */
  class DD2AvoidTrajGen : public AvoidTrajGen
  {
  protected : 
    std::vector<DD2StateT> wp_statet_;                 /**< Vector defining the successive state-times to reach */

    ManaPCTraj* builtPCTraj_;                          /**< Control trajectory currently built */
    ManaPSTraj* builtPSTraj_;                          /**< State trajectory currently built */

    DD2CTraj* builtCTraj_;                                /**< Control trajectory currently built */
    DD2STraj* builtSTraj_;                                /**< State trajectory currently built */

  public : 
    static double coef_min_available_speed;
    static double coef_max_available_speed;
//     static double MIN_AVAILABLE_SPEED;
//     static double MAX_AVAILABLE_SPEED;

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    DD2AvoidTrajGen();

  public : 
    /**
     *  \brief default constructor
     */
    DD2AvoidTrajGen(ManaSys* sys, XYThS_Path* path, double max_length, double width, double fov=0.0);

    /**
     *  \brief destructor
     */
    virtual ~DD2AvoidTrajGen();

    /*************** Getters and setters **************/

  protected : 
    /**
     *  \brief return a pointer to the Mana last registered localization
     *  \return a pointer to the Mana last registered localization
     */
    DD2State* loc() const;

    /**
     *  \brief return a pointer to the Mana trajectory generator
     *  \return a pointer to the Mana trajectory generator
     */
    ManaTrajGen* trajgen();

    /**
     *  \brief return a pointer to the Mana control trajectory
     *  \return a pointer to the Mana control trajectory
     */
    inline const DD2PCTraj* pctraj() const { return ((DD2PCTraj*) pctraj_); }

//     /**
//      *  \brief return a pointer to the Mana state trajectory
//      *  \return a pointer to the Mana state trajectory
//      */
//     inline const DD2PSTraj* pstraj() const { return ((DD2PSTraj*) pstraj_); }

    /**
     *  \brief return a pointer to the differential drive state trajectory
     *  \return a pointer to the differential drive state trajectory
     */
    inline const DD2STraj* straj() const { return ((DD2STraj*) straj_); }

  public : 
    /**
     *  \brief return a pointer to the current system
     */
    virtual const ManaSys* sys() const { return ((ManaSys*)sys_); }

    /************* Initialization functons ************/

  protected : 
    /**
     *  \brief Initialize the parameters of the Mana trajectory generator
     */
    virtual void initDD2TrajGen();

    /************* Trajectory computation *************/

  protected : 
    /**
     *  \brief compute the time coefficient used to select the time to reach waypoints
     *  \param minDistToObs : the current minimal distance to obstacles
     *
     *  The return coefficient belongs to [1.0/coef_max_available_speed ; 1.0/coef_min_available_speed]
     */
    double computeCoefTime(double minDistToObs);

    /*
     *  \brief select states from waypoints
     */
    virtual void selectStatesFromWPs();

    /*
     *  \brief generate a trajectory from the found waypoints
     *  \return 0 if a free trajectory has been found, -1 else
     */
    virtual int computeTrajBtwnStates();

//     /*
//      *  \brief generate a trajectory from the found waypoints
//      *  \return 0 if a free trajectory has been found, -1 else
//      */
//     virtual int computeTrajBtwnStates_old();

    /****** Trajectory computation (2nd version) ******/

  protected : 
    /*
     *  \brief compute the state-time tree corresponding to each possible trajectories
     */
    virtual void computeStateTimeTree();

    /***************** Collision check ****************/

  protected : 
    /*
     *  \brief return the time needed by the robot to stop
     *  \return the time needed by the robot to stop
     */
    virtual double timeToStop() const;

    /*
     *  \brief compute the minimal distance traveled by the robot from its current to stop
     *  \return the minimal distance traveled by the robot from its current to stop
     */
    virtual double distToStop() const;

    /****************** I/O functions *****************/

  public : 
    /**
     *  \brief compute the obstacle avoidance trajectory
     *  \param state : the current state (position and speed) of the robot
     *  \return 1 if the goal has been reached
     *  \return 0 if the process succeeded,
     *  \return -1 if the avoid trajgen computation failed
     *  \return -2 if an I/O problem occured
     *  \return -3 if the avoid trajgen settings is not valid
     *  \return -4 if a possible collision or ICS may occur
     */
    int computeAvoidTraj(const DD2State& s);
    
//     /**
//      *  \brief get back the resulting state trajectory
//      *  \return the resulting state trajectory
//      */
//     virtual DD2PSTraj* getSTraj();
    
    /**
     *  \brief get back the resulting control trajectory
     *  \return the resulting control trajectory
     */
    virtual DD2PCTraj* getPCTraj();

    /**
     *  \brief get back the resulting state trajectory
     *  \return the resulting state trajectory
     */
    virtual DD2STraj* getSTraj();

    /**
     *  \brief get back the resulting shift trajectory
     *  \return a new instance of the resulting shift trajectory
     */
    virtual ShiftRecord* getShiftRecord();

#if DEBUG_ATG
    /***************** Debug functions ****************/

    /*
     *  \brief draw the state trajectory found on the image
     *  \param image : a pointer to the image to modify
     */
    virtual void drawPSTrajOnImage(RGBImage* image);
#endif
  };

};

#endif // MPC_DD2AVOIDTRAJGEN_HPP
