#ifndef MPC_GRAPHPASSAGEWAY_HPP
#define MPC_GRAPHPASSAGEWAY_HPP

#include "avoidObs/inc/Passageway.hpp"
#include "avoidObs/inc/PassagewayTree.hpp"
#include "common/inc/Graph.hpp"
#include "graphics/inc/RGBImage.hpp"
#include "environment/inc/DistObsMap.hpp"

namespace mpc
{
  /**
   * \enum graph_io_t
   * \brief Enumeration of the graph io status
   *
   * GRAPH_IO_ND    : graph input / output not defined
   * GRAPH_INPUT    : input vertex or edge
   * GRAPH_OUTPUT   : output vertex or edge
   */
  typedef enum {
    GRAPH_IO_ND        =  -1,
    GRAPH_INPUT        =  0,
    GRAPH_OUTPUT       =  1
  } graph_io_t;
};

// namespace mpc
// {
//   typedef struct PASSAGEWAY_WIDTH_ITV
//   {
//     float s;
//     float minWidthB;
//     float maxWidthB;
//     PositionXY maxDistPoint;
//     float dist;
//     PositionXY nearObs;
//   } Passageway;
// };

                                    /**************************************************/
                                    /************** Class GraphPassageway *************/
                                    /**************************************************/

namespace mpc
{
  /** \class GraphPassageway
   *  \brief directed graph class with passageway vertex values
   */
  class GraphPassageway : public DGraph<Passageway>
  {
  protected : 
    float tildY_origin_;                                /**< lateral shift origin */
    float fixed_width_;                                 /**< fixed width of the corridor */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    GraphPassageway();

  public : 
    /**
     *  \brief default / init constructor
     *  \param tildY_origin : lateral shift origin of the graph
     *  \param tildY_origin : fixed width of the graph (in the case of fixed width mode)
     */
    GraphPassageway(float tildY_origin, float fixed_width=0);

    /**
     *  \brief copy constructor
     */
    GraphPassageway(const GraphPassageway& g);

    /**
     *  \brief destructor
     */
    ~GraphPassageway();

    /******************* Accessors ********************/

  public : 
    /**
     *  \brief return the lateral shift origin of the graph
     *  \return the lateral shift origin of the graph
     */
    inline float tildY_origin() const { return this->tildY_origin_; }

    /**
     *  \brief return fixed width of the corridor
     *  \return fixed width of the corridor
     */
    inline float fixed_width() const { return this->fixed_width_; }

    /******************** Searches ********************/

  public : 
    /**
     *  \brief search for a passageway in the graph
     *  \param pwy : the searched passageway
     *  \param it_res : the resulting vertex iterator
     *  \return true if the passageway has been found in the graph, false otherwise
     */
    bool searchForPassageway(const Passageway& pwy, DGVertex_iterator<Passageway>& it_res);

    /**************** Utility functions ***************/

  private : 
    /*
     *  \brief check if the input edge is too short
     *  \param edge : the input edge
     *  \param minDiff : minimal curve abscica difference between both endpoints of the edge
     *  \return true if the input edge is too short, false otherwise
     */
    bool isShortEdge(const DGEdge<Passageway>& edge, float minDiff) const;

    /*
     *  \brief check if the input edge is in collision with obstacles on the input distmap
     *  \param edge : the input edge
     *  \param distmap : the distance to obstacles map
     *  \return true if the input edge is in collision with obstacles, false otherwise
     */
    bool isColEdge(const DGEdge<Passageway>& edge, const DistObsMap& distmap) const;

    /*
     *  \brief check if the edge defined from v1 to v2 is in collision with obstacles on the input distmap
     *  \param edge : the input edge
     *  \param distmap : the distance to obstacles map
     *  \return true if the edge defined from v1 to v2 is in collision with obstacles, false otherwise
     */
    bool isColEdge(const DGVertex<Passageway>& v1, const DGVertex<Passageway>& v2, const DistObsMap& distmap) const;

    /*
     *  \brief select the best way to replace the current edge
     *  \param it_edge : iterator to the edge of the graph to replace
     *  \return GRAPH_INPUT if the best option is to link its ending vertex to the previous vertices of the graph
     *  \return GRAPH_OUTPUT if the best option is to link its starting vertex to the next vertices of the graph
     */
    graph_io_t selectBestEdgeRep(const DGEdge_iterator<Passageway>& it_edge);

    /*
     *  \brief reconnect an edge from the graph and try to reconnect it to the previous nodes of the graph
     *  \param vertexBeg_it : iterator to the starting vertex of the edge to replace
     *  \param vertexEnd_it : iterator to the ending vertex of the edge to replace
     *  \param distmap : distance to obstacles map
     */
//     void reconnectEdgeToPrevNodes(const DGEdge_iterator<Passageway>& it_edge, const DistObsMap& distmap);
    void reconnectEdgeToPrevNodes(DGVertex_iterator<Passageway>& vertexBeg_it, DGVertex_iterator<Passageway>& vertexEnd_it, const DistObsMap& distmap);

    /*
     *  \brief reconnect an edge from the graph and try to reconnect it to the next nodes of the graph
     *  \param vertexBeg_it : iterator to the starting vertex of the edge to replace
     *  \param vertexEnd_it : iterator to the ending vertex of the edge to replace
     *  \param distmap : distance to obstacles map
     */
//     void reconnectEdgeToNextNodes(const DGEdge_iterator<Passageway>& it_edge, const DistObsMap& distmap);
    void reconnectEdgeToNextNodes(DGVertex_iterator<Passageway>& vertexBeg_it, DGVertex_iterator<Passageway>& vertexEnd_it, const DistObsMap& distmap);

    /*
     *  \brief remove an edge from the graph and try to reconnect the graph
     *  \param it_edge : iterator to the edge to replace
     *  \param side : the side of the graph to which the nodes of the edge must be reconnected
     *  \param distmap : distance to obstacles map
     *  \return an iterator to the next edge in the list of edges
     */
    DGEdge_iterator<Passageway> replaceEdge(const DGEdge_iterator<Passageway>& it_edge, graph_io_t side, const DistObsMap& distmap);

  public : 
    /**
     *  \brief remove edges that are either too short or in collision with obstacles
     *  \param minDiff : the minimal distance between two successives vertices of the graph
     *  \param distmap : distance to obstacles map
     */
    void removeShortAndColEdges(float minDiff, const DistObsMap& distmap);

    /**
     *  \brief convert the passageway graph into a passageway tree
     *  \param sys : the robotic system constraints
     *  \return a pointer to the new passageway tree
     */
    PassagewayNode* toPassawagewayTree(const System& sys);

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the graphe
     */
    virtual void display();

//     /**
//      *  \brief draw the current graph on an image
//      *  \param image : the image on which the graph must be drawn
//      *  \param col : the color of the path
//      */
//     void drawOnImage(const RGBImage& image, const RGBColor& col);
  };
};


                                    /**************************************************/
                                    /*************** Class VecPassageway **************/
                                    /**************************************************/

namespace mpc
{
  /** \class VecPassageway
   *  \brief vector of passageways
   */
  class VecPassageway : public std::vector<Passageway*>
  {
    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    VecPassageway();

    /**
     *  \brief copy constructor
     */
    VecPassageway(const VecPassageway& v);

    /**
     *  \brief destructor
     */
    ~VecPassageway();

    /******************* Accessors ********************/

  public : 
    /**
     *  \brief get back the first element of the vector
     *  \return a pointer to the first element of the vector or NULL if the list is empty
     */
    const Passageway* first() const;

    /**
     *  \brief get back the last element of the vector
     *  \return a pointer to the last element of the vector or NULL if the list is empty
     */
    const Passageway* last() const;

    /************** Vector modifications **************/

  public : 
    /**
     *  \brief append an element at the end of the vector
     *  \param p : the new passageway to append
     */
    void append(const Passageway& p);

    /**
     *  \brief prepend an element at the beginning of the vector
     *  \param p : the new passageway to prepend
     */
    void prepend(const Passageway& p);

    /**
     *  \brief concatenate a vector at the end of the current one
     *  \param v : the new vector to concat
     */
    void concat(const VecPassageway* v);

    /**
     *  \brief clear all the values of the current vector
     */
    void clearVec();

    /******************** Searches ********************/

  private : 
    /**
     *  \brief search for a strangling point in the current vector between indices i_min and i_max
     *  \param i_min : index of the starting point of the search
     *  \param i_max : index of the ending point of the search
     *  \param i_strang : index of the resulting strangling point
     *  \return a pointer to the strangling point between indices i_min and i_max
     */
    const Passageway* searchForStrangPoint(int i_min, int i_max, int& i_strang) const;

    /**
     *  \brief search for the vector of strangling points of the current vector (recusrive method)
     *  \param minDiffInd : the minimum number of indices between two successives strangling points
     *  \param i_min : index of the starting point of the search
     *  \param i_max : index of the ending point of the search
     *  \return a vector of the strangling points of the current vector
     *
     *  A new instance of VecPassageway is returned
     */
    VecPassageway* searchForStrangPointsRec(int minDiffInd, int i_min, int i_max) const;

    /**
     *  \brief search for the vector of strangling points of the current vector (recusrive method)
     *  \param minDiff : the minimal distance between two successives strangling points
     *  \param i_min : index of the starting point of the search
     *  \param i_max : index of the ending point of the search
     *  \return a vector of the strangling points of the current vector
     *
     *  A new instance of VecPassageway is returned
     */
    VecPassageway* searchForStrangPointsRec(float minDiff, int i_min, int i_max) const;

  public : 
    /**
     *  \brief search for the vector of strangling points of the current vector
     *  \param minDiffInd : the minimum number of indices between two successives strangling points
     *  \return a vector of the strangling points of the current vector
     *
     *  A new instance of VecPassageway is returned
     */
    VecPassageway* searchForStrangPoints(int minDiffInd) const;

    /**
     *  \brief search for the vector of strangling points of the current vector
     *  \param minDiff : the minimal distance between two successives strangling points
     *  \return a vector of the strangling points of the current vector
     *
     *  A new instance of VecPassageway is returned
     */
    VecPassageway* searchForStrangPoints(float minDiff) const;

    /**
     *  \brief search for the vector of strangling points of the current vector and includes its first and last values
     *  \param minDiff : the minimal distance between two successives strangling points
     *  \return a vector of the strangling points of the current vector
     *
     *  A new instance of VecPassageway is returned
     */
    VecPassageway* searchForStrangPointsAndBounds(float minDiff) const;
  };
};


                                    /**************************************************/
                                    /************ Class GraphTopoPassageway ***********/
                                    /**************************************************/

namespace mpc
{
  /** \class GraphTopoPassageway
   *  \brief directed graph class representing the topological passageways (list of passageways vertex values)
   */
  class GraphTopoPassageway : public DGraph<VecPassageway*>
  {
  protected : 
    float tildY_origin_;                                /**< lateral shift origin */
    float fixed_width_;                                 /**< fixed width of the corridor */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    GraphTopoPassageway();

  public : 
    /**
     *  \brief copy constructor
     */
    GraphTopoPassageway(const GraphTopoPassageway& g);

    /**
     *  \brief init constructor (from a passageway graph)
     *  \param gp : input passageway graph
     */
    GraphTopoPassageway(GraphPassageway& gp);
//     GraphTopoPassageway(const GraphPassageway& gp); :-(

    /**
     *  \brief destructor
     */
    ~GraphTopoPassageway();

    /******************* Accessors ********************/

  public : 
    /**
     *  \brief return the lateral shift origin of the graph
     *  \return the lateral shift origin of the graph
     */
    inline float tildY_origin() const { return this->tildY_origin_; }

    /**
     *  \brief return fixed width of the corridor
     *  \return fixed width of the corridor
     */
    inline float fixed_width() const { return this->fixed_width_; }

    /*************** Graph modifications **************/

  private : 
    /*
     *  \brief recursive function used to build the topo graph from a passageway graph
     *  \param gp : reference to the passageway graph
     *  \param gp_it : iterator to a vertex of the passageway graph
     *  \return an iterator to the new node created in the topo graph
     */
    DGVertex_iterator<VecPassageway*> buildTopoVertex(const GraphPassageway& gp, DGVertex_iterator<Passageway>& gp_it);

  public : 
    /**
     *  \brief clear the graph content
     */
    virtual void clear();

    /**
     *  \brief simplfy the graph into a graph of strangling points
     *  \param minDiff : the minimal distance between two successives strangling points
     */
    void simplifyIntoSPGraph(float minDS_StrPoints);

    /**************** Utility functions ***************/

  public : 
    /*
     *  \brief search for a node in the topo graph containing the input passageway
     *  \param pwy : the input passageway
     *  \param gtp_vertex : the output node of the topo graph corresponding to the passageway (if it has been found)
     *  \return true if the input passageway has been found in the topo graph
     */
    bool includePassageway(const Passageway& pwy, DGVertex_iterator<VecPassageway*>& gtp_vertex);

    /**
     *  \brief convert the topo graph into a simple passageway graph
     *  \return a pointer to the new passageway graph
     *
     *  A new instance of GraphPassageway is returned
     */
    GraphPassageway* toGraphPassageway();

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the graphe
     */
    virtual void display();
  };
};


#endif // MPC_GRAPHPASSAGEWAY_HPP

