#ifndef MPC_BINTREE_HPP
#define MPC_BINTREE_HPP

#include <vector>
#include <stdlib.h>

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class BinNode *****************/
                                    /**************************************************/

  /** \class BinNode
   *  \brief template node class for binary tree construction
   */
  template <typename T> class BinNode
  {
  protected : 
    BinNode<T>* father_;                              /*< father of the current node */
    BinNode<T>* left_;                                /*< left son of the current node */
    BinNode<T>* right_;                               /*< right son of the current node */
    T val_;                                        /*< the value of the node */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */  
    BinNode()
    {
      father_ = NULL;
      left_ = NULL;
      right_ = NULL;
    }

  public : 
    /**
     *  \brief init constructor
     */  
    BinNode(T val)
    {
      father_ = NULL;
      left_ = NULL;
      right_ = NULL;
      val_ = val;
}

    /**
     *  \brief destructor
     */  
    ~BinNode()
    {
      if(left_ != NULL) delete left_;
      if(right_ != NULL) delete right_;
    }
    
    /******************** Accessors *******************/

  public : 
    /**
     *  \brief return the father of the current node
     *  \return a pointer to the father of the current node
     */
    BinNode<T>* father() const
    {
      return this->father_;
    }

    /**
     *  \brief return the left son of the current node
     *  \return a pointer to the left son of the current node
     */
    BinNode<T>* left() const
    {
      return this->left_;
    }

    /**
     *  \brief return the right son of the current node
     *  \return a pointer to the right son of the current node
     */
    BinNode<T>* right() const
    {
      return this->right_;
    }

    /**
     *  \brief return the value of the node
     *  \return the value of the node
     */
    const T& val() const
    {
      return this->val_;
    }

    /**
     *  \brief return a reference to the value of the node
     *  \return a reference to the value of the node
     */
    T& val()
    {
      return this->val_;
    }

//     /**
//      *  \brief set the left son of the current node
//      *  \param node : a pointer to the new left son
//      */
//     void setLeft(BinNode* node)
//     {
//   this->left_ = node;
//   if(node != NULL) node->father_ = this;
//     }

//     /**
//      *  \brief set the right son of the current node
//      *  \param node : a pointer to the new right son
//      */
//     void setRight(BinNode* node)
//     {
//   this->right_ = node;
//   if(node != NULL) node->father_ = this;
//     }

    /**
     *  \brief add a left son to the current node
     *  \param val : the value of the left son
     */
    void setLeft(const T& val)
    {
      if(left_ != NULL) delete left_;
      left_ = new BinNode<T>(val);
    }

    /**
     *  \brief add a right son to the current node
     *  \param val : the value of the right son
     */
    void setRight(const T& val)
    {
      if(right_ != NULL) delete right_;
      right_ = new BinNode<T>(val);
    }

    /**
     *  \brief return true if the current node is a leaf
     *  \return true if the current node is a leaf
     */
    bool isLeaf() const
    {
      if(left_ == NULL && right_ == NULL) return true;
      else return false;
    }

    /**
     *  \brief return true if the current node is a root
     *  \return true if the current node is a root
     */
    bool isRoot() const
    {
      if(father_ == NULL) return true;
      else return false;
    }


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief return a vector containing all the values of the leaves ordered from left to right
     *  \return a vector containing all the values of the leaves ordered from left to right
     *
     *  A new instance of std::vector is returned
     */
    void getLeavesVal(std::vector<T>& vecLeaves) const
    {
      if(this->isLeaf()) vecLeaves.push_back(val_);
      else 
	{
	  if(this->left() != NULL) this->left()->getLeavesVal(vecLeaves);
	  if(this->right() != NULL) this->right()->getLeavesVal(vecLeaves);
	}
    }

//     /**
//      *  \brief display the tree
//      */
//     void display();
  };

};

#endif // MPC_BINTREE_HPP
