#ifndef MPC_GRAPH_HPP
#define MPC_GRAPH_HPP

#include <vector>
#include <list>
#include <iostream>
#include <stdlib.h>

#include "common/inc/Tree.hpp"

namespace mpc
{
  template <typename T> class DGVertex;
  template <typename T> class DGEdge;
  template <typename T> class DGVertex_iterator;
  template <typename T> class DGEdge_iterator;
  template <typename T> class DGraph;

                                    /**************************************************/
                                    /****************** Class DGVertex ****************/
                                    /**************************************************/

  /** \class DGVertex
   *  \brief template vertex class for directed graph construction
   */
  template <typename T> class DGVertex
  {
  protected : 
    std::vector< DGEdge_iterator<T> > inputEdges_;                    /*< vertex's input edge list */
    std::vector< DGEdge_iterator<T> > outputEdges_;                   /*< vertex's output edge list */
    T val_;                                                         /*< the value of the vertex */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */  
    DGVertex() {}

  public : 
    /**
     *  \brief init constructor
     *  \param val : the value associated to the vertex
     */  
    DGVertex(T val)
    {
      this->val_ = val;
    }

    /**
     *  \brief copy constructor
     *  \param v : the vertex to copy
     */  
    DGVertex(const DGVertex<T>& v)
    {
      // copy the vertex value
      this->val_ = v.val;

      // copy the vertex input edge list
      for(typename std::vector< DGEdge_iterator<T> >::iterator it = v.inputEdges_.begin(); it != v.inputEdges_.end(); it++) this->inputEdges_.push_back(*it);

      // copy the vertex output edge list
      for(typename std::vector< DGEdge_iterator<T> >::iterator it = v.outputEdges_.begin(); it != v.outputEdges_.end(); it++) this->outputEdges_.push_back(*it);
    }

    /**
     *  \brief destructor
     */
    ~DGVertex() {}

    /******************* Accessors ********************/

  public : 
    /**
     *  \brief return the value of the node
     */
    const T& val() const 
    {
      return this->val_;
    }

    /**
     *  \brief return a reference to the value of the node
     */
    T& val()
    {
      return this->val_;
    }

    /**
     *  \brief return the number of input edges of the current vertex
     *  \return the number of input edges of the current vertex
     */
    int nbInputEdges() const
    {
      return this->inputEdges_.size();
    }

    /**
     *  \brief return the number of output edges of the current vertex
     *  \return the number of output edges of the current vertex
     */
    int nbOutputEdges() const
    {
      return this->outputEdges_.size();
    }


    /************** Edges modifications ***************/

  public : 
    /*
     *  \brief adding an input edge to the node
     *  \param edge : the new input edge to add
     */
    void addInputEdge(DGEdge_iterator<T> edge)
    {
      inputEdges_.push_back(edge);
    }

    /*
     *  \brief adding an output edge to the node
     *  \param edge : the new output edge to add
     */
    void addOutputEdge(DGEdge_iterator<T> edge)
    {
      outputEdges_.push_back(edge);
    }

    /*
     *  \brief removing an input edge to the node
     *  \param edge : the new input edge to remove
     */
    void removeInputEdge(DGEdge<T>* edge)
    {
      typename std::vector< DGEdge_iterator<T> >::iterator it = this->inputEdges_.begin(); 
      bool found = false;

      while(!found && it != this->inputEdges_.end()) 
	{
	  if(*it == edge) 
	    {
// 	      it = this->inputEdges_.erase(it);
	      this->inputEdges_.erase(it);
	      found = true;
	    }
	  else it++;
	}
    }

    /*
     *  \brief removing an output edge to the node
     *  \param edge : the new output edge to remove
     */
    void removeOutputEdge(DGEdge<T>* edge)
    {
      typename std::vector< DGEdge_iterator<T> >::iterator it = this->outputEdges_.begin(); 
      bool found = false;

      while(!found && it != this->outputEdges_.end()) 
	{
	  if(**it == edge) 
	    {
// 	      it = this->outputEdges_.erase(it);
	      this->outputEdges_.erase(it);
	      found = true;
	    }
	  else it++;
	}
    }

    /*
     *  \brief clear the list of input edges of the vertex
     */
    void clearInputEdges()
    {
      inputEdges_.clear();
    }

    /*
     *  \brief clear the list of input edges of the vertex
     */
    void clearOutputEdges()
    {
      outputEdges_.clear();
    }

    friend class DGraph<T>;
  };


                                    /**************************************************/
                                    /************* Class DGVertex_iterator ************/
                                    /**************************************************/

  template <typename T> class DGVertex_iterator : public std::list< DGVertex<T>* >::iterator 
  {
    /********** Constructors and destructors **********/
    
  public : 
    DGVertex_iterator() : std::list< DGVertex<T>* >::iterator() {}
    DGVertex_iterator(const DGVertex_iterator& it) : std::list< DGVertex<T>* >::iterator(it) {}
    DGVertex_iterator(const typename std::list< DGVertex<T>* >::iterator& it) : std::list< DGVertex<T>* >::iterator(it) {}
  };


                                    /**************************************************/
                                    /******************* Class DGEdge *****************/
                                    /**************************************************/

  /** \class DGEdge
   *  \brief template edge class for directed graph construction
   */
  template <typename T> class DGEdge
  {
  protected : 
    DGVertex_iterator<T> startVertex_;                    /*< starting vertex of the edge */
    DGVertex_iterator<T> endVertex_;                      /*< ending vertex of the edge */
//     T val_;                                            /*< the value of the node */

    /********** Constructors and destructors **********/

//   protected : 
  public : 
    /**
     *  \brief default constructor
     */  
    DGEdge()
    {
      this->startVertex_ = NULL;
      this->endVertex_ = NULL;
    }

//   public : 
//     /**
//      *  \brief init constructor
//      *  \param val : the value associated to the vertex
//      */  
//     DGEdge(T val);

    /**
     *  \brief copy constructor
     *  \param v : the vertex to copy
     */  
    DGEdge(const DGEdge& e)
    {
      this->startVertex_ = e.startVertex_;
      this->endVertex_ = e.endVertex_;
    }

    /**
     *  \brief copy constructor
     *  \param v : the vertex to copy
     */  
    DGEdge(const DGVertex_iterator<T>& startV, const DGVertex_iterator<T>& endV)
    {
      this->startVertex_ = startV;
      this->endVertex_ = endV;
    }

    /**
     *  \brief destructor
     */
    virtual ~DGEdge() {}

    
    /******************* Accessors ********************/

  public : 
    /**
     *  \brief return the starting vertex of the edge
     */
    const DGVertex<T>* startVertex() const 
    {
      return (*startVertex_);
    }

    /**
     *  \brief return the ending vertex of the edge
     */
    const DGVertex<T>* endVertex() const 
    {
      return (*endVertex_);
    }

  protected : 
    /**
     *  \brief return an iterator in the graph to the starting vertex of the edge
     */
    DGVertex_iterator<T> startVertexIt() const 
    {
      return startVertex_;
    }

    /**
     *  \brief return an iterator in the graph to the starting vertex of the edge
     */
    DGVertex_iterator<T> endVertexIt() const 
    {
      return endVertex_;
    }

    friend class DGraph<T>;
  };

                                    /**************************************************/
                                    /************** Class DGEdge_iterator *************/
                                    /**************************************************/

  template <typename T> class DGEdge_iterator : public std::list< DGEdge<T>* >::iterator 
  {
    /********** Constructors and destructors **********/
    
  public : 
    DGEdge_iterator() : std::list< DGEdge<T>* >::iterator() {}
    DGEdge_iterator(const DGEdge_iterator& it) : std::list< DGEdge<T>* >::iterator(it) {}
    DGEdge_iterator(const typename std::list< DGEdge<T>* >::iterator& it) : std::list< DGEdge<T>* >::iterator(it) {}
  };


                                    /**************************************************/
                                    /******************* Class DGraph *****************/
                                    /**************************************************/

  /** \class DGraph
   *  \brief template directed graph class
   */
  template <typename T> class DGraph
  {
  protected : 
    std::list< DGVertex<T>* > vertices_;               /**< list of the graph vertices */
    std::list< DGEdge<T>* > edges_;                    /**< list of the graph edges */    

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    DGraph() {}

    /**
     *  \brief default constructor
     */
    DGraph(const DGraph& g) 
    {
      std::cout << "DGraph copy constructor : Not implemented yet! " << std::endl;
      exit(1);
    }

    /**
     *  \brief destructor
     */
    virtual ~DGraph() 
    {
      this->clear();
    }


    /******************* Accessors ********************/

  public : 
    /**
     *  \brief get back an iterator to the first vertex of the graph's vertices list
     *  \return an iterator to the first vertex of the graph's vertices list
     */
    DGVertex_iterator<T> firstVertex() // const :-(
    {
      return DGVertex_iterator<T>(this->vertices_.begin());
    }

    /**
     *  \brief get back an iterator to the first vertex of the graph's vertices list
     *  \return an iterator to the first vertex of the graph's vertices list
     */
    std::vector<DGVertex_iterator<T> > rootVertices()
    {
      std::vector< DGVertex_iterator<T> > vecRes;

      for(typename std::list< DGVertex<T>* >::iterator it = this->vertices_.begin(); it != this->vertices_.end(); it++)
	{
	  if((*it)->inputEdges_.size() == 0) vecRes.push_back(DGVertex_iterator<T>(it));
	}

      return vecRes;
    }

    /**
     *  \brief return the number of vertices of the graph
     *  \return the number of vertices of the graph
     */
    int nbVertices() const
    {
      return this->vertices_.size();
    }

    /**
     *  \brief return the number of edges of the graph
     *  \return the number of edges of the graph
     */
    int nbEdges() const
    {
      return this->edges_.size();
    }

    /**
     *  \brief return a reference to the list of vertices of the graph
     *  \return a reference to the list of vertices of the graph
     */
    std::list< DGVertex<T>* >& vertices()
    {
      return this->vertices_;
    }

    /**
     *  \brief return a reference to the list of edges of the graph
     *  \return a reference to the list of edges of the graph
     */
    std::list< DGEdge<T>* >& edges()
    {
      return this->edges_;
    }

  public : 
    /*
     *  \brief return the number of predecessor of the input vertex
     *  \param vertex_it : vertex iterator
     *  \return the number of predecessor of the input vertex
     */
    int nbPredecessors(const DGVertex_iterator<T>& vertex_it) const
    {
      return (*vertex_it)->inputEdges_.size();
    }

    /*
     *  \brief return the number of successor of the input vertex
     *  \param vertex_it : vertex iterator
     *  \return the number of predecessor of the input vertex
     */
    int nbSuccessors(const DGVertex_iterator<T>& vertex_it) const
    {
      return (*vertex_it)->outputEdges_.size();
    }

    /*
     *  \brief return the i-th predecessor of the input vertex
     *  \param vertex_it : input vertex iterator
     *  \param i-th : number of the predecessor
     */
    DGVertex_iterator<T> predecessor(const DGVertex_iterator<T>& vertex_it, int i) const
    {
      return (*((*vertex_it)->inputEdges_.at(i)))->startVertex_;
    }

    /*
     *  \brief return the i-th successor of the input vertex
     *  \param vertex_it : input vertex iterator
     *  \param i-th : number of the successor
     */
    DGVertex_iterator<T> successor(const DGVertex_iterator<T>& vertex_it, int i) const
    {
      return (*((*vertex_it)->outputEdges_.at(i)))->endVertex_;
    }

    /*
     *  \brief return an iterator in the graph to the starting vertex of the input edge
     *  \param edge_it : input edge iterator
     *  \return an iterator in the graph to the starting vertex of the input edge
     */
    DGVertex_iterator<T> startVertex(const DGEdge_iterator<T>& edge_it) const
    {
      return (*edge_it)->startVertexIt();
    }

    /*
     *  \brief return an iterator in the graph to the starting vertex of the input edge
     *  \param edge_it : input edge iterator
     *  \return an iterator in the graph to the starting vertex of the input edge
     */
    DGVertex_iterator<T> endVertex(const DGEdge_iterator<T>& edge_it) const
    {
      return (*edge_it)->endVertexIt();
    }


    /*************** Graph modifications **************/

  public : 
    /**
     *  \brief clear the graph content
     */
    virtual void clear()
    {
      for(typename std::list< DGVertex<T>* >::iterator it = vertices_.begin(); it != vertices_.end(); it++) delete (*it);
      vertices_.clear();
      
      for(typename std::list< DGEdge<T>* >::iterator it = edges_.begin(); it != edges_.end(); it++) delete (*it);
      edges_.clear();      
    }

    /**
     *  \brief add a new vertex to the graph
     *  \return a pointer to the vertex added to the graph, or NULL if an error occured
     */
    DGVertex_iterator<T> addVertex(T val)
    {
      // create the new vertex
      DGVertex<T>* newVertex = new DGVertex<T>(val);

      // add the new vertex to the list of vertices
      vertices_.push_back(newVertex);

      // return an iterator to the new vertex
      DGVertex_iterator<T> it = vertices_.end();
      it--;
      return it;
    }

    /**
     *  \brief add a new edge between two vertices of the graph
     *  \param startV : starting vertex of the edge
     *  \param endV : starting vertex of the edge
     */
    void addEdge(DGVertex_iterator<T> startV, DGVertex_iterator<T> endV)
    {
      // create a new edge between both vertices
      DGEdge<T>* newEdge = new DGEdge<T>(startV, endV);

      // add the new edge to the list of edges
      edges_.push_back(newEdge);

      // get back iterator to the new edge
      DGEdge_iterator<T> it = edges_.end();
      it--;

      (*startV)->addOutputEdge(it);
      (*endV)->addInputEdge(it);
    }
    
    /**
     *  \brief remove a vertex from the graph
     *
     *  Every edge starting or ending on this vertex will be removed
     */
    void removeVertex(DGVertex_iterator<T> vertex)
    {
      // remove each edge going to the input vertex
      for(typename std::vector< DGEdge_iterator<T> >::iterator it = (*vertex)->inputEdges_.begin(); it != (*vertex)->inputEdges_.end(); it++) 
	{
	  delete (**it);
	  this->edges_.erase(*it);
	}

      // remove each edge coming from the input vertex
      for(typename std::vector< DGEdge_iterator<T> >::iterator it = (*vertex)->outputEdges_.begin(); it != (*vertex)->outputEdges_.end(); it++) 
	{
	  delete (**it);
	  this->edges_.erase(*it);
	}

      // delete the vertex
      delete *vertex;

      // remove the vertex from the list of vertices of the graph
      vertices_.erase( vertex );
    }

    /**
     *  \brief remove an edge from the graph between two vertices
     *  \return an iterator to the next element of the list
     */
    DGEdge_iterator<T> removeEdge(DGEdge_iterator<T> edge_it) 
    {
      // remove the edge from the list of edges
      DGEdge_iterator<T> resIt = this->edges_.erase(edge_it);

//       if(edge_it == NULL) { std::cout << "edge_it is NULL..." << std::endl; exit(1); }
      if(*edge_it == NULL) { std::cout << "*edge_it is NULL..." << std::endl; exit(1); }

      // get back iterators to both end points of the edge
      DGVertex_iterator<T> v1 = (*edge_it)->startVertex_;
      DGVertex_iterator<T> v2 = (*edge_it)->endVertex_;

      // remove the edge found from the list of edges coming from v1
      bool found1 = false;
      typename std::vector< DGEdge_iterator<T> >::iterator it1 = (*v1)->outputEdges_.begin();
      
      while(!found1 && it1 != (*v1)->outputEdges_.end())
	{
	  if(*it1 == edge_it)
	    {
	      (*v1)->outputEdges_.erase(it1);
	      found1 = true;
	    }
	}
      
      // remove the edge found from the list of edges going to v2
      bool found2 = false;
      typename std::vector< DGEdge_iterator<T> >::iterator it2 = (*v2)->inputEdges_.begin();
      
      while(!found2 && it2 != (*v2)->inputEdges_.end())
	{
	  if(*it2 == edge_it)
	    {
	      (*v2)->inputEdges_.erase(it2);
	      found2 = true;
	    }
	}

      return resIt;
    }

    /**
     *  \brief remove an edge from the graph between two vertices
     */
    void removeEdge(DGVertex_iterator<T> v1, DGVertex_iterator<T> v2) 
    { 
      bool found = false;
      typename std::list< DGEdge<T>* >::iterator it = this->edges_.begin();

      // search for an edge coming from v1 and going to v2 in the list of edges
      while(!found && it != this->edges_.end())
	{
	  if((*it)->startVertex_ == v1 && (*it)->endVertex_ == v2)
	    {
	      this->edges_.erase(it);
	      found = true;
	    }
	  else it++;
	}

      if(found)
	{
	  // remove the edge found from the list of edges coming from v1
	  bool found2 = false;
	  typename std::vector< DGEdge_iterator<T> >::iterator it2 = (*v1)->outputEdges_.begin();
	    
	  while(!found2 && it2 != (*v1)->outputEdges_.end())
	    {
	      if(*it2 == it)
		{
		  (*v1)->outputEdges_.erase(it2);
		  found2 = true;
		}
	    }
	  
	  // remove the edge found from the list of edges going to v2
	  bool found3 = false;
	  typename std::vector< DGEdge_iterator<T> >::iterator it3 = (*v2)->inputEdges_.begin();
	    
	  while(!found3 && it3 != (*v2)->inputEdges_.end())
	    {
	      if(*it3 == it)
		{
		  (*v2)->inputEdges_.erase(it3);
		  found3 = true;
		}
	    }      
	}
    }

    /**
     *  \brief create a new vertex with the input value and set it as the predecessor of a given vertex
     *  \param vertex_it : iterator to the node of the graph where the predecessor must be added
     *  \param val : the input value of the new vertex
     */
    void addPredecessor(DGVertex_iterator<T> vertex_it, const T& val) 
    { 
      // add the new vertex
      DGVertex_iterator<T> newV_it = this->addVertex(val);

      // add the edge between coming from the new vertex and going to vertex_it
      this->addEdge(newV_it, vertex_it);
    }

    /**
     *  \brief create a new vertex with the input value and set it as the successor of a given vertex
     *  \param vertex_it : iterator to the node of the graph where the successor must be added
     *  \param val : the input value of the new vertex
     */
    void addSuccessor(DGVertex_iterator<T> vertex_it, const T& val) 
    { 
      // add the new vertex
      DGVertex_iterator<T> newV_it = this->addVertex(val);

      // add the edge between coming from vertex_it and going to the new vertex
      this->addEdge(vertex_it, newV_it);
    }


    /**************** Utility functions ***************/

  private : 
    /**
     *  \brief recursive function for the tree building
     *  \return a pointer to the root of the current built tree
     *
     *  New instances of Nodes are created
     */
    Node<T>* toTreeRec(DGVertex_iterator<T> graph_vertex) const
    {
      // create a node from the value of the current vertex
      Node<T>* node = new Node<T>((*graph_vertex)->val());

      // calling the recursive function on each successor of the current vertex
      int N = this->nbSuccessors(graph_vertex);
      for(int i=0; i<N; i++)
	{
	  // get back the current successor
	  DGVertex_iterator<T> suc_vertex_it = this->successor(graph_vertex, i);
	  // create a node from the current successor
	  Node<T>* suc_node = this->toTreeRec(suc_vertex_it);
	  // add it to the tree
	  node->addSon(suc_node);
	}
      
      return node;
    }

  public : 
    /**
     *  \brief build a tree from the current graph
     *  \return a pointer to the root of the built tree
     *
     *  New instances of Nodes are created
     */
    Node<T>* toTree()
    {
      // check if the graph is not empty
      if(this->nbVertices() == 0) return NULL;

      // get back the "root" of the graph (if it exists)
      std::vector<DGVertex_iterator<T> > roots = this->rootVertices();

      // check if the root is unique
      if(roots.size() != 1) return NULL;
      DGVertex_iterator<T> root = roots.at(0);

      // build the tree from the unique "root" of the graph found
      return this->toTreeRec(root);
    }

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the graphe
     */
    virtual void display() 
    {
      std::cout << "Graphe vertices (" << this->vertices_.size() << ") " << std::endl;
//       for(typename std::list< DGVertex<T>* >::iterator it = this->vertices_.begin(); it != this->vertices_.end(); it++)
      std::cout << "Graphe edges (" << this->edges_.size() << ") " << std::endl;
    }
  };
};

#endif // MPC_GRAPH_HPP
