#ifndef MPC_STATE_HPP
#define MPC_STATE_HPP

#include "common/inc/Vector.hpp"
#include "common/inc/Config.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class State ******************/
                                    /**************************************************/

  /** \class State
   *  \brief State of a robotic system
   */
  class State : public Vector
  {
  protected : 

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    State();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    State(const State& s);

    /**
     *  \brief init constructor
     *  \param n : dimension of the state
     */
    State(int n);

    /**
     *  \brief destructor
     */
    virtual ~State();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    virtual State* copy() const = 0;


    /******************** Accessors *******************/
  
    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    virtual PositionXYTh posXYTh() const = 0;

    //   /**
    //    *  \brief retrieve a configuration from the current state
    //    *  \return the corresponding configuration
    //    */
    //   virtual Config cfg() const = 0;
  };

};
#endif // MPC_STATE_HPP


#ifndef MPC_PMSTATE_HPP
#define MPC_PMSTATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class PMState *****************/
                                    /**************************************************/

  /** \class PMState
   *  \brief State of a point mass system
   */
  class PMState : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    PMState();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    PMState(const PMState& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param sx : speed sx
     *  \param sy : speed sy
     */
    PMState(double x,double y,double sx,double sy);

    /**
     *  \brief destructor
     */
    virtual ~PMState();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    virtual PMState* copy() const { return new PMState(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    PMState(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[2]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[3]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(2); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(3); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual PMConfig cfg() const { return PMConfig(x(),y()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),0.0); }
  };

};
#endif // MPC_PMSTATE_HPP


#ifndef MPC_PM3DSTATE_HPP
#define MPC_PM3DSTATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class PM3DState ****************/
                                    /**************************************************/

  /** \class PM3DState
   *  \brief State of a 3D point mass system
   */
  class PM3DState : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    PM3DState();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    PM3DState(const PM3DState& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param z : position z
     *  \param sx : speed sx
     *  \param sy : speed sy
     *  \param sz : speed sz
     */
    PM3DState(double x,double y,double z,double sx,double sy,double sz);

    /**
     *  \brief destructor
     */
    virtual ~PM3DState();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    PM3DState* copy() const { return new PM3DState(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    PM3DState(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : position Z
     *  \return position Z
     */
    inline double& z() { return (*this)[2]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[3]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[4]; }

    /**
     *  \brief getter : speed SZ
     *  \return speed SZ
     */
    inline double& sz() { return (*this)[5]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : position Z
     *  \return position Z
     */
    inline double z() const { return this->at(2); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(3); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(4); }

    /**
     *  \brief getter : speed SZ
     *  \return speed SZ
     */
    inline double sz() const { return this->at(5); }

//     /**
//      *  \brief retrieve a configuration from the current state
//      *  \return the corresponding configuration
//      */
//     inline virtual PMConfig cfg() const { return PMConfig(x(),y()); }

//     /**
//      *  \brief return the position of the space (in W = R^3)
//      *  \return the position-time of the space (in W = R^3)
//      */
//     inline virtual PositionXYTh posXYTh() const { cerr << "PM3DState::posXYTh cannot be implemented!" << endl; exit(1); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),0.0); }

    /**
     *  \brief return the position of the space (in W = R^3)
     *  \return the position-time of the space (in W = R^3)
     */
    inline virtual PositionXYZ posXYZ() const { return PositionXYZ(x(),y(),z()); }
  };

};
#endif // MPC_PM3DSTATE_HPP


#ifndef MPC_CLSTATE_HPP
#define MPC_CLSTATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class CLState *****************/
                                    /**************************************************/

  /** \class CLState
   *  \brief State of a car-like system
   */
  class CLState : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    CLState();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    CLState(const CLState& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param ph : steering angle
     *  \param s : speed s
     */
    CLState(double x,double y,double th,double ph,double s);

    /**
     *  \brief destructor
     */
    virtual ~CLState();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    CLState* copy() const { return new CLState(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    CLState(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : steering angle PH
     *  \return heading PH
     */
    inline double& ph() { return (*this)[3]; }

    /**
     *  \brief getter : linear speed S
     *  \return linear speed S
     */
    inline double& s() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : steering angle PH
     *  \return heading PH
     */
    inline double ph() const { return this->at(3); }

    /**
     *  \brief getter : linear speed S
     *  \return linear speed S
     */
    inline double s() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual CLConfig cfg() const { return CLConfig(x(),y(),th()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }
  };

};
#endif // MPC_CLSTATE_HPP


#ifndef MPC_SSSTATE_HPP
#define MPC_SSSTATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class SSState *****************/
                                    /**************************************************/

  /** \class SSState
   *  \brief State of a space ship system
   */
  class SSState : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    SSState();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    SSState(const SSState& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param sx : speed sx
     *  \param sy : speed sy
     *  \param om : angular velocity
     */
    SSState(double x,double y,double th,double sx,double sy,double om);

    /**
     *  \brief destructor
     */
    virtual ~SSState();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    SSState* copy() const { return new SSState(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    SSState(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double& sx() { return (*this)[3]; }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double& sy() { return (*this)[4]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& om() { return (*this)[5]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : speed SX
     *  \return speed SX
     */
    inline double sx() const { return this->at(3); }

    /**
     *  \brief getter : speed SY
     *  \return speed SY
     */
    inline double sy() const { return this->at(4); }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double om() const { return this->at(5); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual SSConfig cfg() const { return SSConfig(x(),y(),th()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }
  };

};
#endif // MPC_SSSTATE_HPP


#ifndef MPC_DDSTATE_HPP
#define MPC_DDSTATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class DDState *****************/
                                    /**************************************************/

  /** \class DDState
   *  \brief State of a differential drive system
   */
  class DDState : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    DDState();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    DDState(const DDState& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param omr : right wheel velocity
     *  \param oml : left wheel velocity
     */
    DDState(double x,double y,double th,double omr,double oml);

    /**
     *  \brief destructor
     */
    virtual ~DDState();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    DDState* copy() const { return new DDState(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    DDState(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : right wheel velocity OMR
     *  \return right wheel velocity OMR
     */
    inline double& omr() { return (*this)[3]; }

    /**
     *  \brief getter : left wheel velocity OML
     *  \return left wheel velocity OML
     */
    inline double& oml() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : right wheel velocity OMR
     *  \return right wheel velocity OMR
     */
    inline double omr() const { return this->at(3); }

    /**
     *  \brief getter : left wheel velocity OML
     *  \return left wheel velocity OML
     */
    inline double oml() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual DDConfig cfg() const { return DDConfig(x(),y(),th()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }
  };

};
#endif // MPC_DDSTATE_HPP


#ifndef MPC_DD2STATE_HPP
#define MPC_DD2STATE_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class DD2State *****************/
                                    /**************************************************/

  /** \class DD2State
   *  \brief State of a wheelchair system
   */
  class DD2State : public State
  {
  protected : 

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    DD2State();

    /**
     *  \brief copy constructor
     *  \param p : the control to copy
     */
    DD2State(const DD2State& s);

    /**
     *  \brief init constructor
     *  \param x : position x
     *  \param y : position y
     *  \param th : heading th
     *  \param s : linear velocity
     *  \param om : angular velocity
     */
    DD2State(double x,double y,double th,double s,double om);

    /**
     *  \brief destructor
     */
    virtual ~DD2State();

    /**
     *  \brief return a copy of the current state
     *  \return a copy of the current state
     */
    DD2State* copy() const { return new DD2State(*this); }

  protected : 
    /**
     *  \brief n constructor
     *  \param n : the size of the state vector
     */
    DD2State(int n);

    /******************** Accessors *******************/

  public : 
    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double& x() { return (*this)[0]; }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double& y() { return (*this)[1]; }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double& th() { return (*this)[2]; }

    /**
     *  \brief getter : liner velocity S
     *  \return linear velocity S
     */
    inline double& s() { return (*this)[3]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& om() { return (*this)[4]; }

    /**
     *  \brief getter : position X
     *  \return position X
     */
    inline double x() const { return this->at(0); }

    /**
     *  \brief getter : position Y
     *  \return position Y
     */
    inline double y() const { return this->at(1); }

    /**
     *  \brief getter : heading TH
     *  \return heading TH
     */
    inline double th() const { return this->at(2); }

    /**
     *  \brief getter : linear velocity S
     *  \return linear velocity S
     */
    inline double s() const { return this->at(3); }

    /**
     *  \brief getter : liner velocity S
     *  \return linear velocity S
     */
    inline double& v() { return (*this)[3]; }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double& w() { return (*this)[4]; }

    /**
     *  \brief getter : linear velocity S
     *  \return linear velocity S
     */
    inline double v() const { return this->at(3); }

    /**
     *  \brief getter : angular velocity W
     *  \return angular velocity W
     */
    inline double w() const { return this->at(4); }

    /**
     *  \brief getter : angular velocity OM
     *  \return angular velocity OM
     */
    inline double om() const { return this->at(4); }

    /**
     *  \brief retrieve a configuration from the current state
     *  \return the corresponding configuration
     */
    inline virtual DD2Config cfg() const { return DD2Config(x(),y(),th()); }

    /**
     *  \brief return the position of the space (in W = R^2 x S)
     *  \return the position-time of the space (in W = R^2 x S)
     */
    inline virtual PositionXYTh posXYTh() const { return PositionXYTh(x(),y(),th()); }
  };

};
#endif // MPC_DD2STATE_HPP


#ifndef MPC_WCSTATE_HPP
#define MPC_WCSTATE_HPP

#define WCState DD2State

#endif // MPC_WCSTATE_HPP


#ifndef MPC_MANASTATE_HPP
#define MPC_MANASTATE_HPP

#define ManaState DD2State

#endif // MPC_MANASTATE_HPP

