#ifndef _UTILS_HPP_
#define _UTILS_HPP_

#include "common/inc/Vec.hpp"

#include <string>
#include <vector>
#include <cmath>
#include <complex>
#include <cstdlib>
#include <stdlib.h>
#include <iostream>
using namespace std;

namespace mpc
{
  // Direction types
  typedef enum {
    LEFT   =  1, // left (trigonometric sens)
    RIGHT  = -1  // right (antitrigonometric sens)
  } leftright_t;

  namespace cd
  {
    // Cardinal directions
    typedef enum {
      E   =  1, // east
      NE  =  2, // north east
      N   =  3, // north
      NW  =  4, // north west
      W   =  5, // west
      SW  =  6, // south west
      S   =  7, // south
      SE  =  8  // south east
    } carddir_t;
  };

  /********************************* characters treatment *******************************/

  bool isAlphaNum(char c);                              // Function which return true if the input charachter is alphanumeric
  bool isNum(char c);                                   // Function which return true if the input charachter is numeric
  string addSuffixe(string s,string suf);               // add a suffixe before ".pgm" extension to a filename
  char* getExtFilename(const char* filename);                 // return the extension type of the filename
  bool isExtFilename(const char* filename, const char* ext);        // return true if filename has the extension type ext


  /********************************** numbers treatment *********************************/

  double min(double a,double b);                        // return the minimum of both values
  double max(double a,double b);                        // return the maximum of both values
  double valAbs(double v);                              // return the absolute value of v
  //void int2String(int nb,char* chres);                  // Function which convert an integer into a string
  void int2String(int nb,char* chres);
//   int round(float nb);                                  // compute the round integer from a float value

  /**
   *  \brief Function which convert an double into a string
   *  \param nb : the number to convert
   *  \param chres : the resulting string
   *  \param fl : the number of character after the point
   */
  void double2String(double nb,char* chres,int fl);

  /***************************** tables and vectors treatment ***************************/

  /**
   *  \brief copy the content of a table
   *  \param from : the table to copy
   *  \param to : the resulting table
   *  \param n : the size of both tables
   *
   *  both from and to must be allocated first and have the same size
   */
  void copyTab(double* from,double* to,int n);

  /**
   *  \brief display a table of doubles
   *  \param tab : the table to display
   *  \param n : the size of the table
   */
  void displayTab(const double* tab,int n);

  /**
   *  \brief display a table of integers
   *  \param tab : the table to display
   *  \param n : the size of the table
   */
  void displayTab(int* tab,int n);

  void inverseTab(char* tab,int n);                     // Function which inverse a table
  void displayVectDble(vector<double>* v);              // display vector v
  void sortVectDble(vector<double>& v);                 // sort the vector of double values
  void delDoubleValues(vector<double>& v);              // delete the double values from the input vector
  void mergeAndSortVectDble(vector<double>& v1,vector<double>& v2); // sort and merge two two vectors and return the result in v1
  void getValuesBetween(double binf,double bsup,        // return a vector vres containing the values of v1 include in the 
			const vector<double>& v1,       // interval [binf;bsup]
			vector<double>& vres);          // the vector v1 is supposed to be sorted

  /**
   *  \brief compute the index of a 1D tab from a MultiD tab
   *  \param dim : the dimension of the multidim tab
   *  \param sampMD : the sampling of the multidim tab
   *  \param indMD : the input indices in the multidim tab
   *  \param ind1D : the resulting index in the 1D tab
   */
  void indMD_2_ind1D(int dim,int* sampMD,int* indMD,int& ind1D);

  /**
   *  \brief compute the index of a MultiD tab from a 1D tab
   *  \param dim : the dimension of the multidim tab
   *  \param sampMD : the sampling of the multidim tab
   *  \param ind1D : the input index in the 1D tab
   *  \param indMD : the resulting indices in the multidim tab
   */
  void ind1D_2_indMD(int dim,int* sampMD,int ind1D,int* indMD);


  /************************************ curve smoothing *********************************/

  /**
   *  \brief smooth a vector of double values
   *  \param v : the vector to smooth
   *  \param n : (2n+1) value used to smooth each parameter
   */
  void smoothVec(vector<double>& v,int n);

  void PascTabV2(vector<double>& v,unsigned int n);     // return n-th of Pascal Triangle
  void PascTab1stV2(vector<double>& v,unsigned int n);  // return n-th of 1st derivative Pascal Triangle
  void deriv1stTabV2(vector<double>& vres,vector<double>& v,vector<double>& vp);         
  //return the first derivative of a table of double

  /*********************************** random functions *********************************/

  int initRandom();                              // initialize the random function and return the number of the sequence used
  int initRandom(int seq);                       // initialize the random function from a defined sequence
  double getRandom();                                   // return a real random value between 0 and 1
  double getRandomM11();                                // return a real random value between -1 and 1
  int getIntRandom(int binf,int bsup);                  // return an integer random value between bsup and binf
  double getDoubleRandom(double binf,double bsup);      // return a double random value between bsup and binf


  /************************************ time functions **********************************/

  double getTimeOfDay();                                // return a real gettimeofday function
  int getTimeOfDayUSec();                               // return only the usec value of the gettimeofday function


  /******************************** intersection functions ******************************/

  void coefEqu(Vec2D p,Vec2D v,double* a,double* b,double *c);
  // compute the coefficients (a,b,c) of the equation given by a point p and a vector v
  bool intersecLine2D(Vec2D p1, Vec2D v1, Vec2D p2, Vec2D v2, Vec2D& pres); 
  // compute the intersection between the line defined by (p1,v1) and the line defined by (p2,v2)
  bool intersecSeg2D(Vec2D p1, Vec2D p2, Vec2D p3, Vec2D p4, Vec2D& pres); 
  // compute the intersection between the segments [p1;p2] and [p3;p4]
  // return false if there is no intersection

  // bool intersecLineCircle2D(Vec p1, Vec v1, Vec p2, double rad2, Vec& itsc1, Vec& itsc2);
  // // compute the two intersections itsc1 and itsc2 between the line defined by (p1,v1) and the circle of center p2 and radius rad2
  bool intersecLineCircle2D(Vec p1, Vec v1, Vec p2, double rad2, Vec& itsc1);
  // compute the nearest intersection itsc1 between the line defined by (p1,v1) and the circle of center p2 and radius rad2 from the point p1


  /*********************************** Angle functions **********************************/

  double normalizedAngle(double a);                     // return the value in ]-PI;PI] of the angle a
  double computeAngleFromBetween(Vec pos1,Vec pos2,Vec pos3,leftright_t lr);
  // compute the angle from pos1 between pos2 and pos3
  double computeMinDiffAngle(double a1,double a2);      // compute the minimal absolute difference between angles a1 and a2 (a1 - a2)
  double computeCurvature(Vec2D p1,double dir1,Vec2D p2,double dir2); // compute an approximation of the curvature between two configs
  bool isInTheFieldOfView(Vec2D posFrom,double dirFrom,Vec2D posTgt,double dirTgt,double angleFOV);
  // return true if Tgt is in the field of view of From with a FOV angle FOV
  double degToRad(const double d);                      // convert degrees to radians
  double radToDeg(const double r);                      // convert radians to degrees


  /******************************** Coordinate functions ********************************/

  Vec2D coordInFrame(Vec2D p,Vec3D f);                  // return the coordinate of the point p in the frame f
  // from frame (0.0)
  Vec2D coordFromFrame(Vec2D p,Vec3D f);                // return the coordinate of the point p in the frame (0.0)
  // from frame f


  /******************************* Integration functions ********************************/
  
  typedef double (equation_t)(void* param, double t);
  typedef double (function_t)(double t, void* param);

  // compute a Runge Kutta integration of the function fct over [binf;bsup] knowing the initial value init
  double RungeKutta(equation_t* fct, void* param, double init, double binf, double bsup, int order);

  //Quadrature tools
  void gauleg(const double x1, const double x2, double x[], double w[], int n);
  double qgaus(function_t* fct, void* param, const double a, const double b);


  /*************************** Equations solvation functions ****************************/

  // compute the root of a first order polynom ax + b = 0
  // return the number of roots found
  int rootPoly1st(double a,double b,double& r1);

  // compute the discriminant of a second order polynom ax^2 + bx + c = 0
  double discriminant(double a,double b,double c);

  // compute the real roots of a second order polynom ax^2 + bx + c = 0
  // return the number of real roots found
  int realRootPoly2nd(double a,double b,double c,double& r1,double& r2);

  // compute the complex roots of a second order polynom ax^2 + bx + c = 0
  // return the number of complex roots found
  int complexRootPoly2nd(double a,double b,double c,complex<double>& r1,complex<double>& r2);

  // compute the real roots of a second order polynom ax^3 + bx^2 + cx + d = 0
  // return the number of real roots found
  int realRootPoly3rd(double a,double b,double c,double d,double& r1,double& r2,double& r3);

  // compute the complex roots of a second order polynom ax^3 + bx^2 + cx + d = 0
  // return the number of complex roots found
  int complexRootPoly3rd(double a,double b,double c,double d,complex<double>& r1,complex<double>& r2,complex<double>& r3);


  /******************************** Intervals functions *********************************/

  /**
   *  \brief search for the interval of a regular sampling to which belongs a value v
   *  \param v : the searched value
   *  \param minB : the minimal bound of the sampling
   *  \param maxB : the maximal bound of the sampling
   *  \param nbSp : the number of values of the sampling (number of intervals + 1)
   *  \param minI : the resulting index of the lowest bound of the searched interval
   *  \param maxI : the resulting index of the highest bound of the searched interval
   *
   *  If v is outside of the interval [minB;maxB], minI and maxI are bot equals
   *  to the nearest extremal bound of [minB;maxB]
   *
   *  If v is equals to a bound of the sampling intervals, the resulting indices are [ind(v),ind(v)+1]
   */
  void searchForItvSampling(double v,double minB,double maxB,int nbSp,int& minI,int& maxI);


  /******************************* Input/output functions *******************************/

  void wait_for_key();                                  // wait for the user to press a key of the keyboard


  /********************************* Distance functions *********************************/

  /**
   *  \brief compute the minimal time needed to travel a distance given constraints
   *  \param dist : the distance to travel
   *  \param v : the initial velocity
   *  \param vmax : the maximal velocity
   *  \param amax : the maximal acceleration
   *  \return the minimal time needed to travel dist given constraints
   */
  double minTravelTime(double dist,double v,double vmax,double amax);

  /**
   *  \brief compute the maximal time needed to travel a distance given constraints
   *  \param dist : the distance to travel
   *  \param v : the initial velocity
   *  \param vmax : the maximal velocity
   *  \param amax : the maximal acceleration
   *  \return the maximal time needed to travel dist given constraints
   *  \return INFINITY if the robot has the time to stop before to reach that position
   */
  double maxTravelTime(double dist,double v,double vmax,double amax);

  /**
   *  \brief compute the minimal time and distance to reach the speed vgoal from vinit given constraints
   *  \param vinit : the initial velocity
   *  \param vgoal : the goal velocity
   *  \param amax : the maximal acceleration
   *  \param tres : the resulting time
   *  \param dres : the resulting distance
   */
  void minTDtoReachSpd(double vinit,double vgoal,double amax,double& tres,double& dres);

  /**
   *  \brief compute the minimal distance travelled during a given time and constraints
   *  \param t : the duration of the motion
   *  \param v : the initial velocity
   *  \param vmax : the maximal velocity
   *  \param amax : the maximal acceleration
   *  \return the minimal distance travelled during a given time and constraints
   */
  double minTravelDist(double t,double v,double vmax,double amax);

  /**
   *  \brief compute the maximal distance travelled during a given time and constraints
   *  \param t : the duration of the motion
   *  \param v : the initial velocity
   *  \param vmax : the maximal velocity
   *  \param amax : the maximal acceleration
   *  \return the maximal distance travelled during a given time and constraints
   */
  double maxTravelDist(double t,double v,double vmax,double amax);
};
#endif
