#ifndef MPC_VEC_HPP
#define MPC_VEC_HPP

#include <vector>
#include <string>

#include "common/inc/Utils.hpp"
 
                                    /**************************************************/
                                    /****************** Class Vector ******************/
                                    /**************************************************/
namespace mpc
{

  /** \class Vector
   *  \brief Generic double vector class
   */
  class Vector
  {
  protected : 
    std::vector<double> coord_;           /**< vector of coordinates */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    Vector();
    /**
     *  \brief copy constructor
     *  \param p : the position to copy
     */
    Vector(const Vector& v);
    /**
     *  \brief init constructor
     *  \param n : number of elements of the vector
     */
    Vector(int n);
    /**
     *  \brief destructor
     */
    virtual ~Vector();

    /******************** Accessors *******************/
  
    /**
     *  \brief accessor of the i-th coordinate
     *  \param i : index of the coordinate to return
     *  \return the i-th coordinate of the vector
     */
    inline virtual double& operator[](int i) { return coord_.at(i); }

    /**
     *  \brief accessor of the i-th coordinate
     *  \param i : index of the coordinate to return
     *  \return the i-th coordinate of the vector
     */
    inline virtual double at(int i) const { return coord_.at(i); }

    /**
     *  \brief size of the vector
     *  \return the size of the vector
     */
    inline virtual int size() const { return coord_.size(); }

    /**
     *  \brief equality test
     *  \param v2 : the second vector to test equality with
     *  \return true if both vectors are equals, false else
     */
    virtual bool equals(const Vector& v2) const;

    /******************* Operators ********************/

    /**
     *  \brief operator = : assignation
     *  \param v2 : the vector to assign to the current one
     */
    void operator=(const Vector& v2);

    /**
     *  \brief operator + : sumation
     *  \param v2 : the second vector to add
     *  \return the sum of both vectors
     */
    Vector operator+(const Vector& v2) const;

    /**
     *  \brief operator + : substraction
     *  \param v2 : the second vector to substract
     *  \return the substraction of both vectors
     */
    Vector operator-(const Vector& v2) const;

    /**
     *  \brief operator == : equality test
     *  \param v2 : the second vector to test equality with
     *  \return true if both vectors are equals, false else
     */
    bool operator==(const Vector& v2) const;

    /**
     *  \brief operator * : scalar multiplication
     *  \param d : the scalar to multiply with
     *  \return the product of the current vector and the scalar
     */
    Vector operator*(double d) const;

    /**
     *  \brief operator * : dot product
     *  \param v2 : the vector to multiply with
     *  \return the dot product of both vectors
     */
    double operator*(const Vector& v2) const;

    /**
     *  \brief operator / : division by a scalar
     *  \param d : the scalar to divide with
     *  \return the division result
     */
    Vector operator/(double d) const;

    /**
     *  \brief operator += : sumation
     *  \param v2 : the second vector to add
     */
    void operator+=(const Vector& v2);

    /**
     *  \brief operator += : substraction
     *  \param v2 : the second vector to substract
     */
    void operator-=(const Vector& v2);  

    /**
     *  \brief operator *= : scalar multiplication
     *  \param d : the scalar to multiply with
     */
    void operator*=(double d);

    /**
     *  \brief operator /= : division by a scalar
     *  \param d : the scalar to divide with
     */
    void operator/=(double d);

    /***************** Math functions *****************/

    /**
     *  \brief norm of the vector
     *  \return the norm of the vector
     */
    double norm() const;

    /**
     *  \brief square norm of the vector
     *  \return the square norm of the vector
     */
    double sqNorm() const;

    /**
     *  \brief compute the normalized vector from the current one
     *  \return the normalized vector
     */
    Vector normalized() const;

    /**
     *  \brief normalize the current vector
     */
    void normalize();

    /*************** Display functions ****************/
  
    /**
     *  \brief display the vector's coordinates
     */
    virtual void display() const;

    /**
     *  \brief display the vector's coordinates
     *  \param number : the number of digit used to display each value
     */
    virtual void display(int number) const;

    /**
     *  \brief convert the vector into a string
     *  \return the resulting string
     */
    virtual const char* toString() const;
  };

};
#endif
