#ifndef MPC_MANASIMPLECONTROLLER_HPP
#define MPC_MANASIMPLECONTROLLER_HPP

#include "controller/inc/SimpleController.hpp"

#include "system/inc/ManaSys.hpp"
// #include "trajGen/inc/WCTrajGen.hpp"
#include "trajGen/inc/ManaTrajGen.hpp"

namespace mpc
{

                                    /**************************************************/
                                    /*********** Class ManaSimpleController ***********/
                                    /**************************************************/

  /** \class ManaSimpleController
   *  \brief Simple controller for a Mana robotic system
   */
  class ManaSimpleController : public SimpleController
  {
  public : 
    static bool fixFinalTH;                            /**< true if the final heading must be fixed */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Only used by subclasses
     */
    ManaSimpleController();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system handled
     *  \param dt : the controller period
     *  \param debug : the debugging level
     */
    ManaSimpleController(ManaSys* sys,double dt = 0.5, int debug=0);

    /**
     *  \brief destructor
     */
    virtual ~ManaSimpleController();
  

    /*************** Getters and Setters **************/

  protected : 
    /**
     *  \brief return a pointer to the Mana trajectory generator
     *  \return a pointer to the Mana trajectory generator
     */
    ManaTrajGen* trajgen();

    /**
     *  \brief return a pointer to the Mana control trajectory
     *  \return a pointer to the Mana control trajectory
     */
    inline const ManaPCTraj* ctraj() const { return ((ManaPCTraj*) ctraj_); }

    /**
     *  \brief return a pointer to the Mana state trajectory
     *  \return a pointer to the Mana state trajectory
     */
    inline const ManaPSTraj* straj() const { return ((ManaPSTraj*) straj_); }

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    virtual const ManaSys* sys() const;


    /*************** Control Computation **************/

  protected : 
    /**
     *  \brief Initialize the parameters of the Mana trajectory generator
     */
    virtual void initManaTrajGen();

    /**
     *  \brief compute the control trajectory to apply to go from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the goal state
     *  \return 0 if the process succeeded, -1 else
     */
    int computePCTraj(const ManaState& s1,const ManaState& s2);

  public : 
    /**
     *  \brief compute the control to apply to go from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the goal state
     *  \return the resulting control
     */
    ManaControl computeControl(const ManaState& s1,const ManaState& s2);

    /**
     *  \brief compute the control to apply to go from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the goal state
     *  \return 0 if the process succeeded, -1 else
     */
    ManaPCTraj* computeCTraj(const ManaState& s1,const ManaState& s2);


    /***************** Debug / Display ****************/

  public : 
    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     * 
     *  A new instance of ManaPCTraj is returned
     */
    virtual const ManaPCTraj* getPCTraj();

    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     * 
     *  A new instance of ManaPSTraj is returned
     */
    virtual const ManaPSTraj* getPSTraj();
  };

};

#endif // MPC_MANASIMPLECONTROLLER_HPP
