#ifndef MPC_DDTRAJDEFORM_HPP
#define MPC_DDTRAJDEFORM_HPP

#include "deformation/inc/TrajDeform.hpp"
#include "motion/inc/STrajectory.hpp"
#include "system/inc/DiffDriveSys.hpp"
#include "trajGen/inc/DDTrajGen.hpp"

#include <list>

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class DDTrajDeform ***************/
                                    /**************************************************/

  /** \class DDTrajDeform
   *  \brief Trajectory deformer TEDDY for a Differential Drive system
   */
  class DDTrajDeform : public TrajDeform
  {
  public : 
    std::list <Vec_DP*>* listCtrl_;                    /**< sequence of controls used to generate the trajectory */
    std::list <int>* feaCtrl_;                         /**< feasibility of the sequence of controls */

    bool usePreviousParam_;                            /**< true if the previous parameters for trajectory generation have to be used */

    Vec_DP constraints_;                               /**< vector of constraints used for the trajectory generation process */

    std::vector<DDStateT> itdTraj_;                    /**< intermediate trajectory (after external forces application) */
    std::vector<DDStateT> itdTraj2_;                   /**< intermediate trajectory (after time adjustment) */

    /********** Constructors and destructors **********/
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    DDTrajDeform();

  public : 
    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     *  \param typeS : pointer to the table of state types (for simulation display)
     *  \param tns : pointer to the table of nearest states (for simulation display)
     *  \param tfi : pointer to the table of internal forces (for simulation display)
     *  \param tfe : pointer to the table of external forces (for simulation display)
     */
    DDTrajDeform(DDSTraj* traj,Environment* env,DiffDriveSys* sys);

    /**
     *  \brief init constructor
     *  \param traj : the initial trajectory to deform
     *  \param env : the environment in which the system evolve
     *  \param sys : the features of the robot system used
     *  \param typeS : pointer to the table of state types (for simulation display)
     *  \param tns : pointer to the table of nearest states (for simulation display)
     *  \param tfi : pointer to the table of internal forces (for simulation display)
     *  \param tfe : pointer to the table of external forces (for simulation display)
     */
    DDTrajDeform(DDSTraj* traj,Environment* env,DiffDriveSys* sys,int** typeS,StateTime*** tns,StateTime** tfi,StateTime** tfe);

    /**
     *  \brief destructor
     */
    virtual ~DDTrajDeform();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline DDSTraj* traj() const { return ((DDSTraj*)(this->traj_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline DiffDriveSys* sys() const { return ((DiffDriveSys*)(this->sys_)); }

    /*************** Forces computation ***************/

  protected : 
    /**
     *  \brief computing of the internal forces
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     */
    DDStateT computeFInt(const DDStateT& prev, const DDStateT& cur, const DDStateT& next, bool last, Vec_DP** param, int *fea);

    /**
     *  \brief compute the total external forces
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     */
    PositionXYT computeFExt(const StateTime& cur, Environment* env, double time, bool last);

    /**
     *  \brief compute the total forces
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     *  \param last : true if the current node is the last node of the trajectory
     */
    CLStateT computeForces(const DDStateT& prev, const DDStateT& cur, const DDStateT& next, Environment* env, double time, bool last);

    /**
     *  \brief apply the internal forces on the current node
     *  \param i : index of the current node on the trajectory
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param next : the next node of the trajectory
     *  \param time : the current time
     */
    virtual void applyFInt(int i,const DDStateT& prev, DDStateT& cur, const DDStateT& next, double time);

    /**
     *  \brief compute the total external forces
     *  \param i : index of the current node on the trajectory
     *  \param cur : the current node of the trajectory
     *  \param env : the environment of the system
     *  \param time : the current time
     */
    virtual void applyFExt(int i,DDStateT& cur, Environment* env, double time);


    /******* Connectivity and Collision checking ******/

  protected : 
    /**
     *  \brief check the connectivity of the trajectory
     *  \param curTime : the current elapsed time
     *  \return the time to the first disconnection found (-1 if no disconnection has been found)
     */
    double checkConnectivity(double curTime);

    /**
     *  \brief check the connectivity between two nodes of the trajectory
     *  \param i : 
     *  \param oldCur : 
     *  \param prev : the previous node of the trajectory
     *  \param cur : the current node of the trajectory
     *  \param force : the resulting force to apply to restore the connectivity
     *
     *  SHOULD NOT BE USED!!!
     */
    void checkConnectivity(int i, const DDStateT& oldCur, const DDStateT& prev, const DDStateT& cur, DDStateT& force);
  
    /**
     *  \brief adjust the time of the nodes of the trajectory to maintain its consistency
     *  \param curStep : index of the node already reached
     */
    void adjustTime(int curStep);


    /******************* Resampling *******************/
  
  protected : 
    /**
     *  \brief resample the trajectory
     *  \param curStep : index of the node already reached
     */
    virtual void resample(int curStep);


    /*************** Deformation process **************/

  protected : 
    /**
     *  \brief init the deformation and check the temporal space between two consecutive states
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual void initDeform(int curStep);
  
    /**
     *  \brief finish the deformation
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int finishDeform(double time,int curStep);

  public : 
    /**
     *  \brief deform the current trajectory
     *  \param time : the current time
     *  \param curStep : index of the node of the trajectory that has been already reached
     */
    virtual int deform(double time, int curStep);

    /*************** Feedback control methods ****************/

    /**
     *  \brief init the list of feedback controls
     *  \param listCtrl : the input list of feedback controls
     */
    void initListCtrl(std::list <Vec_DP*>* listCtrl);

    /***************** Trajectory computation ****************/

  protected : 
    /**
     *  \brief initialize the trajectory generator and compute a trajectory from prev to next
     *  \param prev : the starting node of the trajectory generation
     *  \param next : the goal node of the trajectory generation
     *  \param carModel : the trajectory generator used
     *  \param param : the vector of parameters used for the trajectory generation
     *  \param dt : resulting duration of the generated trajectory
     *  \return the trajectory generator return status
     */
    int genTraj(const DDStateT& prev,const DDStateT& next,DDTrajGen& carModel,Vec_DP** param,double& dt) const;

    /**
     *  \brief compute the intermediate node to add between prev and next (strict case)
     *  \param prev : the starting node of the trajectory generation
     *  \param next : the goal node of the trajectory generation
     *  \param res : the resulting intermediate node
     *  \param param : the vector of parameters used for the trajectory generation
     *  \return 0 -> convergence has been ensured
     *  \return 1 -> convergence has not been ensured
     *
     *  strict case : compute the node only if the convergence has been ensured
     */
    int computeItdStateT_Strict(const DDStateT& prev,const DDStateT& next,DDStateT& res,Vec_DP** param);

    /**
     *  \brief compute the intermediate node to add between prev and next (average case)
     *  \param prev : the starting node of the trajectory generation
     *  \param next : the goal node of the trajectory generation
     *  \param res : the resulting intermediate node
     *  \param param : the vector of parameters used for the trajectory generation
     *  \return 0 -> near convergence has been ensured
     *  \return 1 -> far convergence has been ensured
     *  \return 2 -> convergence hasn't been ensured
     *
     *  average case : compute a intermediate node even if the convergence hasn't been ensured
     */
    int computeItdStateT_Avg(const DDStateT& prev,const DDStateT& next,DDStateT& res,Vec_DP** param);

    /**************** Trajectory modifications ***************/

  public : 
    /**
     *  \brief return the state of the deformed trajectory at time t
     *  \param t : the input time
     *  \return the state of the deformed trajectory at time t
     */
    DDStateT getStateTraj(double t) const;
  
  protected : 
    /**
     *  \brief locate the loops along the trajectory
     *  \param curStep : index of the node already reached by the system
     *  \param vecLoop : the resulting vector of loops
     *  \return the number of loops found (put in the vector vecLoop)
     */
    int locateLoop(int curStep,std::vector<int*> &vecLoop) const; 

    /**
     *  \brief remove the loops from the trajectory
     *  \param curStep : index of the node already reached by the system
     */
    void removeLoop(int curStep);

    /**
     *  \brief insert a state in the deformed trajectory
     *  \param s : the state to insert
     *  \param pos : its position in the trajectory's nodes vector
     *  \param param : the vector of parameters for the trajectory generation used to reach the new state
     *  \param fea : the feasibility status of the new inserted node
     */
    void insertStateInTraj(const DDStateT& s,int pos,Vec_DP** param,int fea);

    /**
     *  \brief delete one or several states from the trajectory including start but not including end
     *  \param start : the first node to delete
     *  \param end : the node after the last node to delete
     */
    void deleteStateFromTraj(int start,int end);

    /******************* Utility functions *******************/
  
  public : 
    /**
     *   \brief initialize the vector of constraints
     */
    void initConstraints();

    /**
     *  \brief display the refined trajectory in an openGL viewer
     */
    void displayTraj() const;
  };

};
#endif // MPC_DDTRAJDEFORM_HPP
