#ifndef DEM_HPP
#define DEM_HPP

//#include "genom/inc/DEM_reader.hpp"
#include "environment/inc/Map2D.hpp"

namespace mpc
{

/**
 * \enum dem_cell_t
 * \brief Enumeration of the different dem cell types
 *
 * DEM_CELL_EMPTY  : the cell hasn't been alredy filled
 * DEM_CELL_FILLED : the cell has been alredy filled
 */
typedef enum {
  DEM_CELL_EMPTY   =  0,
  DEM_CELL_FILLED  =  1
} dem_cell_t;

                                    /**************************************************/
                                    /******************** Class DEM *******************/
                                    /**************************************************/

  /** \class DEM
   * \brief Digital Elevation Map
   */
  class DEM : public Map2D
  {
  protected : 
    float zOrigin_;           /**< Elevation of the '0' value */
    float zScale_;            /**< Scale to store the elevation */
    
    float** z_;               /**< Altitude of each cell */
    
    int** nb_points_;         /**< Number of points fused in the dem */
//     float** sigma_z_;         /**< Standard deviation */
//     float** confidence_;      /**< Confidence of the altitude computed in each cell */

    static float colDepth_;   /**< depth of the colors used to convert the DEM into images */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    DEM();
    
    /**
     *  \brief copy constructor
     *  \param dem : the dem to copy
     */
    DEM(const DEM& dem);

    /**
     *  \brief init constructor
     *  \param nbCols : number of columns of the DEM table
     *  \param nbLines : number of lines of the DEM table
     *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param zOrigin : Elevation of the '0' value 
     *  \param xScale : X size of a cell in meters
     *  \param yScale : Z size of a cell in meters
     *  \param zScale : Scale to store the elevation
     */
    DEM(int nbCols, int nbLines, float xOrigin, float yOrigin, float zOrigin, float xScale, float yScale, float zScale);

    /**
     *  \brief destructor
     */
    virtual ~DEM();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of DEM is allocated
     */
    virtual DEM* copy() const;


    /***************** Initialization *****************/

  private : 
    /*
     *  \brief init tables from their dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTables(int nbCols,int nbLines);
    
    /*
     *  \brief delete all the content of the DEM tables
     */
    void clearTables();


    /******************** Accessors *******************/

  public : 
    /**
     *  \brief get back the altitude of a cell of the DEM
     *  \param i : number of the column
     *  \param j : number of the line
     *  \return the altitude of a cell on the DEM
     */
    float getZ(int i,int j) const;

    /**
     *  \brief get back the altitude of a position (x,y) on the DEM
     *  \param x : absica
     *  \param y : ordinate
     *  \return the altitude of a position (x,y) on the DEM
     */
    float getZ(float x,float y) const;

    /**
     *  \brief get back the number of points seen a cell of the DEM
     *  \param i : number of the column
     *  \param j : number of the line
     *  \return the number of points seen in a cell of the DEM
     */
    float getNbP(int i,int j) const;

    /**
     *  \brief Return the Elevation of the '0' value
     *  \return the Elevation of the '0' value
     */
    float zOrigin() const;
        
    /**
     *  \brief Return the Scale to store the elevation
     *  \return the Scale to store the elevation
     */
    float zScale() const;

    /**
     *  \brief set the content of a non-empty cell of the grid
     *  \param i : the column number
     *  \param j : the line number
     *  \param z : the altitude value
     *  \param nbPoints : the number of points seen corresponding to the cell
     *  \param sigma_z : the standard deviation
     *  \param confidence : the confidence over the previous data
     */
    void setCell(int i, int j, float z, int nbPoints, float sigma_z=0, float confidence=0);

    /**
     *  \brief set the content of a cell as empty
     *  \param i : the column number
     *  \param j : the line number
     */
    void setEmptyCell(int i, int j);

    /**
     *  \brief return true is the cell corresponding to theinput index is empty
     *  \param i : the column number
     *  \param j : the line number
     *  \return true is the cell corresponding to theinput index is empty
     *
     *  The user must have checked if the input coord are valid
     */
    bool isEmptyCell(int i, int j) const;


    /**************** DEM modifications ***************/

  private : 
    /**
     *  \brief add a new altitude value in a cell
     *  \param i : the abscica of the cell
     *  \param j : the ordinate of the cell
     *  \param z : the altitude value
     *  \param nbP : the number of points corresponding to the new value
     *  \return 0 if the process succeed, -1 if the input coordinate are out of the grid
     */
    int addValueInCell(int i,int j,float z,int nbP=1);

  public : 
    /**
     *  \brief add a new altitude value in a cell
     *  \param i : the abscica of the cell
     *  \param j : the ordinate of the cell
     *  \param z : the altitude value
     *  \param nbP : the number of points corresponding to the new value
     *  \return 0 if the process succeed, -1 if the input coordinate are out of the grid
     */
    int addValueInCoord(double x,double y,float z,int nbP=1);

    /**
     *  \brief merge the current DEM with the input one (2nd version)
     *  \param dem2 : the input DEM
     *  \result return 0 if the process succeed, -1 else
     */
    int mergeDEM(const DEM& dem2);
    

//     /***** Boolean obstacles building & evaluation ****/

//   protected : 
//     /**
//      *  \brief build boolean obstacles from the DEM
//      *  \param maxSlope : the maximal slope the robot is able to climb
//      *  \return 0 if the process succeed, -1 else
//      */
//     int buildBoolObs(float maxSlope);

//     /**
//      *  \brief increase the dimensions of the boolean obstacles
//      *  \param dist : the distance that the obstacles are inceased by
//      *  \return 0 if the process succeed, -1 else
//      */
//     int increaseBoolObs(float dist);

//   public : 
//     /**
//      *  \brief check if a position (x,y) on the DEM is free
//      *  \param x : absica
//      *  \param y : ordinate
//      */
//     bool isFree(float x,float y) const;

//     /**
//      *  \brief check if a position (x,y) on the DEM is obstructed
//      *  \param x : absica
//      *  \param y : ordinate
//      */
//     bool isObs(float x,float y) const;


    /******** Initialization functions (DEBUG) ********/

    /**
     *  \brief initialize the DEM with a flat terrain
     *  \param nbCols : number rof columns
     *  \param nbLines : number rof lines
     *  \param xScale : X size of each cell
     *  \param yScale : Y size of each cell
     *  \param zMin : minimal Z value used to convert the DEM into an image
     *  \param zMax : maximal Z value used to convert the DEM into an image
     *  \param zPlane : altitude of the plane
     */
    void initFlatDEM(int nbCols, int nbLines, float xScale, float yScale, float zMin, float zMax, float zPlane);

    /**
     *  \brief initialize the DEM with a random terrain
     *  \param nbCols : number rof columns
     *  \param nbLines : number rof lines
     *  \param xScale : X size of each cell
     *  \param yScale : Y size of each cell
     *  \param zMin : minimal Z value used to convert the DEM into an image
     *  \param zMax : maximal Z value used to convert the DEM into an image
     *  \param probSeen : the probability of each cell to have be already seen
     */
    void initRandDEM(int nbCols, int nbLines, float xScale, float yScale, float zMin, float zMax, float probSeen=1.0);

    /************ Input / Output functions ************/

  private : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const;

  public : 
    /**
     *  \brief save all information about the DEM into an output file
     *  \param filename : the output filename
     */
    int save(const char* filename) const;

    /**
     *  \brief load the DEM from an input file
     *  \param filename : the input filename
     */
    int load(const char* filename);

    /**************** Utility functions ***************/

  protected : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param iMin : the number of the first column to keep
     *  \param iMax : the number of the last column to keep
     *  \param jMin : the number of the first line to keep
     *  \param jMax : the number of the last line to keep
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of DEM is created
     */
    virtual DEM* subMap(int iMin, int iMax, int jMin, int jMax) const;

  public : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param xMin : minimal bound over the X axis
     *  \param xMax : maximal bound over the X axis
     *  \param yMin : minimal bound over the Y axis
     *  \param yMax : maximal bound over the Y axis
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of DEM is created
     */
    DEM* subMap(float xMin, float xMax, float yMin, float yMax) const;

    /**
     *  \brief extract a sub-map of the current map
     *  \param (xCenter,yCenter) : coordinates of the center of the sub-map
     *  \param size : size of the sub-map
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of DEM is created
     */
    DEM* subMap(float xCenter, float yCenter, float size) const;

    /**
     *  \brief display the main features of the DEM
     *  \param dispGrid : display the grid if true
     */
    void display(bool dispGrid=false) const;


//     /********************* Friends ********************/

//     friend mpc::DEM_reader;
  };


  

};

#endif // DEM_HPP
