#ifndef DISTOBSMAP_HPP
#define DISTOBSMAP_HPP

#include "environment/inc/Map2D.hpp"
#include "environment/inc/BoolObsMap.hpp"
#include "common/inc/Position.hpp"

#include <cmath>
#include <list>

namespace mpc
{

  typedef struct DIST_CELL
  {
    int x;
    int y;
  } pos_cell_xy;

  /**
   * \enum dm_cell_t
   * \brief Enumeration of the different distance map cell types
   *
   * DM_CELL_ND   : not defined cell
   * DM_CELL_CUR  : currently processed cell
   * DM_CELL_DONE : already processed cell
   * DM_CELL_FAR  : currently processed cell, but too far to be computed
   */
  typedef enum {
    DM_CELL_ND   =  -1,
    DM_CELL_CUR  =   0,
    DM_CELL_DONE =   1,
    DM_CELL_FAR  =   2
  } dm_cell_t;

  typedef struct FREE_SEGMENT
  {
    PositionXY p1;
    PositionXY p2;
    float length;
    float minB;
    float maxB;
    PositionXY maxDistPos;
    float dist;
    PositionXY nearObs;
  } FreeSegment;
  
                                    /**************************************************/
                                    /****************** Class dm3_cell ****************/
                                    /**************************************************/

  /* \class dm3_cell
   * \brief cell containing all information for the distmap computation
   */
  class DM3_Cell
  {
  public :
    int posX;                                          /**< X pos of the cell in the grid */
    int posY;                                          /**< Y pos of the cell in the grid */
    int distX;                                         /**< distance to the closest obstacle (|dx|) */
    int distY;                                         /**< distance to the closest obstacle (|dy|) */
    int dist;                                          /**< distance to the closest obstacle (dx^2 + dy^2) */
    int nearX;                                         /**< X pos of the nearest obstacle */
    int nearY;                                         /**< Y pos of the nearest obstacle*/
    obs2D_cell_t cellStatus;                           /**< status of the current cell (CELL_UNKWN, CELL_OBS or CELL_FREE) */
    std::list<DM3_Cell*>::iterator it_tcd;             /**< pointer (iterator) to the position of the cell in the sorted lists */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    DM3_Cell();

  public : 
    /**
     *  \brief copy constructor
     */
    DM3_Cell(const DM3_Cell& cell);

    /**
     *  \brief init constructor
     */
    DM3_Cell(int x, int y, obs2D_cell_t status, int distX=-1, int distY=-1, int nearX=-1, int nearY=-1, std::list<DM3_Cell*>::iterator *it=NULL);

    /**
     *  \brief destructor
     */
    ~DM3_Cell();
  };


                                    /**************************************************/
                                    /***************** Class DistObsMap ***************/
                                    /**************************************************/

  /** \class DistObsMap
   * \brief map containing the distance to the closest obstacle for every cell of a grid
   */
  class DistObsMap : public BoolObsMap
  {
  protected : 
//     float** distObs_;             /**< Table containing the distance the closest obstacle of the map */
    pos_cell_xy** dist_;           /**< table containing the distance (in number of cells) from the closest obstacle */
    pos_cell_xy** near_;           /**< table containing the position in the grid of the closest obstacle */    

  private : 
    pos_cell_xy* tabBorder_;       /**< table containing the border points of the obstacles */
    int nbBorder_;                 /**< size of the border table */
    pos_cell_xy* newTabBorder_;    /**< table containing the border points of the obstacles */
    int newNbBorder_;              /**< size of the border table */
    dm_cell_t** status_;           /**< table containing the status of the distance map cells during its computation */

  public : 
    static bool unkwnIsObs_;              /* If true, unknown regions of the map are considered as obstacles. Distance to obstacles is computed to them */


    //////////////////////////// THIRD VERSION OF THE DISTMAP COMPUTATION

  protected : 
    int nbFilledCells_;                   /**< Number of filled cells */
    int maxSqDistToObs_;                  /**< maximale square distance to obstacles (and max size of tabCptDist_) */
    DM3_Cell*** tabCells_;                /**< Table of cells containing all information for the distmap computation */
    std::list<DM3_Cell*>** tabCptDist_;   /**< Table of list of cells sorted by distance to the nearest obstacle */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    DistObsMap();
    
  public : 
    /**
     *  \brief copy constructor
     *  \param map : the map to copy
     */
    DistObsMap(const DistObsMap& map);

    /**
     *  \brief init constructor
     *  \param nbCols : number of columns of the map table
     *  \param nbLines : number of lines of the map table
     *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param xScale : X size of a cell in meters
     *  \param yScale : Z size of a cell in meters
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    DistObsMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float thresSlope, float thresVarSlope, float safeDist);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    DistObsMap(const SlopeMap* slopemap, float thresSlope=0.25*M_PI, float safeDist=0.5);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param thresVarSlope : the maximal slope derivative that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     */
    DistObsMap(const VarSlopeMap* slopemap, float thresSlope=0.35*M_PI, float thresVarSlope=2.0*M_PI, float safeDist=0.5);

    /**
     *  \brief init constructor (from an obstacle map)
     *  \param obsmap : the obstacle map
     */
    DistObsMap(const BoolObsMap* obsmap);

    /**
     *  \brief destructor
     */
    virtual ~DistObsMap();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of DistObsMap is allocated
     */
    virtual DistObsMap* copy() const;

    /***************** Initialization *****************/

  private : 
    /*
     *  \brief init the distance table from its dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTableDist(int nbCols,int nbLines);

    /*
     *  \brief init tables from their dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTables(int nbCols,int nbLines);
    
    /*
     *  \brief delete all the content of the slope map tables
     */
    void clearTables();

    /*
     *  \brief return true if the cell at index (i,j) is a border of an obstacle
     *  \return true if the cell at index (i,j) is a border of an obstacle
     */
    bool isBorderObs(int i,int j) const;

    /*
     *  \brief init the table used for the distance computation from its dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTableCptDist(int nbCols,int nbLines);


    //////////////////////////// THIRD VERSION OF THE DISTMAP COMPUTATION

    /*
     *  \brief init the table used for the distance computation from its dimesions (for the 3rd version of the distmap computation)
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    void initTableCptDist3(int nbCols,int nbLines);


    /******************** Accessors *******************/

  public : 
    /*
     *  \brief set the cell (i,j) as undefined (no knowledge about the distance to obstacles)
     *  \param (i,j) : index of the cell in the grid
     */
    void setUndefCell(int i,int j);

    /*
     *  \brief set the parameters of cell (i,j)
     *  \param (i,j) : index of the cell in the grid     
     *  \param dist : the new distance the the closest obstacle
     *  \param near : the new index of the closest obstacle
     */
    void setDist(int i, int j, pos_cell_xy dist, pos_cell_xy near);

    /*
     *  \brief set the parameters of cell (i,j)
     *  \param (i,j) : index of the cell in the grid     
     *  \param state : status of the cell (CELL_UNKWN, CELL_OBS, CELL_FREE)
     *  \param dist : the new distance the the closest obstacle
     *  \param near : the new index of the closest obstacle
     */
    void setDistAndState(int i, int j, obs2D_cell_t state, pos_cell_xy dist, pos_cell_xy near);

  public : 
    /**
     *  \brief get the type of cell (i,j)
     *  \param (i,j) : index of the cell in the grid
     *  \param cellSt : the resulting cell type
     *  \param distX, distY : the resulting distance
     *  \param nearX, nearY : the nearest obstacle from the input position
     */
    void getCellInfo(int i, int j, obs2D_cell_t& cellSt, int& distX, int& distY, int& nearX, int& nearY) const;

    /**
     *  \brief get the type of cell (i,j)
     *  \param (i,j) : index of the cell in the grid
     *  \param cellSt : the resulting cell type
     *  \param dist : the resulting distance
     *  \param near : the nearest obstacle from the input position
     */
    void getCellInfo(int i, int j, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const;

    /**
     *  \brief get the type of cell (i,j)
     *  \param (x,y) : input position on the map
     *  \param cellSt : the resulting cell type
     *  \param dist : the resulting distance
     *  \param near : the nearest obstacle from the input position
     */
    void getInfo(float x, float y, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const;

    /**
     *  \brief search for the cell maximizing its minimal distance to obstacles on a line (x1,y1) -> (x2,y2)
     *  \param (x1,y1) : coord of the first end point of the line
     *  \param (x2,y2) : coord of the second end point of the line
     *  \param pos : resulting cell position
     *  \param cellSt : resulting cell state
     *  \param dist : resulting cell minimal distance to obstacles
     *  \param near : resulting cell nearest obstacle position
     *  \return true if a free position has been found, false else
     */
    bool getCellMaxDistOnLine(float x1, float y1, float x2, float y2, PositionXY& pos, obs2D_cell_t& cellSt, float& dist, PositionXY& near) const;

    /**
     *  \brief search for the cell maximizing its minimal distance to obstacles on a line (x1,y1) -> (x2,y2)
     *  \param (x1,y1) : coord of the first end point of the line
     *  \param (x2,y2) : coord of the second end point of the line
     *  \param vecRes : vector containing the free segments found
     *  \return the number of free segments found
     */
    int getFreeSegmentsOnLine(float x1, float y1, float x2, float y2, vector<FreeSegment>& vecRes) const;

    /**
     *  \brief return true if the segment defined between (x1,y1) and (x2,y2) is free
     *  \param (x1,y1) : coord of the first end point of the segment
     *  \param (x2,y2) : coord of the second end point of the sement
     *  \return true if the segment defined between (x1,y1) and (x2,y2) is free
     */
    bool isFreeSegmentsOnLine(float x1, float y1, float x2, float y2) const;

    //////////////////////////// THIRD VERSION OF THE DISTMAP COMPUTATION

  private : 
//     /*
//      *  \brief create a new cell in the grid, and return a pointer to it
//      *  \param k : index of the column in the grid of the new cell
//      *  \param l : index of the line in the grid of the new cell
//      */
//     DM3_Cell* init_cell(int k, in l);

    /*
     *  \brief remove a cell from the sorted list
     *  \param cell : a pointer to the cell to remove
     */
    void removeCellFromSortedList(DM3_Cell* cell);

    /*
     *  \brief add a cell into the sorted list
     *  \param cell : the new cel to add into the list
     *  \param dist : the distance from the cell to add to its current closest neighbor
     *  \return an iterator to to the cell in the sorted list
     */
    std::list<DM3_Cell*>::iterator addCellInSortedList(DM3_Cell* cell, int dist);

    /*
     *  \brief update the distance of a cell in the grid from one of its neighbors
     *  \param (k,l) : coordinates in the grid of the cell to update
     *  \param (i,j) : coordinates in the grid of the neighbor from which the cell (k,l) must be updated
     */
    void updateDist(int k, int l, int i, int j, int curDist=-1);

    //////////////////////////////////////////////////////////////////////


    /************** Obstacle map building *************/

  private : 
//     /*
//      *  \brief compute the distance from the center of a cell (i,j) to the closest obstacle
//      *  \param (i,j) : index of the cell in the grid     
//      *  \return the distance to the closest obstacle from the center of the cell
//      */
//     float computeDistCellObs(int i, int j) const;

    /*
     *  \brief compute the distance to the closest obstacle for every cell
     */
    void computeDistObs();

    /*
     *  \brief compute the distance to the closest obstacle for every cell (2nd version)
     */
    void computeDistObs2();

    /*
     *  \brief compute the distance to the closest obstacle for every cell (3rd version)
     */
    void computeDistObs3();

    /*
     *  \brief fill the dist and near table from the cell table of the 3rd  version of the distmap computation
     */
    void fill_tabDistNear_from_tabCell();

    /*
     *  \brief set every unknown cell as defining an obstacle
     */
    void setUnknwnAsObs();

  public : 
    /**
     *  \brief build the obstacle map from an input slope map
     *  \param sm : the input slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromSM(const SlopeMap* sm, float thresSlope=0.25*M_PI, float safeDist=0.5);

    /**
     *  \brief build the obstacle map from an input slope map
     *  \param sm : the input slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromVSM(const VarSlopeMap* vsm, float thresSlope=0.25*M_PI, float thresVarSlope=2.0*M_PI, float safeDist=0.5);

    /**
     *  \brief build the obstacle map from an input slope map
     *  \param sm : the input slope map
     *  \param thresSlope : the maximal slope that the robot can climb
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromVSM3(const VarSlopeMap* vsm, float thresSlope=0.25*M_PI, float thresVarSlope=2.0*M_PI, float safeDist=0.5);

    /**
     *  \brief build the distance map from an obstacle map
     *  \param obsmap : the input obstacle map
     *  \param safeDist : the minimal distance that the robot may keep to be safe for collisions
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromObsMap(const BoolObsMap* obsmap);


    /************ Input / Output functions ************/

  protected : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    virtual void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    virtual int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    virtual void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const;


    /**************** Utility functions ***************/

  private : 
    /*
     *  \brief display the status table
     */
    void displayTabStatus() const;

    /*
     *  \brief display the status table
     */
    void displayTabBoolObs() const;

    /*
     *  \brief display the cell table sorted by distance
     */
    void displayTabCptDist() const;

    /*
     *  \brief display the table of cells (3rd distmap computation version)
     */
    void displayTabCells() const;

  public : 
    /**
     *  \brief display the main features of the dist obs map
     *  \param dispGrid : display the grid if true
     */
    virtual void display(bool dispGrid) const;
  };


  

};

#endif // DISTOBSMAP_HPP
