#ifndef ENVIRONMENT_HPP
#define ENVIRONMENT_HPP

#include "common/inc/Config.hpp"
// #include "Obstacle.hpp"
#include "environment/inc/Obs.hpp"
#include "environment/inc/Map.hpp"
//using namespace mpc;

#include <boost/thread/thread.hpp>
#include <list>

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class Environment ****************/
                                    /**************************************************/

  /** \class Environment
   * \brief Representation of the environment of a robotic system
   */
  class Environment
  {
  protected : 
    Map* map_;
    //   vector<ObsDisk2D*> *lObs_;           // should be deleted
    //   vector<Obstacle2D*> *lStObs_;        // should be deleted
    vector<Obs2D*> *lObs2D_;                            /**< list of the obstacles obtained by a perception process (given in a local frame) */
    vector<Obs2D*> *lFixedObs2D_;                       /**< list of the a priori known obstacles (given in a global fixed frame (0,0,0)) */

    //   double Xdflt_;
    //   double Ydflt_;

    double minX_;                                       /**< minimal X value of the static environment */
    double minY_;                                       /**< minimal Y value of the static environment */
    double maxX_;                                       /**< maximal X value of the static environment */
    double maxY_;                                       /**< maximal Y value of the static environment */

//     PositionXYThT frame_;                               /**< the frame of the environment built by the perception process */
    PositionXYTh frame_;                                /**< the frame of the environment built by the perception process */

    //   // security components
    //   boost::mutex mutexEnv;                              // mutex used to prevent environment access conflicts

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    Environment();

    /**
     *  \brief copy constructor
     *  \param e : the environment to copy
     *  \param frame : the new frame in which the environment will be set
     */
    Environment(const Environment& e,const PositionXYTh& frame);
//     Environment(const Environment& e,const PositionXYThT& frame);

    /**
     *  \brief copy constructor with frame change
     *  \param e : the environment to copy
     *  \param 
     */
    Environment(const Environment& e);

    //   Environment(Map* m);                                // init constructor
    //   Environment(Map* m, vector<ObsDisk2D*> *l, vector<Obstacle2D*> *ls); // init constructor

    /**
     *  \brief destructor
     */
    virtual ~Environment();

    /******************* Accessors ********************/
    //   int getNbObs() const;                               // return the number of dynamic obstacles
    //   int getNbStObs() const;                             // return the number of static obstacles
    //   void appendObs(ObsDisk2D* obs);                     // append a new dynamic obstacle
    //   void appendStObs(Obstacle2D* obs);                  // append a new static obstacle
    //   void removeObs(int i);                              // remove the i-th dynamic obstacle
    //   void removeStObs(int i);                            // remove the i-th static obstacle
    //   void clearObs();                                    // remove all dynamic obstacles of the environment
    //   void clearStObs();                                  // remove all static obstacles of the environment
    //   ObsDisk2D* getObs(int i) const;                     // return the i-th dynamic obstacle
    //   Obstacle2D* getStObs(int i) const;                  // return the i-th static obstacle
    //   Map* getMap() const;                                // return the current static map
    //   void setMap(Map* map);                              // set the current map
    //  const PMConfig getCenter() const;                  // return the center of the environment
    double getDim() const;                              // return the dimensions of the environment
  
    double getDfltX() const;                            // return the default X value
    double getDfltY() const;                            // return the default Y value
    void setDfltX(double x);                            // set the default X value
    void setDfltY(double y);                            // set the default Y value

    int loadMap(char* nomfic);                          // load a new map

    /**
     *  \brief return the current static map
     *  \return the current static map
     */
    inline Map* map() const { return this->map_; }

  public : 
    /**
     *  \brief return the minimal X value of the static environment
     *  \return the minimal X value of the static environment
     */
    inline double minX() const { return this->minX_; }

    /**
     *  \brief return the maximal X value of the static environment
     *  \return the maximal X value of the static environment
     */
    inline double maxX() const { return this->maxX_; }

    /**
     *  \brief return the minimal Y value of the static environment
     *  \return the minimal Y value of the static environment
     */
    inline double minY() const { return this->minY_; }

    /**
     *  \brief return the maximal Y value of the static environment
     *  \return the maximal Y value of the static environment
     */
    inline double maxY() const { return this->maxY_; }

    /**
     *  \brief return the center of the environment
     *  \return the center of the environment
     */
    PositionXY getCenter() const;

    /**
     *  \brief return the center of the environment (in the local frame)
     *  \return the center of the environment (in the local frame)
     */
    PositionXY getLocalCenter() const;

    /**
     *  \brief return the radius of the environment
     *  \return the radius of the environment
     */
    double getRadius() const;

    /**
     * \brief return the number of 2D obstacles
     * \return the number of 2D obstacles
     */
    int getNbObs2D() const;
  
    /**
     *  \brief append a new 2D obstacle
     *  \param obs : the new obstacle
     */
    void appendObs2D(const Obs2D& obs);

    /**
     *  \brief remove all 2D obstacles of the environment
     */
    void clearObs2D();

    //   /**
    //    *  \brief remove the i-th 2D obstacle
    //    *  \param i : the index of the obstacle to remove
    //    */
    //   void removeObs2D(int i);

    /**
     *  \brief return the i-th 2D obstacle in the global frame
     *  \param i : index of the obstacle to retrieve
     *  \return a pointer to a copy of the  i-th 2D obstacle in the global frame
     */
    Obs2D* getObs2D(int i) const;

    /**
     *  \brief return the i-th 2D obstacle in the local frame
     *  \param i : index of the obstacle to retrieve
     *  \return the i-th 2D obstacle in the local frame
     */
    const Obs2D& getLocalObs2D(int i) const;

    /**
     * \brief return the number of 2D obstacles
     * \return the number of 2D obstacles
     */
    int getNbFixedObs2D() const;
  
    /**
     *  \brief append a new 2D obstacle
     *  \param obs : the new obstacle
     */
    void appendFixedObs2D(const Obs2D& obs);

    /**
     *  \brief remove all 2D obstacles of the environment
     */
    void clearFixedObs2D();

    /**
     *  \brief return the i-th 2D obstacle
     *  \param i : index of the obstacle to retrieve
     *  \return the i-th 2D obstacle
     */
    const Obs2D& getFixedObs2D(int i) const;

    /**
     *  \brief return the i-th 2D obstacle in the local frame
     *  \param i : index of the obstacle to retrieve
     *  \return the i-th 2D obstacle in the local frame
     */
    const Obs2D& getLocalFixedObs2D(int i) const;

    /**
     *  \brief return the frame in which is set the environment
     *  \return the frame in which is set the environment
     */
    inline const PositionXYTh& frame() const { return this->frame_; }
//     inline const PositionXYThT& frame() const { return this->frame_; }
  
    /*************** Environment update ***************/
  
    /**
     *  \brief simulate a random motion of all dynamic obstacles
     *  \param the time at which the environment is evaluated
     */
    void simuRandMotion(double t);

    /************* Obstruction functions **************/

  public : 
    //   virtual bool obstruct(const Config& c);             // return true if the obstacles obstruct the config c
    //   virtual bool obstruct(const Config& c1,const Config& c2); // return true if the obstacles obstruct the line between c1 and c2

    //   /**
    //    *  \brief test if the environment obstruct the position pos
    //    *  \param pos : the position to test
    //    */
    //   virtual bool obstruct(const PositionXY pos,double t);

    //   /**
    //    *  \brief test if the environment obstruct the position pos
    //    *  \param pos : the position to test
    //    *  \param t : the input time at which to test collisions
    //    *  \param near : the nearest position of the obstacle found
    //    *  \param dist : the minimal distance between pos and the nearest obstacle
    //    */
    //   virtual bool obstruct(const PositionXY pos,double t,PositionXY& near,double& dist);

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& pos,double t) const;

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position-time of the segment
     *  \param p2 : the second end position-time of the segment
     */
    virtual bool obstructSgmt(const PositionXYT& p1,const PositionXYT& p2) const;

    /**
     *  \brief test if the static environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstructSgmt(const PositionXYT& p1,const PositionXYT& p2,PositionXYT& near,double& dist) const;

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool staticObstruct(const PositionXY pos,PositionXY& near,double& dist);

    /**
     *  \brief test if the static environment obstruct the position pos
     *  \param pos : the position to test
     */
    virtual bool staticObstruct(const PositionXY& pos) const;

    /**
     *  \brief test if the static environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     */
    virtual bool staticObstructSgmt(const PositionXY& p1,const PositionXY& p2) const;

    /**
     *  \brief test if the static environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool staticObstructSgmt(const PositionXY& p1,const PositionXY& p2,PositionXY& near,double& dist) const;

    /*************** Sampling functions ***************/

    /**
     *  \brief sample the obstructed positions of the obstacle of the static environment
     *  \param sampleX,sampleY : sampling scale of the resulting map
     *  \return an occupancy grid of the environment
     *
     *  A new instance of occupancy grid is returned
     */
    virtual Map* sample(int sampleX,int sampleY) const;

    /****************** IO functions ******************/

    /**
     *  \brief load a segmented environment from an input file
     *  \param filename : the input filename of the environment to load
     *  \param dist : the secure distance considered to build boxes around the obstacles
     *  \return 0 if the process succeed, -1 else
     */
    int loadSegmentedEnv(const char* filename,double dist=0);

    /*************** Display functions ****************/

//     /**
//      *  \brief draw the environment in an openGL viewer
//      *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
//      *  \param time : the time at which the environment must be displayed
//      */
//     virtual void toGL(bool dispFM=false, double time=0) const;

    /**
     *  \brief draw the environment in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
     */
    virtual void toGL(const PositionXYTh& frame,bool dispFM) const;
//     virtual void toGL(const PositionXYThT& frame,bool dispFM) const;
  
    /**
     *  \brief draw the environment in an openGL viewer
     *  \param dispFM : true if the forecast model of the dynamic obstacles must be displayed
     *  \param curTime : the current time at which must be displayed the static environment
     */
    virtual void toGL(bool dispFM=false,double curTime=0) const;

    /**
     *  \brief display the main features of the environment
     */
    virtual void display() const;

    /**
     *  \brief display the number of obstacles of the environment
     */
    virtual void displayNbObs() const;


#ifdef GAZEBO_SCENE
    /*************** Sampling functions ***************/
    int loadFromXMLFile(char* filename);                // load the environment from an xml file (gazebo world)
#endif
  };

};
#endif // ENVIRONMENT_HPP

