#ifndef MPC_OBS_HPP
#define MPC_OBS_HPP

#include "common/inc/Object.hpp"
#include "common/inc/PositionTime.hpp"
#include "common/inc/Speed.hpp"
#include "environment/inc/Map.hpp"
//using namespace mpc;

/**
 * \enum bshape_t
 * \brief Enumeration of the different bounding boxes
 *
 * UNDEFINED_OBS : undefined obstacle
 * POINT_OBS2D   : ponctual obstacle
 * SPHERE_OBS2D  : spherical obstacle
 * POLYGON_OBS2D : polygonal obstacle
 * BOX_OBS2D     : rectangular obstacle
 */
typedef enum {
  UNDEFINED_OBS     =  0,
  POINT_OBS2D       =  1,
  DISK_OBS2D        =  2,
  POLYGON_OBS2D     =  3,
  BOX_OBS2D         =  4
} obst_t;

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Obs ********************/
                                    /**************************************************/

  /** \class Obs
   * \brief Generic obstacle perception data class
   */
  class Obs
  {
  protected : 
    //   Position* lastPosT_;                            /**< Last position-time at which the obstacle has been perceived */
    bool moving_;                                    /**< true if the obstacle seems to be moving */
    obst_t type_;                                    /**< type of the obstacle */

    static double thresChange_;                      /**< probability threshold to change its motion after one sec (random motion mode) */
    static double maxAngleChange_;                   /**< maximal angle change after one sec (random motion mode) */
    static double maxSpdChange_;                     /**< maximal speed change after one sec (random motion mode) */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    Obs();

  public : 
    /**
     *  \brief copy constructor
     *  \param o : the obstacle to copy
     */
    Obs(const Obs& o);

    /**
     *  \brief destructor
     */
    virtual ~Obs();

    /******************** Accessors *******************/

    /**
     *  \brief return true if the obstacle is moving
     *  \return true if the obstacle is moving
     */
    bool isMoving() const { return moving_; }
  };

};
#endif // MPC_OBS_HPP


#ifndef MPC_OBS2D_HPP
#define MPC_OBS2D_HPP

#include "common/inc/BoundingShape.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class Obs2D *******************/
                                    /**************************************************/

  /** \class Obs2D
   * \brief Generic 2D obstacle perception data class
   */
  class Obs2D : public Obs, public Object2D
  {
  protected : 
    PositionXYThT lastPosT_;                          /**< Last position-time at which the obstacle has been perceived */
    SpeedXYTh lastSpd_;                               /**< Last estimated speed of the obstacle */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    Obs2D();

    /**
     *  \brief init constructor
     *  \param post : the initial position-time of the obstacle
     *  \param spd : the initial speed of the obstacle
     */
    Obs2D(const PositionXYThT& post,const SpeedXYTh& spd, const BoundingShape& bs);
  
  public : 
    /**
     *  \brief copy constructor
     *  \param o : the obstacle to copy
     */
    Obs2D(const Obs2D& o);

    /**
     *  \brief destructor
     */
    virtual ~Obs2D();

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual Obs2D* copy() const = 0;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual Obs2D* copyFromFrame(const PositionXYTh& frame) const = 0;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual Obs2D* copyFromFrame(const PositionXYThT& frame) const = 0;

    /**
     *  \brief return a pointer on a copy of the current obstacle from frame (0,0,0,0) to the input frame
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual Obs2D* copyInFrame(const PositionXYTh& frame) const = 0;

    /**
     *  \brief return a pointer on a copy of the current obstacle from frame (0,0,0,0) to the input frame
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual Obs2D* copyInFrame(const PositionXYThT& frame) const = 0;

    /******************** Accessors *******************/
  
  public : 
    /**
     *  \brief Position accessor
     *  \return the position of the object in its workspace (W)
     */
    virtual const PositionXYTh& pos() const { return (this->lastPosT_); }

    /**
     *  \brief Position accessor
     *  \param t : time at which the position of the object is evaluated
     *  \return the position of the object in its workspace (W) at time t
     */
    virtual const PositionXYTh& pos(double t) const { return this->forecastPosT(t); }

  protected : 
    /**
     *  \brief Position accessor
     *  \return the position of the object in its workspace (W)
     */
    virtual PositionXYTh& pos() { return (this->lastPosT_); }

  public : 
    /**
     *  \brief return the last position-time at which the obstacle has been perceived
     */
    inline const PositionXYThT& lastPosT() const { return (this->lastPosT_); }

    /**
     *  \brief return the last position-time at which the obstacle has been perceived
     */
    inline const SpeedXYTh& lastSpd() const { return (this->lastSpd_); }

    /**
     *  \brief return the forecast position of the obstacle at time t
     *  \param t : the input time at which the position of the obstacle is evaluated
     *  \return the position forecast at time t
     */
    virtual PositionXYThT forecastPosT(double t) const;

    /**
     *  \brief register a new position-time of the obstacle
     *  \param post : the new position-time perceived
     */
    virtual void registerPosT(const PositionXYThT& post,const SpeedXYTh& spd);

  public : // !!!!!!!!!!!!!!!!!!!!!!!!!! a remettre en protected !!!!!!!!!!!!!!!!!!!!!!!!!!!
    /**
     *  \brief return the last position-time at which the obstacle has been perceived
     */
    inline PositionXYThT& lastPosT() { return (this->lastPosT_); }

    /**
     *  \brief return the last position-time at which the obstacle has been perceived
     */
    inline SpeedXYTh& lastSpd() { return (this->lastSpd_); }

    /*************** Environment update ***************/
  
    /**
     *  \brief simulate a random motion of all dynamic obstacles
     *  \param the time at which the environment is evaluated
     */
    void simuRandMotion(double t);

    /************* Obstruction functions **************/

  public : 
    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& pos,double t=0) const = 0;

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const = 0;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t=0) const = 0;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t,PositionXY& near,double& dist) const = 0;

    /*************** Sampling functions ***************/

    /**
     *  \brief sample the obstructed positions of the obstacle of the static environment
     *  \param map : the occupancy grid on which the obstacle must be sampled
     *  \param sampleX,sampleY : sampling scale of the resulting map
     *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
     */
    virtual void sample(Map* map,double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const;

    /*************** Display functions ****************/

    /**
     *  \brief draw the object in an openGL viewer
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(bool dispFM=false) const = 0;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYTh& frame,bool dispFM=false) const = 0;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYThT& frame,bool dispFM=false) const = 0;
  };

};
#endif // MPC_OBS2D_HPP


#ifndef MPC_DISKOBS2D_HPP
#define MPC_DISKOBS2D_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class DiskObs2D *****************/
                                    /**************************************************/

  /** \class DiskObs2D
   * \brief 2D disk-shaped obstacle perception data class
   */
  class DiskObs2D : public Obs2D
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    DiskObs2D();
  
  public : 
    /**
     *  \brief init constructor
     *  \param post : the initial position-time of the obstacle
     *  \param spd : the initial speed of the obstacle
     *  \param rad : the radius of the disk obstacle
     */
    DiskObs2D(const PositionXYT& post,const SpeedXY& spd,double rad);
  
    /**
     *  \brief copy constructor
     *  \param o : the obstacle to copy
     */
    DiskObs2D(const DiskObs2D& o);

    /**
     *  \brief destructor
     */
    virtual ~DiskObs2D();

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual DiskObs2D* copy() const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual DiskObs2D* copyFromFrame(const PositionXYTh& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual DiskObs2D* copyFromFrame(const PositionXYThT& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual DiskObs2D* copyInFrame(const PositionXYTh& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual DiskObs2D* copyInFrame(const PositionXYThT& frame) const;

    /******************** Accessors *******************/

    /**
     *  \brief Bounding shape accessor
     *  \return the bounding shape of the object
     */
    inline BDisk& bshape() const { return *((BDisk*)bshape_); }

    /**
     *  \brief return the radius of the disk shape of the obstacle
     *  \return the radius of the disk shape of the obstacle
     */
    inline double radius() const { return this->bshape().radius(); }

    /************* Obstruction functions **************/

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& pos,double t=0) const;

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t=0) const;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t,PositionXY& near,double& dist) const;

    //   /*************** Sampling functions ***************/

    //   /**
    //    *  \brief sample the obstructed positions of the obstacle of the static environment
    //    *  \param sampleX,sampleY : sampling scale of the resulting map
    //    *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
    //    *  \return an occupancy grid of the environment
    //    *
    //    *  A new instance of occupancy grid is returned
    //    */
    //   virtual Map* sample(double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the object's features
     */
    virtual void display() const;

    /**
     *  \brief draw the object in an openGL viewer
     */
    inline virtual void toGL() const { this->toGL(false); }

    /**
     *  \brief draw the object in an openGL viewer
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(bool dispFM) const;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYTh& frame,bool dispFM) const;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYThT& frame,bool dispFM=false) const;
  };

};
#endif // MPC_DISKOBS2D_HPP


#ifndef MPC_BBOXOBS2D_HPP
#define MPC_BBOXOBS2D_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class BBoxObs2D *****************/
                                    /**************************************************/

  /** \class BBoxObs2D
   * \brief 2D bounding box-shaped obstacle perception data class
   */
  class BBoxObs2D : public Obs2D
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    BBoxObs2D();
  
  public : 
    /**
     *  \brief init constructor
     *  \param post : the initial position-time of the obstacle
     *  \param spd : the initial speed of the obstacle
     *  \param lg : the length of the box
     *  \param wd : the width of the box
     */
    BBoxObs2D(const PositionXYThT& post,const SpeedXYTh& spd,double lg,double wd);

    /**
     *  \brief init constructor (from a segment and a secure distance)
     *  \param p1 : the first end of the segment
     *  \param p2 : the second end of the segment
     *  \param dist : the secure distance from the segment to the boundaries of the box
     *  \param t : the initial time at which is defined the obstacle
     *  \param spd : the initial speed of the obstacle
     */
    BBoxObs2D(const PositionXY& p1,const PositionXY& p2,double dist,double t=0,const SpeedXYTh& spd=SpeedXYTh(0,0,0));
  
    /**
     *  \brief copy constructor
     *  \param o : the obstacle to copy
     */
    BBoxObs2D(const BBoxObs2D& o);

    /**
     *  \brief destructor
     */
    virtual ~BBoxObs2D();

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual BBoxObs2D* copy() const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual BBoxObs2D* copyFromFrame(const PositionXYTh& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual BBoxObs2D* copyFromFrame(const PositionXYThT& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual BBoxObs2D* copyInFrame(const PositionXYTh& frame) const;

    /**
     *  \brief return a pointer on a copy of the current obstacle
     *  \param frame : the frame in which we want to express the coordinates of the obstacle
     *  \return a pointer on a copy of the current obstacle
     */
    virtual BBoxObs2D* copyInFrame(const PositionXYThT& frame) const;

    /******************** Accessors *******************/

    /**
     *  \brief Bounding shape accessor
     *  \return the bounding shape of the object
     */
    inline BBox& bshape() const { return *((BBox*)bshape_); }

    /**
     *  \brief getter : return the length of the box
     *  \return the length of the box
     */
    inline double lg() const { return this->bshape().lg(); }
  
    /**
     *  \brief getter : return the width of the box
     *  \return the width of the box
     */
    inline double wd() const { return this->bshape().wd(); }

    /************* Obstruction functions **************/

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& pos,double t=0) const;

    /**
     *  \brief test if the environment obstruct the position pos
     *  \param pos : the position to test
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& pos,double t,PositionXY& near,double& dist) const;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t=0) const;

    /**
     *  \brief test if the environment obstruct the segment p1p2
     *  \param p1 : the first end position of the segment
     *  \param p2 : the second end position of the segment
     *  \param t : the input time at which to test collisions
     *  \param near : the nearest position of the obstacle found
     *  \param dist : the minimal distance between pos and the nearest obstacle
     */
    virtual bool obstruct(const PositionXY& p1,const PositionXY& p2,double t,PositionXY& near,double& dist) const;

    //   /*************** Sampling functions ***************/

    //   /**
    //    *  \brief sample the obstructed positions of the obstacle of the static environment
    //    *  \param sampleX,sampleY : sampling scale of the resulting map
    //    *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
    //    *  \return an occupancy grid of the environment
    //    *
    //    *  A new instance of occupancy grid is returned
    //    */
    //   virtual Map* sample(double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const;

    /*************** Sampling functions ***************/

    /**
     *  \brief sample the obstructed positions of the obstacle of the static environment
     *  \param map : the occupancy grid on which the obstacle must be sampled
     *  \param sampleX,sampleY : sampling scale of the resulting map
     *  \param minY,maxX,minY,maxY : the bounds of the window in which the obstacle must be sampled
     */
    virtual void sample(Map* map,double minX,double maxX,double minY,double maxY,int sampleX,int sampleY) const;

    /*************** Display functions ****************/
  
    /**
     *  \brief display the object's features
     */
    virtual void display() const;

    /**
     *  \brief draw the object in an openGL viewer
     */
    inline virtual void toGL() const { this->toGL(false); }

    /**
     *  \brief draw the object in an openGL viewer
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(bool dispFM) const;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYTh& frame,bool dispFM=false) const;

    /**
     *  \brief draw the object in an openGL viewer
     *  \param frame : the frame of the perception system from which is construct the environment model
     *  \param dispFM : true if the forecast model must be displayed
     */
    virtual void toGL(const PositionXYThT& frame,bool dispFM) const;
  };

};
#endif // MPC_BBOXOBS2D_HPP
