#ifndef ROADMAP_HPP
#define ROADMAP_HPP

//#include "Config.hpp"
#include "common/inc/Position.hpp"
#include "environment/inc/Environment.hpp"

#include <vector>

namespace mpc
{

  class RoadNode
  {
  protected : 
    PositionXY* pos_;                                   // position of the node
    std::vector<RoadNode*>* lNeighb_;                  // list of neighbours of the current node
    bool visited_;                                     // true if the node has been already visited
    double E_;                                         // estimation from the current node to the goal node
    double L_;                                         // length of the path computed up to this node
    RoadNode* prev_;                                   // previous node from which has been computed the current node

    /********** Constructors and destructors **********/
  protected : 
    RoadNode();                                        // default constructor
  public : 
    RoadNode(const RoadNode& n);                       // copy constructor
    RoadNode(PositionXY& pos);                           // init constructor
    RoadNode(PositionXY& pos, RoadNode* ngh);            // init constructor
    virtual ~RoadNode();                               // destructor

    /************** Getters and setters ***************/
    inline const PositionXY& getPos() const { return *pos_; }   // return the position of the node
    RoadNode* getNeighbour(int i) const;               // return the i-th neighbour of the node
    int getNbNeighbour() const;                        // return the number of neighbours of n
    void addNeighbour(RoadNode* n);                    // add a new neighbour to the node
    bool isVisited() const;                            // return true if the roadnode has been already visited
    void visit();                                      // mark the node as visited
    inline virtual double getE() const {return this->E_;}     // return the estimation from the current node to the goal node
    inline virtual double getL() const {return this->L_;}     // return the length of the path computed up to this node
    inline virtual double getH() const {return (E_ + L_);}    // return the heuristic of the current node
    inline RoadNode* getPrev() const {return this->prev_;} // return the previous node from which has been computed the current node

    /**************** Travel functions ****************/
    virtual void reinit();                             // reinitialize all parameters of the node for the travel of the roadmap
    virtual void computeLfrom(RoadNode* prev);         // compute and set the length of the path to the current node 
    // from the previous node prev
    virtual void computeEto(RoadNode* goal);           // compute and set the estimation from the current 
    // node to the node goal

    /*************** Display functions ****************/
    virtual void display();                            // display the current node
  };

  class DRoadNode : public RoadNode
  {
  protected : 
    static double maxDObs_;                            // maximal distance to which an additional weight is added 
    // to the heuristic to visit that node
    double DObs_;                                      // distance from the position of the node to the nearest obstacles
    double sumIDO_;                                    // sum of the distances to obstacles

    /********** Constructors and destructors **********/
  protected : 
    DRoadNode();                                       // default constructor
  public : 
    DRoadNode(const DRoadNode& n);                     // copy constructor
    DRoadNode(PositionXY& pos,double dobs);              // init constructor
    DRoadNode(PositionXY& pos, RoadNode* ngh,double dobs); // init constructor
    virtual ~DRoadNode();                              // destructor

    /************** Getters and setters ***************/
    static double getMaxDistObs();                     // get the maximal distance to which an additional weight is added
    static void setMaxDistObs(double d);               // set the maximal distance to which an additional weight is added
    inline double D() const { return this->DObs_; }    // return the distance from the position of the node to the nearest obstacles
    double invD() const;                               // return the weight added to move away from obstacles
    inline double getSIDO() const { return this->sumIDO_; } // return the sum of the distances to obstacles
    inline virtual double getH() const { return (E_ + L_ + getSIDO()); } // return the heuristic of the current node

    /**************** Travel functions ****************/
    virtual void reinit();                             // reinitialize all parameters of the node for the travel of the roadmap
    virtual void computeLfrom(RoadNode* prev);         // compute and set the length of the path to the current node 
    // from the previous node prev
    virtual void computeSIDOfrom(DRoadNode* prev);     // compute the sum of distance to obstacles

    /*************** Display functions ****************/
    virtual void display();                            // display the current node
  };

  class RoadMap 
  {
  protected : 
    std::vector<RoadNode*> *lNode_;                    // list of nodes of the map
    Environment* env_;                                 // environment on which is constructed the map

    double maxDistNgh_;                                // maximal distance between two nodes
    int nbNodesNeeded_;                                // number of nodes needed on the map

    /********** Constructors and destructors **********/
  protected : 
    RoadMap();                                         // default constructor
  public : 
    RoadMap(const RoadMap& rm);                        // copy constructor
    RoadMap(Environment* env);                         // init constructor

    /************** Getters and setters ***************/
    virtual RoadNode* getNode(int i) const;            // return the i-th node of the roadmap
    virtual RoadNode* getNode(PositionXY& cfg) const;  // return the node of the roadmap which can be found in position cfg
						       // if it exists, return NULL else
    virtual int getIndNode(const PositionXY& cfg) const; // return the index of the node in the roadmap which can be found in position cfg
						       // return -1 if it doesn't exists
    int getNbNodes() const;                            // return the number of nodes of the roadmap
    double getMaxDistNgh() const;                      // return the maximal distance between two nodes
    void setMaxDistNgh(double d);                      // set the maximal distance between two nodes
    int getNbNodesNeeded() const;                      // return the number of nodes needed on the map
    void setNbNodesNeeded(int n);                      // set the number of nodes needed on the map
    Environment* getEnv() const;                       // return the environment of the roadmap

    /************** RoadMap Construction **************/
    virtual void construct();                          // construct the roadmap
    void addNode(RoadNode* n);                         // add a new node to the roadmap
    RoadNode* findNearNode(const PositionXY& pos) const; // find the nearest node of the roadmap from the position pos
    void reinit();                                     // reinitialize all parameters of the nodes of the roadmap for its travel

    /*************** Display functions ****************/
    virtual void display() const;                      // display the main features of the roadmap
    virtual void toGL(double time) const;              // display the roadmap in an openGL viewer
  };

  class DRoadMap : public RoadMap
  {
    /********** Constructors and destructors **********/
  protected : 
    DRoadMap();                                        // default constructor
  public : 
    DRoadMap(const DRoadMap& rm);                      // copy constructor
    DRoadMap(Environment* env);                        // init constructor

    /************** RoadMap Construction **************/
    virtual void construct();                          // construct the roadmap
  
    /***************** I/O functions ******************/
    virtual int load(const char* nomfic);              // load the roadmap from an input file
						       // return 0 if the process succeed, -1 else
    virtual int save(const char* nomfic);              // save the roadmap into an output file
						       // return 0 if the process succeed, -1 else
    
    /*************** Display functions ****************/
    virtual void toGL(double time) const;              // display the roadmap in an openGL viewer
  };

};
#endif // ROADMAP_HPP
