#ifndef SLOPEMAP_HPP
#define SLOPEMAP_HPP

#include "environment/inc/Map2D.hpp"
#include "environment/inc/DEM.hpp"

namespace mpc
{

                                    /**************************************************/
                                    /****************** Class SlopeMap ****************/
                                    /**************************************************/

  /** \class SlopeMap
   * \brief map representing the slope of a 3D robot environment
   */
  class SlopeMap : public Map2D
  {
  protected : 
    float** slope_;               /**< Slope of each cell */
    map2D_cell_t** cellState_;    /**< State of the slope map cells (CELL_FILLED or CELL_EMPTY) */

  protected : 
    float maxDist_;               /*< Maximal distance considered between two cells to compute the slope */
    int maxNbCells_;              /*< Maximal niumber of cells corresponding to maxDist_ */

    static float colDepth_;       /**< depth of the colors used to convert the map into images */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    SlopeMap();
    
  public : 
    /**
     *  \brief copy constructor
     *  \param map : the map to copy
     */
    SlopeMap(const SlopeMap& map);

    /**
     *  \brief init constructor
     *  \param nbCols : number of columns of the map table
     *  \param nbLines : number of lines of the map table
     *  \param xOrigin : X coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param yOrigin : Z coordinate of the top left corner of the (0,0) pixel in the global frame
     *  \param xScale : X size of a cell in meters
     *  \param yScale : Z size of a cell in meters
     */
    SlopeMap(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param maxDist : the maximal distance considered between two cells to compute the slope
     */
    SlopeMap(const DEM* dem, float maxDist=0);

    /**
     *  \brief destructor
     */
    virtual ~SlopeMap();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of SlopeMap is allocated
     */
    virtual SlopeMap* copy() const;


    /***************** Initialization *****************/

  protected : 
    /*
     *  \brief init tables from their dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    virtual void initTables(int nbCols,int nbLines);
    
    /*
     *  \brief delete all the content of the slope map tables
     */
    virtual void clearTables();


    /******************** Accessors *******************/

  private : 
    /*
     *  \brief set the cell (i,j) as undefined (no knowledge about the slope)
     *  \param (i,j) : index of the cell in the grid     
     */
    void setUndefCell(int i,int j);

    /*
     *  \brief set the slope of cell (i,j)
     *  \param (i,j) : index of the cell in the grid     
     *  \param slope : the new slope
     */
    void setSlope(int i,int j,float slope);

  public : 
    /**
     *  \brief get informations about the cell (i,j)
     *  \param (i,j) : index of the cell in the grid
     *  \param cellSt : the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     *  \param slope : the resulting slope
     */
    void getCellInfo(int i, int j, map2D_cell_t& cellSt, float& slope) const;


    /*************** Slope map building ***************/

  private : 
    /*
     *  \brief compute the slope between two cells (i,j) and (k,l)
     *  \param dem : the input dem
     *  \param (i,j) : index of the cell in the grid
     *  \param zIJ : altitude in the cell of index (i,j)
     *  \param (k,l) : index of the second cell in the grid
     *  \param slope : the resulting slope
     *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     *
     *  The cell (i,j) is supposed to be filled on the dem
     */
    map2D_cell_t computeSlope(const DEM* dem, int i, int j, float zIJ, int k, int l, float& slope) const;

    /*
     *  \brief compute the slope in a cell (i,j)
     *  \param dem : the input dem
     *  \param (i,j) : index of the cell in the grid
     *  \param slope : the resulting slope
     *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     */
    map2D_cell_t computeSlope(const DEM* dem, int i, int j, float& slope) const;

    /*
     *  \brief initialize the maximal number of cells considered to compute the slope
     *  \param maxDist : the maximal distance considered to compute the slope
     */
    void initMaxNbCells(float maxDist);

  public : 
    /**
     *  \brief build the slope map from an input data elevation map
     *  \param dem : the input dem
     *  \param maxDist : the maximal distance considered between two cells to compute the slope
     *  \return 0 if the process succeed, -1 else
     */
    int buildFromDEM(const DEM* dem, float maxDist=0);


    /************ Input / Output functions ************/

  protected : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    virtual void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    virtual int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    virtual void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const;


    /**************** Utility functions ***************/

  protected : 
    /**
     *  \brief extract a sub-map of the current map
     *  \param iMin : the number of the first column to keep
     *  \param iMax : the number of the last column to keep
     *  \param jMin : the number of the first line to keep
     *  \param jMax : the number of the last line to keep
     *  \return a pointer to the sub-map if the process succeeded, NULL otherwise
     *
     *  A new instance of SlopeMap is created
     */
    virtual SlopeMap* subMap(int iMin, int iMax, int jMin, int jMax) const;

  public : 
    /**
     *  \brief display the main features of the slope map
     *  \param dispGrid : display the grid if true
     */
    virtual void display(bool dispGrid) const;
  };

};

#endif // SLOPEMAP_HPP
