#ifndef VARSLOPEMAP_HPP
#define VARSLOPEMAP_HPP

#include "environment/inc/Map2D.hpp"
#include "environment/inc/DEM.hpp"
#include "environment/inc/SlopeMap.hpp"

namespace mpc
{

                                    /**************************************************/
                                    /**************** Class VarSlopeMap ***************/
                                    /**************************************************/

  /** \class VarSlopeMap
   * \brief map representing the slope derivative of a 3D robot environment
   */
  class VarSlopeMap : public SlopeMap
  {
  protected : 
    float** varSlope_;            /**< Slope derivative of each cell */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    VarSlopeMap();
    
  public : 
    /**
     *  \brief copy constructor
     *  \param map : the map to copy
     */
    VarSlopeMap(const VarSlopeMap& map);

    /**
     *  \brief init constructor (from a data elevation map)
     *  \param dem : the input dem
     *  \param maxDist : the maximal distance considered between two cells to compute the slope
     */
    VarSlopeMap(const DEM* dem, float maxDist=0);

    /**
     *  \brief destructor
     */
    virtual ~VarSlopeMap();

    /**
     *  \brief return a copy of the current map
     *  \return a pointer to the copy of the current map
     *
     *  A new instance of SlopeMap is allocated
     */
    virtual VarSlopeMap* copy() const;


    /***************** Initialization *****************/

  private : 
    /*
     *  \brief init the slope derivative table from its dimesions
     *  \param number of columns of the table
     *  \param number of lines of the table
     */
    virtual void initTableVS(int nbCols,int nbLines);
    
    /*
     *  \brief delete all the content of the slope map tables
     */
    virtual void clearTables();


    /******************** Accessors *******************/

  private : 
    /*
     *  \brief set the slope of cell (i,j)
     *  \param (i,j) : index of the cell in the grid     
     *  \param varSlope : the new slope derivative
     */
    void setVarSlope(int i, int j, float varSlope);

  public : 
    /**
     *  \brief get informations about the cell (i,j)
     *  \param (i,j) : index of the cell in the grid
     *  \param cellSt : the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     *  \param slope : the resulting slope
     *  \param varSlope : the resulting slope derivative
     */
    void getCellInfo(int i, int j, map2D_cell_t& cellSt, float& slope, float& varSlope) const;


    /*************** Slope map building ***************/

  private : 
    /*
     *  \brief compute the slope between two cells (i,j) and (k,l)
     *  \param dem : the input dem
     *  \param (i,j) : index of the cell in the grid
     *  \param slopeIJ : altitude in the cell of index (i,j)
     *  \param (k,l) : index of the second cell in the grid
     *  \param varSlope : the resulting slope derivative
     *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     *
     *  The cell (i,j) is supposed to be filled on the dem
     */
    map2D_cell_t computeVarSlope(int i, int j, float slopeIJ, int k, int l, float& varSlope) const;

    /*
     *  \brief compute the slope in a cell (i,j)
     *  \param dem : the input dem
     *  \param (i,j) : index of the cell in the grid
     *  \param varSlope : the resulting slope derivative
     *  \return the resulting cell state (CELL_FILLED is the slope has been computed, CELL_EMPTY otherwise)
     */
    map2D_cell_t computeVarSlope(int i, int j, float& varSlope) const;

  public : 
    /**
     *  \brief build the slope map from an input data elevation map
     *  \param dem : the input dem
     *  \param maxDist : the maximal distance considered between two cells to compute the slope
     *  \return 0 if the process succeed, -1 else
     */
    int buildVSFromDEM(const DEM* dem, float maxDist=0);


    /************ Input / Output functions ************/

  protected : 
    /*
     *  \brief init the Z color parameters
     *  \param minColZ : output minimal Z value used to convert the DEM into an image
     *  \param maxColZ : output maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : output Z scale value used to convert the DEM into an image
     */
    virtual void initZ_ColorParams(float &minColZ, float &maxColZ, float &scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \return the colour corresponding to a cell of the DEM
     */
    virtual int greyColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ) const;

    /*
     *  \brief return the colour corresponding to a cell of the DEM
     *  \param minColZ : minimal Z value used to convert the DEM into an image
     *  \param maxColZ : maximal Z value used to convert the DEM into an image
     *  \param scaleColZ : Z scale value used to convert the DEM into an image
     *  \param R : the output red colour
     *  \param G : the output green colour
     *  \param B : the output blue colour
     */
    virtual void rgbColPix(int i,int j,float minColZ, float maxColZ, float scaleColZ,int& R,int& G,int& B) const;


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief display the main features of the var slope map
     *  \param dispGrid : display the grid if true
     */
    virtual void display(bool dispGrid) const;
  };


  

};

#endif // VARSLOPEMAP_HPP
