#ifndef ITGFUNC_HPP
#define ITGFUNC_HPP

#include "math/inc/Operation.hpp"

                                    /**************************************************/
                                    /****************** Class ItgFunc *****************/
                                    /**************************************************/

namespace mpc
{
  class ItgFunc : public Operation
  {
    std::vector<Disc_Point> disc_points_;              /**< discontinuity points vector */
    bool dp_updated_;                                  /**< true if the discontinuity points vector has been updated */
    double cst_;                                       /**< integration constant */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    ItgFunc();

  public : 
    /**
     *  \brief init constructor
     *  \param f : the operand
     *  \param cst : integration constant
     *
     *  Input pointer is directly assigned to the structure
     */
    ItgFunc(Function* f, double cst=0);

//     /**
//      *  \brief init constructor
//      *  \param f : the first operand
//      *
//      *  A copy of the input functions is performed
//      */
//     ItgFunc(const Function& f);

    /**
     *  \brief copy constructor
     */
    ItgFunc(const ItgFunc& op);

    /**
     *  \brief destructor
     */
    virtual ~ItgFunc();

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual ItgFunc* copy() const;


    /**************** Quadrature tools ****************/

  private : 
    /*
     *  \brief Quadrature tools...
     */  
    void gauleg(const double x1, const double x2, double x[], double w[], int n) const;

    /*
     *  \brief Quadrature tools...
     */
    double qgaus(double (ItgFunc::*func)(double) const, const double a, const double b) const;


    /**************** Setters & Getters ***************/

  private : 
    /*
     *  \brief return the value of the integrated function at a given parameter t
     *  \param t : the input parameter value
     *  \return the value of the integrated function at a given parameter t
     */
    double getIntegratedFuncVal(double t) const;

  public : 
    /**
     *  \brief return the value of the integration function at a given parameter t
     *  \param t : the input parameter value
     *  \return the value of the integration function at a given parameter t
     */
    virtual double getVal(double t) const;
  
    /******************** Operators *******************/

  public : 
    /**
     *  \brief multiply the current function by a constant value
     *  \param v : the input multiplication coefficient
     */
    virtual void operator*=(double v);

    /************ Integration & Derivation ************/

  public : 
    /**
     *  \brief return the derivation of the current function
     *  \return the derivation of the current function
     */
    virtual Function* derivation() const;

    //   virtual Function* derivation(const char* nameVar) = 0; // return the derivation of the current polynome wrt the variable named nameVar

  protected : 
    /**
     *  \brief return the integration of the current function
     *  \return the integration of the current function
     */
    virtual Function* integration() const;

  public : 
    /**
     *  \brief return the integration of the current function
     *  \param cste : the integration constant at t=0
     *  \return the integration of the current function
     */
    virtual Function* integration(double cste) const;

    /**
     *  \brief return the integration of the current function
     *  \param cste,binf : cste is the integration constant value defined at t=binf
     *  \return the integration of the current function
     */
    virtual Function* integration(double binf,double cste) const;

    /****************** Roots search ******************/

  public : 
    /**
     *  \brief return the real roots of this function
     *  \param r : the list of roots of the function
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r) const;

    /**
     *  \brief return the roots of the equation eq - v = 0
     *  \param r : the list of roots of the function
     *  \param v : the input constant value
     *  \return the number of roots of the function
     */
    virtual int realRoots(std::vector<double>& r,double v) const;

    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief display the polynomial
     */
    virtual void display() const;

    /**
     *  \brief shift the function on Y axis from value v
     *  \param v : the shifting value
     */
    virtual void shiftY(double v);

    /**
     *  \brief shift the time parameter of the function
     *  \param t : the shifting value
     */
    virtual void shiftTime(double t);

    /***************** Discontinuities ****************/

    /**
     *  \brief search for the discontinuities of the function
     *  \param discPoints : resulting discontinuity points
     */
    virtual void getVecDiscPoints(std::vector<Disc_Point>& discPoints) const;

    /**
     *  brief update the discontinuity points from the pointed function
     */
    void updateDiscPoints();
  };

};
#endif // ITGFUNC_HPP
