#ifndef MULTIITVFUNCTION_HPP
#define MULTIITVFUNCTION_HPP

#include "math/inc/Polynomial.hpp"
#include "math/inc/Function.hpp"
#include <list>
#include <stdlib.h>

// class mpc::Polynomial;

                                    /**************************************************/
                                    /**************** Class ItvFunction ***************/
                                    /**************************************************/

namespace mpc
{
  class Polynomial;

  /**  \class ItvFunction
   *  \brief Class defining an interval of a piecewise polynomial function
   */
  class ItvFunction
  {
  public : 
    double binf_;                                   /**< lower bound of the interval */
    double bsup_;                                   /**< upper bound of the interval */
    Function* fct_;                                 /**< function defined on the interval */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    ItvFunction();

    /**
     *  \brief copy constructor
     *  \param itvf : the interval to copy
     *
     *  The function pointed by the input interval is copied to
     */
    ItvFunction(const ItvFunction& itvf);

    /**
     *  \brief init constructor
     *  \param binf : the lower bound of the interval
     *  \param bsup : the upper bound of the interval
     *  \param f : a pointer to the function defined on the current interval
     *
     *  f is not copied here, a new instance of the function must be allocated before
     */
    ItvFunction(double binf,double bsup,Function* f);  // init constructor

    /**
     *  \brief desctructor
     */
    ~ItvFunction();

    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief shift the time bounds of the current interval
     *  \param dt : the shifting time value
     */
    void shiftTime(double dt);

    /**
     *  \brief return true if t \in [binf_;bsup_]
     *  \param t : the input value at which the interval must be evaluated
     *  \return true if t \in [binf_;bsup_]
     */
    bool isDefinedAt(double t) const;

    /**
     *  \brief display the interval
     */
    void display() const;

    /**
     *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
     *  \param polyCoef : the resulting coefficients of the polynomial function
     *  \param binf : the lower boudary of the interval
     *  \param bsup : the upper boudary of the interval
     *  \return the number of values in the resulting vector, or -1 if the current function is not a polynomial function
     */
    int getPolyCoef(std::vector<double>& polyCoef, double& binf, double& bsup) const;


    /**************** Static functions ****************/

  public : 
    /**
     *  \brief return true if t is include in one of the intervals
     *  \param t : the input value at which the intervals must be evaluated
     *  \param l : the list of intervals to evaluate
     */
    static bool isIncludeIn(double t,std::list<ItvFunction>* l);
  };

};

                                    /**************************************************/
                                    /************* Class MultiItvFunction *************/
                                    /**************************************************/

namespace mpc
{
  /** \class MultiItvFunction
   *  \brief Class defining a piecewise polynomial function
   */
  class MultiItvFunction : public Function
  {
  protected : 
    std::list<ItvFunction*>* vecItvFunc_;           /**< list of intervals defining the piecewise function */

    /********** Constructors and destructors **********/

  public : 
    /**
     *  \brief default constructor
     */
    MultiItvFunction();

    /**
     *  \brief copy constructor
     *  \param f : the function to copy
     *
     *  Each function defined on an interval is copied here
     */
    MultiItvFunction(const MultiItvFunction& f);

    /**
     *  \brief init constructor (from a unique function defined on interval [binf;bsup])
     *  \param f : function defined on the interval
     *  \param binf : lower bound of the function 
     *  \param bsup : upper bound of the function 
     *
     *  The pointer of function f is directly assigned to the new interval created (no copy)
     */
    MultiItvFunction(const Function* f,double binf=0,double bsup=-1);
//     MultiItvFunction(const Function* f,double binf,double bsup);

    /**
     *  \brief init constructor (from a vector of polynomial coefficients)
     *  \param vecPolys : the vector of polynomial coefficients
     *
     *  The structure of each interval vector must be described as follows : 
     *  vecPolys.at(i).at(0) -> lower boundary of the current interval
     *  vecPolys.at(i).at(1) -> upper boundary of the current interval
     *  vecPolys.at(i).at(2+) -> coefficient of degree 0+ of the current interval
     */
    MultiItvFunction(const std::vector<std::vector<double> >& vecPolys);

    /**
     *  \brief destructor
     */
    virtual ~MultiItvFunction();

    /**
     *  \brief return a copy of the current function
     *  \return a copy of the current function
     *
     *  A new instance of function is allocated and returned here
     */
    virtual MultiItvFunction* copy() const;

  
    /**************** Setters & Getters ***************/

    /**
     *  \brief return the value of the polynomial at a given parameter t
     *  \param t : the input value at which the function must be evaluated
     *  \return the value of the polynomial at a given parameter t
     */
    virtual double getVal(double t) const;

    /**
     *  \brief return the function used at the input parameter t
     *  \param t : the input parameter
     *  \return a pointer to the function defined at the value t if one is defined, return NULL else
     */
    Function* getFunction(double t) const;

    /**
     *  \brief return the function used at the input parameter t
     *  \param t : the input parameter
     *  \param binf : the resulting lower bound
     *  \return a pointer to the function defined at the value t if one is defined, return NULL else
     */
    Function* getFunctionBInf(double t, double& binf) const;

    /**
     *  \brief return the minimal bound of the function
     *  \return the minimal bound of the function
     */
    virtual double getBInf() const;
  
    /**
     *  \brief return the maximal bound of the function
     *  \return the maximal bound of the function
     */
    virtual double getBSup() const;

    /**
     *  \brief destroy all intervals of the function
     */
    void clearItv();

    /**
     *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
     *  \param i : the input index of the resulting polynomial interval
     *  \param polyCoef : the resulting coefficients of the polynomial function
     *  \param binf : the lower boudary of the interval
     *  \param bsup : the upper boudary of the interval
     *  \return the number of values in the resulting vector, or -1 if the current function is not a polynomial function
     */
    int getPolyCoef(int i, std::vector<double>& polyCoef, double& binf, double& bsup) const;

    /**
     *  \brief get back the polynomial coefficients of the current interval if it corresponds to a ploynomial function
     *  \param polyCoef : the resulting coefficients of the polynomial function
     *  \return 0 if the process succeed, -1 else
     *
     *  The structure of each interval vector is described as follows : 
     *  vec.at(0) -> lower boundary of the current interval
     *  vec.at(1) -> upper boundary of the current interval
     *  vec.at(2+) -> coefficient of degree 0+ of the current interval
     */
    int getPolyCoef(std::vector<std::vector<double> >& polyCoef) const;
  

    /******************** Operators *******************/

    /**
     *  \brief multiply the current MultiItvFunction by a constant value
     *  \param v : the constant value
     */
    virtual void operator*=(double v);

    /**
     *  \brief multiply the current piecewise function by a polynomial function
     *  \param f2 : the input polynomial function
     */
    virtual void operator*=(const Polynomial& f2);

    /**
     *  \brief multiply the current piecewise function by an other piecewise function
     *  \param f2 : the input piecewise function
     */
    virtual void operator*=(const MultiItvFunction& f2);

    /**
     *  \brief add a piecewise function to the current piecewise function
     *  \param f2 : the input piecewise function
     */
    virtual void operator+=(const MultiItvFunction& f2);

    /**
     *  \brief substract a piecewise function to the current piecewise function
     *  \param f2 : the input piecewise function
     */
    virtual void operator-=(const MultiItvFunction& f2);

    /************ Integration & Derivation ************/

    /**
     *  \brief return the derivation of the current function
     *  \return the derivation of the current function
     *
     *  A new instance of piecewise function is allocated and returned here
     */
    virtual MultiItvFunction* derivation() const;

    /**
     *  \brief return the integration of the current function
     *  \return the integration of the current function
     *
     *  A new instance of piecewise function is allocated and returned here
     */
    virtual MultiItvFunction* integration() const;

    /**
     *  \brief return the integration of the current function
     *  \param cste : the constant value defined at t=0
     *  \return the integration of the current function
     *
     *  A new instance of piecewise function is allocated and returned here
     */
    virtual MultiItvFunction* integration(double cste) const;

    /**
     *  \brief return the integration of the current function
     *  \param cste,binf : cste is the constant value defined at t=binf
     *  \return the integration of the current function
     *
     *  A new instance of piecewise function is allocated and returned here
     */
    virtual MultiItvFunction* integration(double binf,double cste) const;
  

    /***************** Absolute value *****************/

    /**
     *  \brief compute the absolute value of the current function
     */
    void addAbsoluteValue();


    /****************** Roots search ******************/

//     /**
//      *  \brief return the real roots of this function
//      *  \param r : the resulting list of the roots of the current function
//      */
//     virtual int realRoots(std::vector<double>& r) const;

    /**
     *  \brief return the real roots of the equation f(t) - v = 0, where f is the current function
     *  \param r : the resulting list of the roots of the equation
     */
    virtual int realRoots(std::vector<double>& r,double v=0.0) const;
  
    /**************** Interval changes ****************/

    /**
     *  \brief change one or several intervals of the current piecewise function by f
     *  \param [bi;bf] is the interval on which the function must be changed
     *  \param f : the function to replace the current one on interval [bi;bf]
     *
     *  The input function f is copied to be set in the new interval created
     */
    void changeItv(double bi,double bs,MultiItvFunction* f);

    /**
     *  \brief change one or several intervals of the current piecewise function by f
     *  \param [bi;bf] is the interval on which the function must be changed
     *  \param f : the function to replace the current one on interval [bi;bf]
     *
     *  The input function f is copied to be set in the new interval created
     */
    void changeItv(double bi,double bs,Function* f);

    /**
     *  \brief add a lower bound over the function
     *  \param bmin : the lower bound to add on the current function
     *  \param lRes : the list of intervals that have been bounded
     */
    void addMinBound(double bmin,std::list<ItvFunction>* lRes=NULL, bool debug=false);

    /**
     *  \brief add an upper bound over the function
     *  \param bmin : the upper bound to add on the current function
     *  \param lRes : the list of intervals that have been bounded
     */
    void addMaxBound(double bmax,std::list<ItvFunction>* lRes=NULL);

    /**
     *  \brief add a lower bound over the function
     *  \param bmin : the lower bound to add on the current function
     *  \param lRes : the list of intervals that have been bounded
     */
    void addMinBound(MultiItvFunction* f,std::list<ItvFunction>* lRes=NULL);

    /**
     *  \brief add an upper bound over the function
     *  \param bmin : the upper bound to add on the current function
     *  \param lRes : the list of intervals that have been bounded
     */
    void addMaxBound(MultiItvFunction* f,std::list<ItvFunction>* lRes=NULL);

    //   std::list<ItvFunction>* addMinBound2(MultiItvFunction* f);            // add a minimal bound over the function, return the bounded intervals
    //   std::list<ItvFunction>* addMaxBound2(MultiItvFunction* f);            // add a maximal bound over the function, return the bounded intervals
  
    /**
     *  \brief get the list of intervals defined between bounds [bmin;bmax]
     *  \param [bmin;bmax] : the evaluated bounds
     */
    std::list<ItvFunction*> getListItvBetween(double bmin,double bmax);

    /**
     *  \brief get the list of value at which the definition of the function change
     *  \return the resulting list
     * 
     *  Avector<double> new instance of std::vector<double> is returned
     */
    std::vector<double>* getListChangePoints();

    /**
     *  \brief shift the time bounds of all time intervals
     *  \param dt : the shifting time value
     */
    void shiftTime(double dt);

    /**
     *  \brief concatenate two piecewise polynomial functions
     *  \param f : the input fonction to concatenate
     */
    void concat(const MultiItvFunction& f);

  
    /**************** Utility functions ***************/

    /**
     *  \brief display the piecewise polynomial function
     */
    virtual void display() const;

    /**
     *  \brief shift the current function on Y axis from value v
     *  \param v : the shift value
     */
    virtual void shiftY(double v);

    /**
     *  \brief check the consistency of intervals defined in the current piecewise function
     *  \return false if there is an unconsistency in the intervals defined
     */
    bool checkItvConsistency() const;


    /***************** Discontinuities ****************/

    /**
     *  \brief search for the discontinuities of the function
     *  \param discPoints : resulting discontinuity points
     */
    virtual void getVecDiscPoints(std::vector<Disc_Point>& discPoints) const;
  };

};
#endif // MULTIITVFUNCTION
