#ifndef MPC_PCTRAJ_HPP
#define MPC_PCTRAJ_HPP

#include "motion/inc/CTrajectory.hpp"
#include "math/inc/MultiItvFunction.hpp"

#include "system/inc/ManaSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class PCTraj ******************/
                                    /**************************************************/

  /** \class PCTraj
   *  \brief Piecewise polynomial control trajectory executed by a system
   */
  class PCTraj : public CTrajectory
  {
  protected : 
    ctraj_t type_;                                    /**< type of the control trajectory */
    int nbC_;                                         /**< dimension of the control vector */
    MultiItvFunction** tabFunc_;                      /**< table containing the piecewise parametric profiles of the functions defining the control trajectory */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    PCTraj();

    /**
     *  \brief init constructor
     *  \param dim : dimension of teh control vector
     *  \param startTime : the starting time of the control trajectory
     *  \param endTime : the ending time of the control trajectory
     */
    PCTraj(int dim);

  public : 
    /**
     *  \brief copy constructor
     */
    PCTraj(const PCTraj& traj);

    /**
     *  \brief destructor
     */
    virtual ~PCTraj();

    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     *
     *  Return a new instance of PCTraj
     */
    virtual PCTraj* copy() const = 0;

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief return the dimension of the control vector
     *  \return the dimension of the control vector
     */
    inline int nbC() const { return this->nbC_; }

//     /**
//      *  \brief getter : first control of the trajectory
//      *  \return the first control of the trajectory
//      */
//     virtual const Control& firstCtrl() const = 0;

//     /**
//      *  \brief getter : last control of the trajectory
//      *  \return the last control of the trajectory
//      */
//     virtual const Control& lastCtrl() const = 0;

//     /**
//      *  \brief getter : get the control defined at time t
//      *  \param t : the input time
//      *  \return return the control defined at time t
//      */
//     virtual const Control& at(double t) const = 0;

    /**
     *  \brief return the type of the trajectory
     *  \return the type of the trajectory
     */
    inline ctraj_t type() const { return this->type_; }

    /**
     *  \brief return the start time of the control trajectory
     *  \return the start time of the control trajectory
     */
    double startTime() const;

    /**
     *  \brief return the end time of the control trajectory
     *  \return the end time of the control trajectory
     */
    double endTime() const;


    /******** Control trajectory modifications ********/

  public : 
    /**
     *  \brief concatenate two trajectories
     *  \param traj : the trajectory to concatenate to the current one
     */
    virtual void concat(const PCTraj& traj);

    /**
     *  \brief shift the trajectory by time t
     *  \param t : the time to shift the trajectory
     */
    virtual void shift(double t);


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief sample the control trajectory
     *  \return a sampled control trajectory
     *  \param nb : the number of samples
     */
    virtual CTraj* sample(int nb=20) const = 0;

    /**
     *  \brief sample the control trajectory
     *  \return a sampled control trajectory
     *  \param dt : the time interval between two samples
     */
    virtual CTraj* sample(double dt) const = 0;

    /**
     *  \brief diplay the control trajectory
     */
    virtual void display() const = 0;
  };


                                    /**************************************************/
                                    /**************** Class DD2PCTraj ****************/
                                    /**************************************************/

  /**
   * \enum dd2_sampling_t
   * \brief Enumeration of the different control trajectory types
   *
   * DD2_SPG_VEL   : Sample the velocity profiles of the DD2 trajectory
   * DD2_SPG_ACC   : Sample the acceleration profiles of the DD2 trajectory
   */
  typedef enum {
    DD2_SPG_VEL  =  0,
    DD2_SPG_ACC  =  1
  } dd2_sampling_t;

  /** \class DD2PCTraj
   *  \brief Piecewise polynomial control trajectory executed by a DD2 robotic system (linear velocity - angular velocity)
   */
  class DD2PCTraj : public PCTraj
  {
    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    DD2PCTraj();

  public : 
    /**
     *  \brief copy constructor
     */
    DD2PCTraj(const DD2PCTraj& traj);

    /**
     *  \brief init constructor (from a constant control)
     *  \param v : the constant linear velocity control
     *  \param w : the constant angular velocity control
     *  \param t : the duration of the control
     */
    DD2PCTraj(double v, double w, double t);

    /**
     *  \brief init constructor (from a linear control)
     *  \param v0 : the initial linear velocity
     *  \param a : the constant linear acceleration
     *  \param w0 : the initial angular velocity
     *  \param mu : the constant angular acceleration
     *  \param t : the duration of the control
     */
    DD2PCTraj(double v0, double a, double w0, double mu, double t);

    /**
     *  \brief init constructor (from two polynomial functions)
     *  \param polyV : the polynomial function representing the linear velocity control
     *  \param polyW : the polynomial function representing the angular velocity control
     *  \param t : the duration of the control
     */
    DD2PCTraj(const Polynomial& polyV, const Polynomial& polyW, double t);

    /**
     *  \brief init constructor (from two polynomial acceleration functions and two initial velocities)
     *  \param v0 : the initial linear velocity
     *  \param polyAcc : the polynomial function representing the linear acceleration control
     *  \param w0 : the initial angular velocity
     *  \param polyAAcc : the polynomial function representing the angular acceleration control
     *  \param t : the duration of the control
     */
    DD2PCTraj(double v0, const Polynomial& polyAcc, double w0, const Polynomial& polyAAcc, double t);

    /**
     *  \brief init constructor (from the coefficients of the linear and angular velocity profiles)
     *  \param vel : vector containing the coefficient of the piecewise polynomial linear velocity profile
     *  \param avel : vector containing the coefficient of the piecewise polynomial angular velocity profile
     */
    DD2PCTraj(const std::vector<std::vector<double> >& vecVel, const std::vector<std::vector<double> >& vecAVel);

    /**
     *  \brief destructor
     */
    virtual ~DD2PCTraj();

    /**
     *  \brief return a copy of the current control trajectory
     *  \return a copy of the current control trajectory
     *
     *  Return a new instance of PCTraj
     */
    virtual DD2PCTraj* copy() const;

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : first control of the trajectory
     *  \return the first control of the trajectory
     */
    virtual DD2Control firstCtrl() const;

    /**
     *  \brief getter : last control of the trajectory
     *  \return the last control of the trajectory
     */
    virtual DD2Control lastCtrl() const;

    /**
     *  \brief getter : get the control defined at time t
     *  \param t : the input time
     *  \return return the control defined at time t
     */
    virtual DD2Control at(double t) const;

    /**
     *  \brief return the linear velocity value to apply at time t
     *  \param t : the input time
     *  \return the linear velocity value to apply at time t
     */
    double v(double t) const;

    /**
     *  \brief return the angular velocity value to apply at time t
     *  \param t : the input time
     *  \return the angular velocity value to apply at time t
     */
    double w(double t) const;

    /**
     *  \brief return the linear acceleration value to apply at time t
     *  \param t : the input time
     *  \return the linear acceleration value to apply at time t
     */
    double acc(double t) const;

    /**
     *  \brief return the angular acceleration value to apply at time t
     *  \param t : the input time
     *  \return the angular acceleration value to apply at time t
     */
    double aacc(double t) const;

    /**
     *  \brief return a pointer to the linear velocity function
     *  \return a pointer to the linear velocity function
     */
    inline const MultiItvFunction* fctVel() const { return tabFunc_[0]; }

    /**
     *  \brief return a pointer to the linear velocity function
     *  \return a pointer to the linear velocity function
     */
    inline const MultiItvFunction* fctAVel() const { return tabFunc_[1]; }

    /**
     *  \brief return a pointer to the linear acceleration function
     *  \return a pointer to the linear acceleration function
     */
    inline const MultiItvFunction* fctAcc() const { return tabFunc_[2]; }

    /**
     *  \brief return a pointer to the angular acceleration function
     *  \return a pointer to the angular acceleration function
     */
    inline const MultiItvFunction* fctAAcc() const { return tabFunc_[3]; }


//     /**************** State computation ***************/

//   public : 
//     /**
//      *  \brief return the state reached from s at time t after application of the control trajectory
//      *  \param s0 : the initial state from which starts the control trajectory
//      *  \param t : the time at which the output state is evaluated
//      *  \param sys : the robotic system constraints
//      *  \return the state reached from s at time t after application of the control trajectory
//      */
//     DD2State getStateReachedFrom(const DD2State& s0, double t, const DD2Sys* sys=NULL) const;

//     /**
//      *  \brief return the state reached from s at time t after application of the control trajectory
//      *  \param s0 : the initial state from which starts the control trajectory
//      *  \param t : the time at which the output state is evaluated
//      *  \param sys : the robotic system constraints
//      *  \return the state reached from s at time t after application of the control trajectory
//      */
//     DD2StateT getStateReachedFrom(const DD2StateT& s0, double t, const DD2Sys* sys=NULL) const;


    /**************** Boundary functions **************/

  public : 
    /**
     *  \brief add a minimal and a maximal bound over the linear acceleration profile
     *  \param minV : the minimal bound
     *  \param maxV : the maximal bound
     */
    void addAccBounds(double minA,double maxA);

    /**
     *  \brief add a minimal and a maximal bound over the angular acceleration profile
     *  \param minAV : the minimal bound
     *  \param maxAV : the maximal bound
     */
    void addAAccBounds(double minAA,double maxAA);

    /**
     *  \brief add a minimal and a maximal bound over the linear velocity profile
     *  \param minV : the minimal bound
     *  \param maxV : the maximal bound
     */
    void addVelBounds(double minV,double maxV);

    /**
     *  \brief add a minimal and a maximal bound over the angular velocity profile
     *  \param minAV : the minimal bound
     *  \param maxAV : the maximal bound
     */
    void addAVelBounds(double minAV,double maxAV);


    /**************** Utility functions ***************/

  public : 
    /**
     *  \brief sample the velocity control trajectory
     *  \return a sampled velocity control trajectory
     *  \param nb : the number of samples
     */
    virtual DD2CTraj* sampleVelTraj(int nb=20) const;

    /**
     *  \brief sample the velocity control trajectory
     *  \return a sampled velocity control trajectory
     *  \param dt : the time interval between two samples
     */
    virtual DD2CTraj* sampleVelTraj(double dt) const;

    /**
     *  \brief sample the acceleration control trajectory
     *  \return a sampled acceleration control trajectory
     *  \param nb : the number of samples
     */
    virtual DD2AccCTraj* sampleAccTraj(int nb=20) const;

    /**
     *  \brief sample the acceleration control trajectory
     *  \return a sampled acceleration control trajectory
     *  \param dt : the time interval between two samples
     */
    virtual DD2AccCTraj* sampleAccTraj(double dt) const;

    /**
     *  \brief sample the control trajectory
     *  \return a sampled control trajectory
     *  \param nb : the number of samples
     */
    virtual DD2CTraj* sample(int nb=20) const;

    /**
     *  \brief sample the control trajectory
     *  \return a sampled control trajectory
     *  \param dt : the time interval between two samples
     */
    virtual DD2CTraj* sample(double dt) const;

    /**
     *  \brief diplay the control trajectory
     */
    virtual void display() const;

    /**
     *  \brief return the polynomial coefficients of the linear and angular velocity profiles
     *  \param coefPolyVel : the resulting coefficients of the polynomial linear velocity profiles
     *  \param coefPolyAVel : the resulting coefficients of the polynomial angular velocity profiles
     */
    void getPolyCoef(vector<vector<double> >& coefPolyVel, vector<vector<double> >& coefPolyAVel) const;
  };

};


#define ManaPCTraj DD2PCTraj


#endif // MPC_PCTRAJ_HPP
