#ifndef MPC_PATH_HPP
#define MPC_PATH_HPP

#include "common/inc/Config.hpp"
#include "common/inc/Position.hpp"
#include "graphics/inc/RGBImage.hpp"
//using namespace mpc;

/**
 * \enum path_t
 * \brief Enumeration of the different path types
 *
 * ND_PATH        : Undefined path
 * PM_PATH        : Path of a point mass system
 * DPM_PATH       : Path of a directed point mass system
 * CL_PATH        : Path of a car-like system
 * SS_PATH        : Path of a spaceship system
 * DD_PATH        : Path of a differential drive system
 * WC_PATH        : Path of a wheelchair system
 * STR_LINE_PATH  : Straight line path
 * CIRC_BOW_PATH  : Circle bow path
 * CLOTH_BOW_PATH : Clothoid bow path
 * GEOM_PATH      : Path made of geometrical primitives (straight lines, circle bows, clothoids bows)
 */
typedef enum {
  ND_PATH        =  0,
  PM_PATH        =  1,
  DPM_PATH       =  2,
  CL_PATH        =  3,
  SS_PATH        =  4,
  DD_PATH        =  5,
  WC_PATH        =  2,  // 6
  DD2_PATH       =  2,
  MANA_PATH      =  2,
  STR_LINE_PATH  =  7,
  CIRC_BOW_PATH  =  8,
  CLOTH_BOW_PATH =  9,
  GEOM_PATH      = 10,
  XYTHS_PATH     = 11
} path_t;

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class Path *******************/
                                    /**************************************************/

/** \class Path
 *  \brief Path followed by a system
 */
class Path
{
protected : 
  path_t type_;                                    /**< type of the path */

  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor
   */
  Path();

public : 
  /**
   *  \brief destructor
   */
  virtual ~Path();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the type of the path
   *  \return the type of the path
   */
  inline path_t type() const { return this->type_; }

  /**
   *  \brief compute the length of the path
   *  \return the length of the path
   */
  virtual double getLength() const = 0;
  
  /**************** Path modification ***************/

public : 
  /**
   *  \brief reverse the current path
   */
  virtual void reverse() = 0;
    
  /**************** Utility functions ***************/

public : 
  /**
   *  \brief diplay the path
   */
  virtual void display() const = 0;

  /**
   *  \brief display in an openGL viewer
   */
  virtual void toGL() const = 0;
  
  /**
   *  \brief return the barycenter of the path in the workspace
   */
  virtual mpc::PositionXY center() const = 0;

  /**
   *  \brief compute an approximative radius of the set of all configurations of the path
   *  \return an approximative radius of the set of all configurations of the path
   */
  virtual double radius() const = 0;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic) = 0;
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const = 0;
};

};
#endif // MPC_PATH_HPP


#ifndef MPC_CPATH_HPP
#define MPC_CPATH_HPP

#include <vector>

namespace mpc
{
                                    /**************************************************/
                                    /******************* Class CPath ******************/
                                    /**************************************************/

/** \class CPath
 *  \brief CPath followed by a system
 */
class CPath : public Path
{
protected : 
  std::vector<mpc::Config*> listCfg_;                   /**< list of the configurations describing the path */

  /********** Constructors and destructors **********/

protected : 
  /**
   *  \brief default constructor
   */
  CPath();

public : 
  /**
   *  \brief destructor
   */
  virtual ~CPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief getter : size of the list of configurations
   *  \return the size of the list of configuratons
   */
  inline int size() const { return this->listCfg_.size(); }

  /**
   *  \brief compute the length of the path
   *  \return the length of the path
   */
  double getLength() const;

  /**
   *  \brief compute the length of the path up to the n-th node
   *  \param n : the input index of node
   *  \return the length of the path up to the n-th node
   */
  double getLengthUpTo(int n) const;

  /**
   *  \brief compute the length of the path between the a-th and the b-th node
   *  \param a : the first index of node
   *  \param b : the second index of node
   *  \return the length of the path between the a-th and the b-th node
   */
  double getLengthBetween(int a,int b) const;

  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual const mpc::Config& at(int i) const = 0;

  /**
   *  \brief return true if the path is empty
   *  \return true if the path is empty
   */
  bool empty() const;

  
  /******** Configurations list modification ********/

public : 
//   /**
//    *  \brief append a new configuration to the path
//    *  \param c : the configuration to append
//    */
//   virtual void append(ConfigR* c);

//   /**
//    *  \brief prepend a new configuration to the path
//    *  \param c : the configuration to prepend
//    */
//   virtual void prepend(ConfigR*);

//   /**
//    *  \brief concatenate a path to the current one
//    *  \param p : the input path to concatenate
//    */
//   virtual void concat(const Path& p);

  /**
   *  \brief reverse the current path
   */
  void reverse();

  /**
   *  \brief clear the list of configurations
   */
  void clear();
  
  /******************** Operators *******************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual mpc::Config& operator[](int i) = 0;
  
  /**************** Utility functions ***************/

public : 
  /**
   *  \brief diplay the path
   */
  virtual void display() const;

//   /**
//    *  \brief return true if the path is free on the map
//    *  \param map : the input map
//    *  \return true if the path is free on the map
//    */
//   bool isFreeOn(Map *map) const;

//   /**
//    *  \brief trace the current path on a map
//    */
//   void traceOn(Map *map)const;

  /**
   *  \brief display in an openGL viewer
   */
  void toGL() const;
  
  /**
   *  \brief return the barycenter of the path in the workspace
   */
  virtual mpc::PositionXY center() const;

  /**
   *  \brief compute an approximative radius of the set of all configurations of the path
   *  \return an approximative radius of the set of all configurations of the path
   */
  virtual double radius() const;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic) = 0;
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const = 0;
};

};
#endif // MPC_CPATH_HPP


#ifndef MPC_PMPATH_HPP
#define MPC_PMPATH_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class PMPath ******************/
                                    /**************************************************/

/** \class PMPath
 *  \brief Path followed by a point mass system
 */
class PMPath : public CPath
{
  /********** Constructors and destructors **********/

public : 
  /**
   *  \brief default constructor
   */
  PMPath();

  /**
   *  \brief copy constructor
   *  \param t : the path to copy
   */
  PMPath(const PMPath& p);

  /**
   *  \brief destructor
   */
  virtual ~PMPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual const mpc::PMConfig& at(int i) const;

  /**
   *  \brief getter : middle configuration of the path
   *  \return the middle configuration of the path
   */
  virtual const mpc::PMConfig& middleCfg() const;

  /**
   *  \brief getter : first configuration of the path
   *  \return the first configuration of the path
   */
  virtual const mpc::PMConfig& firstCfg() const;

  /**
   *  \brief getter : last configuration of the path
   *  \return the last configuration of the path
   */
  virtual const mpc::PMConfig& lastCfg() const;

  /******** Configurations list modification ********/

public : 
  /**
   *  \brief append a new configuration to the path
   *  \param c : the configuration to append
   */
  void append(const mpc::PMConfig& c);

  /**
   *  \brief prepend a new configuration to the path
   *  \param c : the configuration to prepend
   */
  void prepend(const mpc::PMConfig& c);

  /**
   *  \brief concatenate a path to the current one
   *  \param p : the input path to concatenate
   */
  void concat(const PMPath& p);

  /******************** Operators *******************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual mpc::PMConfig& operator[](int i);

  /**************** Utility functions ***************/

public : 
//   /**
//    *  \brief diplay the path
//    */
//   virtual void display() const;

//   /*
//    *  \brief search for the closest point of the path from the input position
//    *  \param (x,y) : the input position
//    *  \param dist : the resulting minimal distance
//    *  \param closestCfg :  the closest configuration of the path
//    */
//   void closestCfgFrom(double x, double y, double& dist, PMConfig& cfg);

  /*
   *  \brief search for the minimal distance from the path to the input position
   *  \param (x,y) : the input position
   *  \return the resulting minimal distance
   */
  double minDistFrom(double x, double y);

  /*
   *  \brief get back the index of the input coord in a grid
   *  \param (nbCols, nbLines) : number of columns and lines of the image
   *  \param (xOrigin, yOrigin) : coordinates of the top left corner of the image
   *  \param (xScale, yScale) : scales of the image (length between two cells of the image)
   *  \param (x,y) : input coordinates (in the world)
   *  \param (i,j) : output coordinates in the image (column, line)
   */
  static void getIndexPos(int nbCols, int nbLines, float xOrigin, float yOrigin, float xScale, float yScale, float x,float y,int& i,int& j);

  /**
   *  \brief draw the path on a rgb image
   *  \param image : the input image
   *  \param (xOrigin, yOrigin) : coordinates of the top left corner of the image
   *  \param (xScale, yScale) : scales of the image (length between two cells of the image)
   */
  void drawOnRGBImage(RGBImage& image, float xOrigin, float yOrigin, float xScale, float yScale, const RGBColor& color = RGBColor(0.5,0.5,0.5)) const;

  /**
   *  \brief get the boundaries of the path
   *  \param minX : the minimal X coordinate
   *  \param maxX : the maximal X coordinate
   *  \param minY : the minimal Y coordinate
   *  \param maxY : the maximal Y coordinate
   *  \return 0 if the process succeeded, -1 otherwise (if the path is empty)
   */
  int getBounds(float& minX, float& maxX, float& minY, float& maxY) const;

  /************** Path transformations **************/

public : 
  /**
   *  \brief shift the path coordinates on the plane
   *  \param dx : shift on the X axis
   *  \param dy : shift on the Y axis
   */
  void shiftXY(double dx, double dy);

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic);
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const;
};

};
#endif // MPC_PMPATH_HPP


#ifndef MPC_DPMPATH_HPP
#define MPC_DPMPATH_HPP

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class DPMPath ******************/
                                    /**************************************************/

/** \class DPMPath
 *  \brief Path followed by a directed point mass system
 */
class DPMPath : public PMPath
{
  /********** Constructors and destructors **********/

public : 
  /**
   *  \brief default constructor
   */
  DPMPath();

  /**
   *  \brief copy constructor
   *  \param t : the path to copy
   */
  DPMPath(const DPMPath& p);

  /**
   *  \brief destructor
   */
  virtual ~DPMPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual const mpc::DPMConfig& at(int i) const;

  /**
   *  \brief getter : middle configuration of the path
   *  \return the middle configuration of the path
   */
  virtual const mpc::DPMConfig& middleCfg() const;

  /**
   *  \brief getter : first configuration of the path
   *  \return the first configuration of the path
   */
  virtual const mpc::DPMConfig& firstCfg() const;

  /**
   *  \brief getter : last configuration of the path
   *  \return the last configuration of the path
   */
  virtual const mpc::DPMConfig& lastCfg() const;

  /******** Configurations list modification ********/

public : 
  /**
   *  \brief append a new configuration to the path
   *  \param c : the configuration to append
   */
  void append(const mpc::DPMConfig& c);

  /**
   *  \brief prepend a new configuration to the path
   *  \param c : the configuration to prepend
   */
  void prepend(const mpc::DPMConfig& c);

  /**
   *  \brief concatenate a path to the current one
   *  \param p : the input path to concatenate
   */
  void concat(const DPMPath& p);

  /**
   *  \brief return an inverted path from the current one
   *  \return an inverted path from the current one
   */
  DPMPath* invert() const;

  /******************** Operators *******************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual mpc::DPMConfig& operator[](int i);

  /**************** Utility functions ***************/

public : 
//   /**
//    *  \brief diplay the path
//    */
//   virtual void display() const;

  /************** Path transformations **************/

public : 
  /**
   *  \brief lateral shift of the path coordinates on the plane
   *  \param dist : shift distance
   */
  void shiftLat(double dist);

  /**
   *  \brief rotate the path around its origin
   *  \param ang : rotation angle
   */
  void rotate(double ang);

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic);
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const;

  /**
   *  \brief load the path (for a robufast car) from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool loadRobufastPath(char* nomfic);
};

};
#endif // MPC_DPMPATH_HPP


#ifndef MPC_XYTHS_PATH_HPP
#define MPC_XYTHS_PATH_HPP

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class XYThS_Path ****************/
                                    /**************************************************/

/** \class XYThS_Path
 *  \brief Path followed by a directed planar system
 */
class XYThS_Path : public DPMPath
{
  /********** Constructors and destructors **********/

public : 
  /**
   *  \brief default constructor
   */
  XYThS_Path();

  /**
   *  \brief copy constructor
   *  \param t : the path to copy
   */
  XYThS_Path(const XYThS_Path& p);

  /**
   *  \brief copy constructor (from a DPMPath)
   *  \param t : the path to copy
   */
  XYThS_Path(const DPMPath& p);

  /**
   *  \brief destructor
   */
  virtual ~XYThS_Path();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual const mpc::XYThS_Config& at(int i) const;

  /**
   *  \brief getter : middle configuration of the path
   *  \return the middle configuration of the path
   */
  virtual const mpc::XYThS_Config& middleCfg() const;

  /**
   *  \brief getter : first configuration of the path
   *  \return the first configuration of the path
   */
  virtual const mpc::XYThS_Config& firstCfg() const;

  /**
   *  \brief getter : last configuration of the path
   *  \return the last configuration of the path
   */
  virtual const mpc::XYThS_Config& lastCfg() const;

  /**
   *  \brief put in cfg the configuration of the path given at curve abscica s
   *  \param s : the input curve abscica
   *  \param cfg : the resulting configuration
   */
  void getCfgAtCAbs(double s, XYThS_Config& cfg) const;


  /******** Configurations list modification ********/

public : 
  /**
   *  \brief append a new configuration to the path
   *  \param c : the configuration to append
   */
  void append(const mpc::XYThS_Config& c);

  /**
   *  \brief prepend a new configuration to the path
   *  \param c : the configuration to prepend
   */
  void prepend(const mpc::XYThS_Config& c);

  /**
   *  \brief concatenate a path to the current one
   *  \param p : the input path to concatenate
   */
  void concat(const XYThS_Path& p);

  /******************** Operators *******************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual mpc::XYThS_Config& operator[](int i);

//   /************* Input/Output functions *************/

// public : 
//   /**
//    *  \brief load the path from an input file
//    *  \param nomfic : name of the input file
//    */
//   virtual bool load(char* nomfic);
  
//   /**
//    *  \brief save the path in an output file
//    *  \param nomfic : name of the output file
//    */
//   virtual bool save(char* nomfic) const;
};

};
#endif // MPC_XYTHS_PATH_HPP


#ifndef MPC_CLPATH_HPP
#define MPC_CLPATH_HPP

namespace mpc
{
                                    /**************************************************/
                                    /****************** Class CLPath ******************/
                                    /**************************************************/

/** \class CLPath
 *  \brief Path followed by a car-like system
 */
class CLPath : public CPath
{
  /********** Constructors and destructors **********/

public : 
  /**
   *  \brief default constructor
   */
  CLPath();

  /**
   *  \brief copy constructor
   *  \param t : the path to copy
   */
  CLPath(const CLPath& p);

  /**
   *  \brief destructor
   */
  virtual ~CLPath();

  /************** Getters and setters ***************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual const CLConfig& at(int i) const;

  /**
   *  \brief getter : middle configuration of the path
   *  \return the middle configuration of the path
   */
  virtual const CLConfig& middleCfg() const;

  /**
   *  \brief getter : first configuration of the path
   *  \return the first configuration of the path
   */
  virtual const CLConfig& firstCfg() const;

  /**
   *  \brief getter : last configuration of the path
   *  \return the last configuration of the path
   */
  virtual const CLConfig& lastCfg() const;

  /******** Configurations list modification ********/

public : 
  /**
   *  \brief append a new configuration to the path
   *  \param c : the configuration to append
   */
  void append(const CLConfig& c);

  /**
   *  \brief prepend a new configuration to the path
   *  \param c : the configuration to prepend
   */
  void prepend(const CLConfig& c);

  /**
   *  \brief concatenate a path to the current one
   *  \param p : the input path to concatenate
   */
  void concat(const CLPath& p);

  /******************** Operators *******************/

public : 
  /**
   *  \brief return the i-th element of the path
   *  \param i : index of the element to return
   *  \return the i-th element of the path
   */
  virtual CLConfig& operator[](int i);

  /**************** Utility functions ***************/

public : 
//   /**
//    *  \brief diplay the path
//    */
//   virtual void display() const;

  /************* Input/Output functions *************/

public : 
  /**
   *  \brief load the path from an input file
   *  \param nomfic : name of the input file
   */
  virtual bool load(char* nomfic);
  
  /**
   *  \brief save the path in an output file
   *  \param nomfic : name of the output file
   */
  virtual bool save(char* nomfic) const;
};

};
#endif // MPC_CLPATH_HPP


#ifndef MPC_DD2PATH_HPP
#define MPC_DD2PATH_HPP

#define DD2Path DPMPath

#endif // MPC_DD2PATH_HPP


#ifndef MPC_WCPATH_HPP
#define MPC_WCPATH_HPP

#define WCPath DD2Path

#endif // MPC_WCPATH_HPP


#ifndef MPC_MANAPATH_HPP
#define MPC_MANAPATH_HPP

#define ManaPath DD2Path

#endif // MPC_MANAPATH_HPP
