#ifndef MPC_SHIFTRECORD_HPP
#define MPC_SHIFTRECORD_HPP

#include "motion/inc/Shift.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class ShiftRecord ***************/
                                    /**************************************************/

  /** \class ShiftRecord
   *  \brief Shift recording
   */
  class ShiftRecord
  {
    /********** Constructors and destructors **********/

  public :   
    /**
     *  \brief default constructor
     */
    ShiftRecord();

    /**
     *  \brief destructor
     */
    virtual ~ShiftRecord();
  };
};

#endif // MPC_SHIFTRECORD_HPP


#ifndef MPC_DSHIFTRECORD_HPP
#define MPC_DSHIFTRECORD_HPP

#include <vector>

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class DShiftRecord **************/
                                    /**************************************************/

  /** \class DShiftRecord
   *  \brief Discrete shift recording
   */
  class DShiftRecord : public ShiftRecord
  {
  protected : 
    std::vector<Shift*> listShift_;                 /**< list of the discrete shifts */

    /********** Constructors and destructors **********/

  public :   
    /**
     *  \brief default constructor
     */
    DShiftRecord();

    /**
     *  \brief copy constructor
     *  \param t : the trajectory to copy
     */
    DShiftRecord(const DShiftRecord& r);

    /**
     *  \brief destructor
     */
    virtual ~DShiftRecord();

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : size of the list
     *  \return the size of the list
     */
    inline int size() const { return this->listShift_.size(); }

    /**
     *  \brief getter : middle value of the record
     *  \return the middle value of the record
     */
    virtual const Shift& middle() const = 0;

    /**
     *  \brief getter : first value of the record
     *  \return the first value of the record
     */
    virtual const Shift& first() const = 0;

    /**
     *  \brief getter : last value of the record
     *  \return the last value of the record
     */
    virtual const Shift& last() const = 0;

    /**
     *  \brief getter : i-th element of the record
     *  \return return the i-th element of the record
     */
    virtual const Shift& at(int i) const = 0;

    /**
     *  \brief return true if the record is empty
     *  \return true if the record is empty
     */
    bool empty() const;

    /**************** list modification ***************/

  public : 
//     /**
//      *  \brief shift the record by curve absica s
//      *  \param t : the curve abscica to shift the record
//      */
//     void shift(double s);
    /**
     *  \brief pop out the last element of the record
     */
    void pop_back();

    /**
     *  \brief remove the i-th element of the record
     *  \param i : index of the value to remove
     */
    void remove(int i);

    /**
     *  \brief remove the all elements of the record between start and end
     *  \param i : index of the value to remove
     *
     *  including start but not including end
     */
    void remove(int start,int end);

    /**
     *  \brief clear the content of the trajectory
     */
    void clear();

    /******************** Operators *******************/

  public : 
    /**
     *  \brief getter : i-th element of the record
     *  \return return the i-th element of the record
     */
    virtual Shift& operator[](int i) = 0;
  };
};

#endif // MPC_DSHIFTRECORD_HPP


#ifndef MPC_YTHS_DSHIFTRECORD_HPP
#define MPC_YTHS_DSHIFTRECORD_HPP

#include <vector>

namespace mpc
{
                                    /**************************************************/
                                    /************* Class YThS_DShiftRecord ************/
                                    /**************************************************/

  /** \class YThS_DShiftRecord
   *  \brief Discrete recording of the lateral and angular shift between two paths
   */
  class YThS_DShiftRecord : public DShiftRecord
  {
    /********** Constructors and destructors **********/

  public :   
    /**
     *  \brief default constructor
     */
    YThS_DShiftRecord();

    /**
     *  \brief copy constructor
     *  \param r : the record to copy
     */
    YThS_DShiftRecord(const YThS_DShiftRecord& r);

    /**
     *  \brief destructor
     */
    virtual ~YThS_DShiftRecord();

    /************** Getters and setters ***************/

  public : 
    /**
     *  \brief getter : middle value of the record
     *  \return the middle value of the record
     */
    virtual const YThSShift& middle() const;

    /**
     *  \brief getter : first value of the record
     *  \return the first value of the record
     */
    virtual const YThSShift& first() const;

    /**
     *  \brief getter : last value of the record
     *  \return the last value of the record
     */
    virtual const YThSShift& last() const;

    /**
     *  \brief getter : i-th element of the record
     *  \return return the i-th element of the record
     */
    virtual const YThSShift& at(int i) const;

    /**************** List modification ***************/

    /**
     *  \brief append a new value at the end of the record
     *  \param s : the new value to append
     */
    virtual void append(const YThSShift& s);

    /**
     *  \brief concatenate two trajectories
     *  \param t : the trajectory to concatenate to the current one
     */
    virtual void concat(const YThS_DShiftRecord& r);

    /**
     *  \brief insert the value s at index position i in the list
     *  \param i : index to insert the new value
     *  \param s : the new value to insert
     */
    virtual void insert(int i, const YThSShift& s);

    /******************** Operators *******************/

  public : 
    /**
     *  \brief getter : i-th element of the record
     *  \return return the i-th element of the record
     */
    virtual YThSShift& operator[](int i);
  };

};

#endif // MPC_YTHS_DSHIFTRECORD_HPP
