#ifndef MPC_CTRAJ_PLANNER
#define MPC_CTRAJ_PLANNER

#include "trajGen/inc/TrajGen.hpp"
#include "planner/inc/TrajPlanner.hpp"
#include "common/inc/ControlTime.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class CTrajPlanner ***************/
                                    /**************************************************/

  /** \class CTrajPlanner
   *  \brief Trajectory planner from an input control sequence
   */
  class CTrajPlanner : public TrajPlanner
  {
  protected : 
    TrajGen* trajgen_;                                   /**< Trajectory generator to compute the state
							    trajectory from the input sequence of controls*/
  

    /************* Constructors and destructors ************/
  
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    CTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e   : environment in which the system evolve
     */
    CTrajPlanner(System* sys,Environment* e);

    /**
     *  \brief destructor
     */
    virtual ~CTrajPlanner();
  };

};
#endif


#ifndef MPC_CCCTRAJ_PLANNER
#define MPC_CCCTRAJ_PLANNER

#include "trajGen/inc/CLTrajGen.hpp"

namespace mpc
{
  /**************************************************/
  /************** Class CCCTrajPlanner **************/
  /**************************************************/

  /** \class CCCTrajPlanner
   *  \brief Trajectory planner for a cycab system from an input control sequence
   */
  class CCCTrajPlanner : public CTrajPlanner
  {
    /************* Constructors and destructors ************/
  
  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    CCCTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e   : environment in which the system evolve
     */
    CCCTrajPlanner(System* sys,Environment* e);

    /**
     *  \brief destructor
     */
    virtual ~CCCTrajPlanner();

    /********************** Accessors **********************/

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const CLState& initS() const { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const State& goalS() const { return *((CLState*)(this->goalS_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    inline virtual State& initS() { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    inline virtual State& goalS() { return *((CLState*)(this->goalS_)); }

    /**
     *  \brief return the trajectory generator used
     *  \return the trajectory generator used
     */
    inline CLTrajGen* trajgen() const { return ((CLTrajGen*)(this->trajgen_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual CLSTraj* traj() { return ((CLSTraj*)(this->traj_)); }

  public : 
    /**
     *  \brief return the robotic system's constraints
     *  \return the robotic system's constraints
     */
    inline virtual const CarLikeSys* sys() const { return ((CarLikeSys*)(this->sys_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual const CLSTraj* getTraj() const;

    /**************** Trajectory Computation ***************/

  protected : 
    /**
     *  \brief compute the state reach from an initial state and an input control
     *  \param s1 : the initial state
     *  \param ctrl : the input control time
     */
    CLStateT computeStateReachedFrom(const CLStateT& s1,const CCControlT& ctrl);

  public : 
    /**
     *  \brief compute the trajectory from a predefined input sequence of controls
     *  \param initS : the initial state of the trajectory
     */
    void computeDefinedTraj(const CLState& initS);

    /**
     *  \brief compute the trajectory from a predefined input sequence of controls
     *  \param initS : the initial state of the trajectory
     */
    void computeDefinedTraj2(const CLState& initS);

    /**
     *  \brief compute the trajectory from a predefined input sequence of controls
     *  \param initS : the initial state of the trajectory
     */
    void computeDefinedTraj3(const CLState& initS);
  };

};
#endif // MPC_CCCTRAJ_PLANNER
