#ifndef MPC_PVDTRAJPLANNER_HPP
#define MPC_PVDTRAJPLANNER_HPP

#include "planner/inc/TrajPlanner.hpp"
#include "system/inc/System.hpp"

#include "planner/inc/PFPathPlanner.hpp"

#include "system/inc/PointMassSys.hpp"
#include "system/inc/CarLikeSys.hpp"
#include "system/inc/WheelChairSys.hpp"


/**
 * \enum pvd_sampling_t
 * \brief Enumeration of the different path types
 *
 * PVD_TIME_SPG    : PVD by sampling over the time
 * PVD_SPACE_SPG   : PVD by sampling over the space
 */
typedef enum {
  PVD_TIME_SPG   =  0,
  PVD_SPACE_SPG  =  1,
} pvd_sampling_t;

namespace mpc
{
                                    /**************************************************/
                                    /************** Class PVDTrajPlanner **************/
                                    /**************************************************/

  /** \class PVDTrajPlanner
   *  \brief Trajectory planner by path-velocity decomposition
   */
  class PVDTrajPlanner : public TrajPlanner
  {
  protected : 
    Path* path_;                                   /**< the path followed by the robot */

    double timeStep_;                              /**< the time step used to compute the trajectory */
    static double maxSpeedCoef_;                   /**< maximal speed coefficient */
    static double maxAccCoef_;                     /**< maximal acceleration coefficient */

    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    PVDTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     */
    PVDTrajPlanner(System* sys, Environment* e = NULL);

    /**
     *  \brief destructor
     */
    virtual ~PVDTrajPlanner();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual Path* path() const { return this->path_; }

    /**
     *  \brief return the maximum speed coefficient
     *  \return the maximum speed coefficient
     */
    inline static double getMSC() { return PVDTrajPlanner::maxSpeedCoef_; }

    /**
     *  \brief set the maximum speed coefficient
     *  \param msc : the new maximum speed coefficient
     */
    static void setMSC(double msc);

    /**
     *  \brief return the maximum acceleration coefficient
     *  \return the maximum acceleration coefficient
     */
    inline static double getMAC() { return PVDTrajPlanner::maxAccCoef_; }

    /**
     *  \brief set the maximum acceleration coefficient
     *  \param mac : the new maximum acceleration coefficient
     */
    static void setMAC(double mac);

    /**
     *  \brief return the time step value
     *  \return the time step value
     */
    inline double getTimeStep() const { return this->timeStep_; }

    /**
     *  \brief set the time step value
     *  \param ts : the new time step value
     */
    inline void setTimeStep(double ts) { this->timeStep_ = ts; }

  protected : 
    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    virtual System* sys() const = 0;
  };

};
#endif // MPC_PVDTRAJPLANNER_HPP


#ifndef MPC_PMPVDTRAJPLANNER_HPP
#define MPC_PMPVDTRAJPLANNER_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************* Class PM_PVDTrajPlanner ************/
                                    /**************************************************/

  /** \class PM_PVDTrajPlanner
   *  \brief Trajectory planner by path-velocity decomposition for a point mass system
   */
  class PM_PVDTrajPlanner : public PVDTrajPlanner
  {
    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    PM_PVDTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     */
    PM_PVDTrajPlanner(PointMassSys* sys);

    /**
     *  \brief destructor
     */
    virtual ~PM_PVDTrajPlanner();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const PMPath* getPath() const { return ((PMPath*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return a copy of the planned trajectory
     *
     *  A new instance of PMSTraj is returned
     */
    virtual PMSTraj* getTraj() const;

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const PMState& initS() const { return *((PMState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const PMState& goalS() const { return *((PMState*)(this->goalS_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual PMState& initS() { return *((PMState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual PMState& goalS() { return *((PMState*)(this->goalS_)); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual PMPath* path() const { return ((PMPath*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual PMSTraj* traj() { return ((PMSTraj*)(this->traj_)); }

    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    inline virtual PointMassSys* sys() const { return ((PointMassSys*)(this->sys_)); }


    /**************** Trajectory Computing *****************/

  public : 
    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \param p  : the precomputed path bewteen s1 and s2
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTraj(const PMState& s1,const PMState& s2,const PMPath* p);
  };

};
#endif // MPC_PMPVDTRAJPLANNER_HPP


#ifndef MPC_DD2PVDTRAJPLANNER_HPP
#define MPC_DD2PVDTRAJPLANNER_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************* Class DD2_PVDTrajPlanner ***********/
                                    /**************************************************/

  /** \class DD2_PVDTrajPlanner
   *  \brief Trajectory planner by path-velocity decomposition for a differential drive
   */
  class DD2_PVDTrajPlanner : public PVDTrajPlanner
  {
  protected : 
    double maxVel_;                                        /**< maximal available velocity along the path */
    double maxAcc_;                                        /**< maximal available acceleration along the path */

  public : 
    static pvd_sampling_t spg_type_;                       /**< sampling type */

    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    DD2_PVDTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the robotic system constraints
     */
    DD2_PVDTrajPlanner(DiffDrive2Sys* sys);

    /**
     *  \brief destructor
     */
    virtual ~DD2_PVDTrajPlanner();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const DD2Path* getPath() const { return ((DD2Path*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return a copy of the planned trajectory
     *
     *  A new instance of DD2STraj is returned
     */
    virtual DD2STraj* getTraj() const;

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const DD2State& initS() const { return *((DD2State*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const DD2State& goalS() const { return *((DD2State*)(this->goalS_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual DD2State& initS() { return *((DD2State*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual DD2State& goalS() { return *((DD2State*)(this->goalS_)); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual DD2Path* path() const { return ((DD2Path*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual DD2STraj* traj() { return ((DD2STraj*)(this->traj_)); }

    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    inline virtual DiffDrive2Sys* sys() const { return ((DiffDrive2Sys*)(this->sys_)); }

    /**************** Trajectory Computing *****************/

  protected : 
    /*
     *  \brief 
     */
    void computeItdTime(double initSpd, double a1, double goalSpd, double pathLength, double& tmid, double& ttot);

    /**
     *  \brief compute the trajectory from s0 to sf
     *  \param s0 : the initial state
     *  \param sf : the goal state
     *  \param p  : the precomputed path bewteen s1 and s2
     *  \param cruiseVel : the cruise velocity (that we try to keep all along the trajectory)
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    int computeTraj(const DD2State& s0, const DD2State& sf, const DPMPath* p, float cruiseVel);

  public : 
    /**
     *  \brief compute the trajectory from an input path
     *  \param p  : the precomputed path
     *  \param cruiseVel : the cruise velocity (that we try to keep all along the trajectory)
     *  \param startVel : the starting velocity
     *  \param endVel : the ending velocity
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTraj(const DPMPath* p, float cruiseVel=0, float startVel=0, float endVel=0);

    /*
     *  \brief replan a part of the input state trajectory with a new maximal velocity
     *  \param straj : the input state trajectory
     *  \param maxVel : the new maximal velocity
     *  \param startTime : the starting time from which the trajectory must be modified
     *  \param startVel : the velocity registered at the starting time
     *  \param endTime : the ending time until which the trajectory must be modified
     *  \return the trajectory planner status at the end of the planning process
     *            ERR_ROBOT            : wrong robot features
     *            PARAMS_NOT_VALID     : not valid input parameters
     *            TRAJ_CPTD            : the trajectory has been computed
     */
    int replanTraj(DD2STraj* straj, float maxVel, float startTime, float startVel, float endTime);
  };

};
#endif // MPC_DD2PVDTRAJPLANNER_HPP


#ifndef MPC_CLPVDTRAJPLANNER_HPP
#define MPC_CLPVDTRAJPLANNER_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************* Class CL_PVDTrajPlanner ************/
                                    /**************************************************/

  /** \class CL_PVDTrajPlanner
   *  \brief Trajectory planner by path-velocity decomposition for a car-like system
   */
  class CL_PVDTrajPlanner : public PVDTrajPlanner
  {
    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    CL_PVDTrajPlanner();

  public : 
    /**
     *  \brief init constructor
     */
    CL_PVDTrajPlanner(CarLikeSys* sys);

    /**
     *  \brief destructor
     */
    virtual ~CL_PVDTrajPlanner();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual const CLPath* getPath() const { return ((CLPath*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return a copy of the planned trajectory
     *
     *  A new instance of CLSTraj is returned
     */
    virtual CLSTraj* getTraj() const;

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const CLState& initS() const { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const CLState& goalS() const { return *((CLState*)(this->goalS_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual CLState& initS() { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual CLState& goalS() { return *((CLState*)(this->goalS_)); }

    /**
     *  \brief return the current path
     *  \return the current path
     */
    inline virtual CLPath* path() const { return ((CLPath*)(this->path_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual CLSTraj* traj() { return ((CLSTraj*)(this->traj_)); }

    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    inline virtual CarLikeSys* sys() const { return ((CarLikeSys*)(this->sys_)); }

    /**************** Trajectory Computing *****************/

  public : 
    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \param p  : the precomputed path bewteen s1 and s2
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTraj(const CLState& s1,const CLState& s2,const CLPath* p);
  };

};
#endif // MPC_CLPVDTRAJPLANNER_HPP
