#ifndef MPC_ROADMAPTP_HPP
#define MPC_ROADMAPTP_HPP

#include "planner/inc/TrajPlanner.hpp"
#include "planner/inc/RoadMapPP.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class RoadMapTP ****************/
                                    /**************************************************/

  /** \class RoadMapTP
   *  \brief Trajectory planner using a roadmap
   */
  class RoadMapTP : public TrajPlanner
  {
  protected : 
    RoadMapPP* rmpp_;                              /**< the roadmap path planner used to compute an initial path */

    double timeStep_;                              /**< the time step used to compute the trajectory */
    static double maxSpeedCoef_;                   /**< maximal speed coefficient */
    static double maxAccCoef_;                     /**< maximal acceleration coefficient */

    static double minDObsCoef_;                    /**< coefficient between the minimal distance to obstacles 
						      and the length of the trajectories computed*/

  public : 
    static double minLengthTraj_;                  /**< minimal length of the trajectories computed */
    static double maxLengthTraj_;                  /**< maximal length of the trajectories computed */


    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     */
    RoadMapTP();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     */
    RoadMapTP(System* sys,Environment* e);

    //   /**
    //    *  \brief init constructor
    //    *  \param sys : robotic system handled
    //    *  \param rm : roadmap defined over the environment
    //    */
    //   RoadMapTP(System* sys,RoadMap* rm);

    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     *  \param rm : roadmap defined over the environment
     */
    RoadMapTP(System* sys,Environment* e,RoadMap* rm);

    /**
     *  \brief destructor
     */
    virtual ~RoadMapTP();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the maximum speed coefficient
     *  \return the maximum speed coefficient
     */
    inline static double getMSC() { return RoadMapTP::maxSpeedCoef_; }

    /**
     *  \brief set the maximum speed coefficient
     *  \param msc : the new maximum speed coefficient
     */
    static void setMSC(double msc);

    /**
     *  \brief return the maximum acceleration coefficient
     *  \return the maximum acceleration coefficient
     */
    inline static double getMAC() { return RoadMapTP::maxAccCoef_; }

    /**
     *  \brief set the maximum acceleration coefficient
     *  \param mac : the new maximum acceleration coefficient
     */
    static void setMAC(double mac);

    /**
     *  \brief return the time step value
     *  \return the time step value
     */
    inline double getTimeStep() const { return this->timeStep_; }

    /**
     *  \brief set the time step value
     *  \param ts : the new time step value
     */
    inline void setTimeStep(double ts) { this->timeStep_ = ts; }

    //   /**
    //    *  \brief return the initial state
    //    *  \return the initial state
    //    */
    //   virtual const State& initS() const;

    //   /**
    //    *  \brief return the initial state
    //    *  \return the initial state
    //    */
    //   virtual const State& goalS() const;

    // protected : 
    //   /**
    //    *  \brief return the initial state
    //    *  \return the initial state
    //    */
    //   virtual State& initS();

    //   /**
    //    *  \brief return the initial state
    //    *  \return the initial state
    //    */
    //   virtual State& goalS();

    /******************** Initialization *******************/

  public : 
    /**
     *  \brief reinit the planner and the roadmap if needed
     */
    void reinit();
  };

};
#endif // MPC_ROADMAPTP_HPP


#ifndef MPC_CLROADMAPTP_HPP
#define MPC_CLROADMAPTP_HPP

#include "system/inc/CarLikeSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class CLRoadMapTP ***************/
                                    /**************************************************/

  /** \class CLRoadMapTP
   *  \brief Trajectory planner for a car-like system using a roadmap
   */
  class CLRoadMapTP : public RoadMapTP
  {
    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     */
    CLRoadMapTP();

    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     */
    CLRoadMapTP(CarLikeSys* sys,Environment* e);

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     *  \param rm : roadmap defined over the environment
     */
    CLRoadMapTP(CarLikeSys* sys,Environment* e,RoadMap* rm);

    /**
     *  \brief destructor
     */
    virtual ~CLRoadMapTP();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const CLState& initS() const { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const CLState& goalS() const { return *((CLState*)(this->goalS_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual const CLSTraj* getTraj() const;

    /**
     *  \brief return the robotic system's constraints
     *  \return the robotic system's constraints
     */
    inline virtual const CarLikeSys* sys() const { return ((CarLikeSys*)(this->sys_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual CLState& initS() { return *((CLState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual CLState& goalS() { return *((CLState*)(this->goalS_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual CLSTraj* traj() { return ((CLSTraj*)(this->traj_)); }


    /**************** Trajectory Computing *****************/

  public : 
    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTrajV2(const CLState& s1,const CLState& s2);

    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTraj(const CLState& s1,const CLState& s2);

  protected : 
    /**
     *  \brief select the target state-time to reach
     *  \param path : the path computed on the roadmap
     *  \param pathLength : the total length of the path
     *  \param vecMD : list of minimal distances to obstacles for all configurations of the path computed
     *  \param s2 : the goal state of the trajectory
     *  \param curST : the current state-time
     *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
     *  \param curLength : length of the path up to the current configuration
     *  \param targetST : the resulting target state-time
     *  \param indTargetCfg : index of the target configuration of the path
     *  \param targetLength : length of the path up to the target configuration
     */
    void selectTargetST(const PMPath& path,double pathLength,
			std::list<double>* vecMD,const CLState& s2,
			const CLStateT& curST,int indCurCfg,double curLength,
			CLStateT& targetST,int& indTargetCfg,double& targetLength);

    /**
     *  \brief try to generate a trajectory between two state-times
     *  \param s1 : the initial state-time
     *  \param s2 : the goal state-time
     *  \param sR : the final state-time reached
     *  \param trajRes : the resulting discrete trajectory
     */
    void genTraj(const CLStateT& s1,const CLStateT& s2,CLStateT& sR,CLSTraj& traj);
  };

};
#endif // MPC_CLROADMAPTP_HPP


#ifndef MPC_WCROADMAPTP_HPP
#define MPC_WCROADMAPTP_HPP

#include "system/inc/WheelChairSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /**************** Class WCRoadMapTP ***************/
                                    /**************************************************/

  /** \class WCRoadMapTP
   *  \brief Trajectory planner for a wheelchair system using a roadmap
   */
  class WCRoadMapTP : public RoadMapTP
  {
    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     */
    WCRoadMapTP();

    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     */
    WCRoadMapTP(WheelChairSys* sys,Environment* e);

  public : 
    //   /**
    //    *  \brief init constructor
    //    *  \param sys : robotic system handled
    //    *  \param rm : roadmap defined over the environment
    //    */
    //   WCRoadMapTP(WheelChairSys* sys,RoadMap* rm);

    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e : environment in which the system evolve
     *  \param rm : roadmap defined over the environment
     */
    WCRoadMapTP(WheelChairSys* sys,Environment* e,RoadMap* rm);

    /**
     *  \brief destructor
     */
    virtual ~WCRoadMapTP();

    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const WCState& initS() const { return *((WCState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const WCState& goalS() const { return *((WCState*)(this->goalS_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual const WCSTraj* getTraj() const;

    /**
     *  \brief return the robotic system's constraints
     *  \return the robotic system's constraints
     */
    inline virtual const WheelChairSys* sys() const { return ((WheelChairSys*)(this->sys_)); }

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual WCState& initS() { return *((WCState*)(this->initS_)); }

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual WCState& goalS() { return *((WCState*)(this->goalS_)); }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    inline virtual WCSTraj* traj() { return ((WCSTraj*)(this->traj_)); }


    /**************** Trajectory Computing *****************/

  public : 
    /**
     *  \brief compute the trajectory from s1 to s2 (old version without collision checking)
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTrajV2(const WCState& s1,const WCState& s2);

    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTraj(const WCState& s1,const WCState& s2);

    /**
     *  \brief compute the trajectory from s1 to s2
     *  \param s1 : the initial state
     *  \param s2 : the final state
     *  \return the trajectory planner status at the end of the planning process
     *            NO_PATH_LDD   : no path has been loaded
     *            ERR_ROBOT     : wrong robot features
     *            ERR_ENV       : wrong environment
     *            ERR_ENV_MAP   : wrong static obstacles map
     *            ERR_ENV_OBS   : wrong dynamic obstacles list
     *            ERR_GOAL_ST   : the goal configuration is not available or free 
     *            ERR_INIT_ST   : the init configuration is not available or free
     *            NO_TRAJ       : no path can be computed
     *            TRAJ_CPTD     : the trajectory has been computed
     */
    virtual int computeTrajV3(const WCState& s1,const WCState& s2);

  protected : 
    /**
     *  \brief select the target state-time to reach
     *  \param path : the path computed on the roadmap
     *  \param pathLength : the total length of the path
     *  \param vecMD : list of minimal distances to obstacles for all configurations of the path computed
     *  \param s2 : the goal state of the trajectory
     *  \param curST : the current state-time
     *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
     *  \param curLength : length of the path up to the current configuration
     *  \param targetST : the resulting target state-time
     *  \param indTargetCfg : index of the target configuration of the path
     *  \param targetLength : length of the path up to the target configuration
     */
    void selectTargetST(const PMPath& path,double pathLength,
			std::list<double>* vecMD,const WCState& s2,
			const WCStateT& curST,int indCurCfg,double curLength,
			WCStateT& targetST,int& indTargetCfg,double& targetLength);

    /**
     *  \brief compute the target state-time to reach from its position and the previous state-time
     *  \param path : the path computed on the roadmap
     *  \param curST : the current state-time
     *  \param indCurCfg : index of the current configuration of the path we suppose to have already reached
     *  \param indTargetCfg : index of the target configuration of the path
     *  \param s2 : the goal state of the trajectory
     *  \param targetST : the resulting target state-time
     */
    void computeTargetST(const PMPath& path,const WCStateT& curST,int indCurCfg,int indTargetCfg,const WCState& s2,WCStateT& targetST);

    /**
     *  \brief try to generate a trajectory between two state-times
     *  \param s1 : the initial state-time
     *  \param s2 : the goal state-time
     *  \param sR : the final state-time reached
     *  \param trajRes : the resulting discrete trajectory
     */
    void genTraj(const WCStateT& s1,const WCStateT& s2,WCStateT& sR,WCSTraj& traj);

    /**
     *  \brief adjust the time and speed of the trajectory once it has been computed
     *  \param sInit : the initial state
     *  \param sGoal : the final state
     *  \param pathLength : the length of the path
     *
     *  The Trajectory results in a three steps velocity profile
     *  - A first step of acceleration up to the maximal speed
     *  - A second step with constant velocity (this part may be empty)
     *  - A last part of deceleration down to the final speed
     */
    void adjustTimeSpeed(const WCState& sInit,const WCState& sGoal,double pathLength);

    /**
     *  \brief compute the elapsed time between two states during the acceleration phase
     *  \param s1 : the initial state-time
     *  \param s2 : the final state-time
     *  \param v1 : the initial speed
     *  \param amax : the maximal acceleration
     *  \param tcst : the remaining time before the constant speed phase
     *  \param vmax : the maximal speed
     *  \param finalDT : the resulting elapsed time
     *  \param finalV : the speed of the final state
     */
    static void elapsedTime_AccPhase(const WCStateT& s1,const WCStateT& s2,double v1,double amax,double tcst,double vmax,double& finalDT,double& finalV);

    /**
     *  \brief compute the elapsed time between two states during the constant speed phase
     *  \param s1 : the initial state-time
     *  \param s2 : the final state-time
     *  \param vmax : the maximal speed
     *  \param tdec : the remaining time before the deceleration phase
     *  \param amax : the maximal acceleration
     *  \param finalDT : the resulting elapsed time
     *  \param finalV : the speed of the final state
     */
    static void elapsedTime_CstPhase(const WCStateT& s1,const WCStateT& s2,double vmax,double tdec,double amax,double& finalDT,double& finalV);

    /**
     *  \brief compute the elapsed time between two states during the acceleration phase
     *  \param s1 : the initial state-time
     *  \param s2 : the final state-time
     *  \param v1 : the initial speed
     *  \param amax : the maximal acceleration
     *  \param finalDT : the resulting elapsed time
     *  \param finalV : the speed of the final state
     */
    static void elapsedTime_DecPhase(const WCStateT& s1,const WCStateT& s2,double v1,double amax,double& finalDT,double& finalV);

    /**
     *  \brief resample the trajectory
     */
    void resampleTraj();
  };

};
#endif // MPC_WCROADMAPTP_HPP
