#ifndef MPC_TRAJPLANNER_HPP
#define MPC_TRAJPLANNER_HPP

#include "motion/inc/STrajectory.hpp"
#include "environment/inc/Environment.hpp"
#include "planner/inc/PathPlanner.hpp"
#include "common/inc/StateTime.hpp"

namespace mpc
{
  /**
   * \enum res_tp_t
   * \brief Enumeration of the different trajectory planner status
   *
   * NO_PATH_LDD   : no path has been loaded
   * ERR_ROBOT     : wrong robot features
   * ERR_ENV       : wrong environment
   * ERR_ENV_MAP   : wrong static obstacles map
   * ERR_ENV_OBS   : wrong dynamic obstacles list
   * ERR_GOAL_ST   : the goal configuration is not available or free 
   * ERR_INIT_ST   : the init configuration is not available or free
   * NO_TRAJ       : no path can be computed
   * TRAJ_CPTD     : the trajectory has been computed
   */
  typedef enum {
    PARAMS_NOT_VALID = -10,
    NO_INPUT_TRAJ    = -9,
    PF_NOT_STARTED   = -8,
    NO_PATH_LDD      = -7,
    ERR_ROBOT        = -6,
    ERR_ENV          = -5,
    ERR_ENV_MAP      = -4,
    ERR_ENV_OBS      = -3,
    ERR_GOAL_ST      = -2,
    ERR_INIT_ST      = -1,
    NO_TRAJ          =  0,
    TRAJ_CPTD        =  1
  } res_tp_t;

  /**
   * \enum res_tg_t
   * \brief Enumeration of the different trajectory generation return types
   *
   * CURV_OVR      : the curvature has been overreached
   * DCURV_OVR     : the curvature derivative has been overreached
   * NO_TRAJ_FEA   : the trajectory computed is not feasable
   * TRAJ_FEA      : the trajectory computed is feasable 
   */
  typedef enum {
    CURV_OVR      = -2,
    DCURV_OVR     = -1,
    NO_TRAJ_FEA   =  0,
    TRAJ_FEA      =  1
  } res_tg_t;

  /**
   * \enum trajPlannerType_t
   * \brief Enumeration of the different controller types
   *
   * CTRLER_NULL : controller not allocated
   * CTRLER_DISCONNECTED : disconnected controller
   * CTRLER_STOPPED : stopped controller
   * CTRLER_STARTED : started controller
   */
  typedef enum {
    ND_TP          = -1,
    SIMPLE_TP      =  0,
    PVD_TP         =  1,
    ROADMAP_TP     =  2
  } trajPlanner_t;

                                    /**************************************************/
                                    /**************** Class TrajPlanner ***************/
                                    /**************************************************/

  /** \class TrajPlanner
   *  \brief Trajectory planner
   */
  class TrajPlanner
  {
  protected : 
    STraj *traj_;                                      /**< the trajectory which will be computed*/
    Environment *env_;                                       /**< the environment in which the system evolve */
    System* sys_;                                            /**< the robotic system handled */
    State* initS_;                                           /**< the initial state */
    State* goalS_;                                           /**< the goal state */
    res_tp_t status_;                                        /**< trajectory planner status */
    trajPlanner_t type_;                                     /**< trajectory planner type */
  
    /************* Constructors and destructors ************/

  protected : 
    /**
     *  \brief default constructor
     *
     *  Only used by subclasses
     */
    TrajPlanner();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : robotic system handled
     *  \param e   : environment in which the system evolve
     */
    TrajPlanner(System* sys,Environment* e);

    /**
     *  \brief destructor
     */
    virtual ~TrajPlanner();
  
    /********************** Accessors **********************/

  public : 
    /**
     *  \brief return the type of the trajectory planner
     *  \return the type of the trajectory planner
     */
    inline virtual trajPlanner_t type() const { return this->type_; }

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual const STraj* getTraj() const;

    /**
     *  \brief return the current environment
     *  \return the current environment
     */
    inline const Environment* env() const { return this->env_; }

    /**
     *  \brief return the current map
     *  \return the current map
     */
    inline const Map* getMap() const { return this->env()->map(); }

    /**
     *  \brief return the robotic system's constraints
     *  \return the robotic system's constraints
     */
    inline virtual const System* sys() const { return this->sys_; }
  
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const State& initS() const = 0;

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual const State& goalS() const = 0;

  protected : 
    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual State& initS() = 0;

    /**
     *  \brief return the initial state
     *  \return the initial state
     */
    virtual State& goalS() = 0;

    /**
     *  \brief return the current trajectory
     *  \return the current trajectory
     */
    virtual STraj* traj();

    //   /******************* Traj Computing ********************/

    // public : 
    //   /**
    //    *  \brief compute the trajectory from s1 to s2
    //    *  \param s1 : the initial state
    //    *  \param s2 : the final state
    //    */
    //   virtual int computeTraj(const State& s1,const State& s2) = 0;
  };


  /***************** Collision Checking ******************/

  /**
   *  \brief check if a collision exists between a trajectory and a static environment
   *  \param traj : the input trajectory
   *  \param env : the input environment
   *  \return true if a collision exists between the trajectory and the static environment
   */
  bool staticCollision(const STraj& traj,const Environment* env);

};
#endif // MPC_TRAJPLANNER_HPP
