#ifndef SSLATTICE_HPP
#define SSLATTICE_HPP

#include "trajGen/inc/TrajGen.hpp"
#include "system/inc/System.hpp"
#include "common/inc/Tree.hpp"
#include "common/inc/StateTime.hpp"
#include "common/inc/ControlTime.hpp"
#include <vector>

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class SSLattice ****************/
                                    /**************************************************/

  /** \class SSLattice
   *  \brief State Space Lattice
   */
  class SSLattice
  {
  protected : 
    int height_;                                    /**< the height of the tree composing the lattice */
    int ctrlSize_;                                  /**< the size of the control vector */
    int nbCtrlSampling_;                            /**< the number of sampling for each control */

    System* sys_;                                   /**< the system for which the state space lattice is built */
    mpc::Node<StateTime*>* tree_;                   /**< the tree representing the lattice */

    std::vector<ControlTime*>* ctrlSp_;             /**< the control sampling */
    StateTime** tabLeaves_;                         /**< table of pointers to the leaves of the lattice tree */
    int J_;                                         /**< index used to fill the leaves tab */

    State* avError_;                                /**< the available error */

    bool displayTraj_;                              /**< true if the trajectories have to be stocked to be displayed */
    int trajSp_;                                    /**< sampling of the trajectories to display them */
    int I_;                                         /**< index used to fill the trajectory vectors */

    double curDT_;                                  /**< time interval between two nodes of the tree for the current lattice computation */
    int nbLeaves_;                                  /**< number of leaves of the current lattice built */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    SSLattice();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the system for which the state space lattice is built
     */
    SSLattice(System* sys);

    /**
     *  \brief destructor
     */
    virtual ~SSLattice();

    /************** Getters and Setters ***************/

  public : 
    /**
     *  \brief set the height of the tree composing the lattice
     *  \param h : the new height of the tree
     */
    inline void setHeight(int h) { this->height_ = h; }

    /**
     *  \brief set the number of sampling for each control
     *  \param nbCS : the new number of sampling for each control
     */
    inline void setNbCS(int nbCS) { this->nbCtrlSampling_ = nbCS; }

    /**
     *  \brief set the trajectory display status
     *  \param td : the new trajectory display status
     */
    inline void setDispTraj(bool td) { this->displayTraj_ = td; }

    /**
     *  \brief set the trajectory sampling
     *  \param ts : the new trajectory sampling
     */
    inline void setTrajSp(int ts) { this->trajSp_ = ts; }

    /**
     *  \brief return the number of leaves of the lattice
     *  \return the number of leaves of the lattice
     */
    inline int nbLeaves() const { return this->nbLeaves_; }

    /**
     *  \brief return the i-th leaf of the lattice tree
     *  \return the i-th leaf of the lattice tree
     */
    StateTime* leafAt(int i) const;

    /************** Lattice construction **************/

  protected : 
    /**
     *  \brief allocate a new state-time and copy the input one
     *  \param s : input state-time
     *  \return a pointer to the copy of the input state-time
     */
    virtual StateTime* copyStateT(const StateTime& s) = 0;

    /**
     *  \brief generate a trajectory to compute the state reached by s by ct
     *  \param st : the initial state-time of the trajectory
     *  \param ct : the control-time applied on the system
     *  \return the final state-time reached
     */
    virtual StateTime* genTraj(const StateTime* st,const ControlTime* ct) = 0;

    /**
     *  \brief sample the available controls
     *  \param ctrlSp : number of samples used for each control
     *  \param dt : the time of application of each control
     *  \return the control sampling vector
     */
    virtual void sampleControl(int nbCtrlSp,double dt) = 0;

    /**
     *  \brief recursive method that construct one node of the tree
     *  \param s : the initial state-time from which the current node is built
     *  \param dt : the time duration of a subtrajectory
     *  \param lvl : the remaining level to build up to the leaves
     *  \return the root of the branch built
     */
    mpc::Node<StateTime*>* constructNode(StateTime* s,double dt,int lvl);

  public : 
    /**
     *  \brief construct the state space lattice
     *  \param s : the initial state-time from which the state space lattice is built
     *  \param t : the time until which the lattice is built
     *  \return 0 if the state space lattice has been constructed successfully
     *  \return -1 if no input system has been defined
     */
    int construct(const StateTime& s,double t);

    /************** Feasibility functions *************/

  protected : 
    /**
     *  \brief compute the available error between two state-times
     *  \return the available error between two state-times
     */
    virtual StateTime* computeAvError(const StateTime* s1,const StateTime* s2) const = 0;

    /**
     *  \brief compute the difference between two state-times
     *  \return the difference between two state-times
     *
     *  A new instance of a state-time is computed, it must be deleted
     */
    virtual StateTime* diffStateT(const StateTime* s1,const StateTime* s2) const = 0;

  public : 
    /**
     *  \brief check if the input state-time is reachable according to the sslattice
     *  \param s : the input state to evaluate
     */
    virtual bool isReachable(const StateTime& s) const;

    /**************** Utility functions ***************/

  protected : 
    /**
     *  \brief delete the control sampling
     */
    void deleteCtrlSp();

    /**
     *  \brief delete the leaves tab
     */
    void deleteTabLeaves();

    /**
     *  \brief compute the number of trajectories that will be computed
     *  \return the number of trajectories that will be computed
     */
    int computeNbTraj() const;

    /**
     *  \brief compute the number of leaves of the tree
     *  \return the number of leaves of the tree
     */
    int computeNbLeaves() const;

    /**
     *  \allocate the trajectory vectors
     *  \param nb : the number of trajectories to allocate
     */
    virtual void allocTrajVec(int nb) = 0;

    /**************** Display functions ***************/

  public : 
    /**
     *  \brief display the list of controls computed
     */
    void displayCtrlSp() const;

    /**
     *  \brief draw the lattice in a gnuplot window
     *  \return 0 if the process succeed, -1 else
     */
    virtual int drawGP() const = 0;
  };

};
#endif // SSLATTICE_HPP
