#ifndef CARLIKESYS_HPP
#define CARLIKESYS_HPP

#include "system/inc/System.hpp"

namespace mpc
{

  /** \class CarLikeSys
   * \brief Car-like system class
   */
  class CarLikeSys : public System
  {
  protected : 
    double length_;                                   /**< length of the robot (between the front and rear wheels) */
    double radius_;                                   /**< radius of the cirle which include the robot */

    double vmax_;                                     /**< maximal speed */
    double amax_;                                     /**< maximal acceleration */

    double maxSteering_;                              /**< maximal steering */
    double maxSteerSpd_;                              /**< maximal steering speed */
    double maxCurvature_;                             /**< maximal curvature (duplicated information with maximal steering) */
    double maxCurvSpd_;                               /**< maximal curvature speed (duplicated information with maximal steering speed) */

    bool canMoveBack_;                                ///< true if the car-like can move backward
  
  public : 

  /********** Constructors and destructors **********/

  /**
   *  \brief default constructor
   */
    CarLikeSys();

  /**
   *  \brief init constructor
   *  \param cfg_filename : configuration filename
   */
    CarLikeSys(char* cfg_filename);

    /**
     *  \brief destructor
     */
    virtual ~CarLikeSys();

    /************** Getters and setters ***************/

    /**
     *  \brief getter : return the length of the robot
     *  \return the length of the robot
     */
    double getLength() const;
    /**
     *  \brief setter : set the length of the robot
     *  \param l : the new length of the robot
     */
    void setLength(const double& l);

    /**
     *  \brief getter : return the radius of the circle including the robot
     *  \return radius of the circle including the robot
     */
    double getRadius() const;
    /**
     *  \brief setter : set the radius of the circle including the robot
     *  \param r : radius of the circle including the robot
     */
    void setRadius(const double& r);
  
    /**
     *  \brief getter : return the maximal speed
     *  \return maximal speed
     */
    double getVMax() const;
    /**
     *  \brief setter : set the maximal speed
     *  \param v : new maximal speed
     */
    void setVMax(const double& v);
    /**
     *  \brief getter : return the maximal acceleration
     *  \return maximal acceleration
     */
    double getAMax() const;
    /**
     *  \brief setter : set the maximal acceleration
     *  \param v : new maximal acceleration
     */
    void setAMax(const double& a);

    /**
     *  \brief getter : return the maximal steering
     *  \return maximal steering
     */
    double getMaxSteering() const;
    /**
     *  \brief setter : set the maximal steering
     *  \param s : new maximal steering
     */
    void setMaxSteering(const double& s);
    /**
     *  \brief getter : return the maximal steering speed
     *  \return maximal steering speed
     */
    double getMaxSteerSpd() const;
    /**
     *  \brief setter : set the maximal steering speed
     *  \param ss : new maximal steering speed
     */
    void setMaxSteerSpd(const double& ss);

    /**
     *  \brief getter : return the maximal curvature
     *  \return maximal curvature
     */
    double getMaxCurvature() const;
    /**
     *  \brief setter : set the maximal curvature
     *  \param c : new maximal curvature
     */
    void setMaxCurvature(const double& c);
    /**
     *  \brief getter : return the maximal curvature speed
     *  \return maximal curvature speed
     */
    double getMaxCurvSpd() const;
    /**
     *  \brief setter : set the maximal curvature speed
     *  \param cs : new maximal curvature speed
     */
    void setMaxCurvSpd(const double& cs);

    /**
     *  \brief getter : return true if the wheelchair can move backward
     *  \return true if the wheelchair can move backward
     */
    inline bool canMoveBack() const { return this->canMoveBack_; }

    /**
     *  \brief getter : return true if the wheelchair can move backward
     *  \return true if the wheelchair can move backward
     */
    inline void setMoveBack(bool cmb) { this->canMoveBack_ = cmb; }

    /************* Initialization functions **************/

    /**
     *  \brief init the car-like system from a configuration file
     *  \param cfg_filename : the configuration filename
     */
    void loadFromCfgFile(char* cfg_filename);

    /*************** Display functions ****************/

    /**
     *  \brief display the features of the car-like system
     */
    virtual void display() const;

    /**
     *  \brief draw the car-like system in an openGL viewer
     */
    virtual void toGL() const;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const;
  };

};
#endif // CARLIKESYS_HPP
