#ifndef DIFFDRIVE2SYS_HPP
#define DIFFDRIVE2SYS_HPP

#include "system/inc/System.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class DiffDrive2Sys **************/
                                    /**************************************************/

  /** \class DiffDrive2Sys
   * \brief Differential drive system class
   */
  class DiffDrive2Sys : public System
  {
  protected :   
    double max_speed;                       ///< maximal linear speed (m/s)
    double max_acc;                         ///< maximal linear acceleration (m/s^2)
    double max_rotation_speed;              ///< maximal rotation speed (rad/s)
    double max_rotation_acc;                ///< maximal rotation acceleration (rad/s^2)
  
    double length;                          ///< lenght size (m)
    double width;                           ///< width size (m)

    bool canMoveBack_;                      ///< true if the robot can move backward
  
  public : 

  /********** Constructors and destructors **********/
    
  /**
   *  \brief default constructor
   */
    DiffDrive2Sys();

    /**
     *  \brief copy constructor
     *  \param sys : the wheelchair system constraints to copy
     */
    DiffDrive2Sys(const DiffDrive2Sys& sys);
    
    /**
     *  \brief init constructor
     *  \param wr : radius of the wheels (m)
     *  \param wb : distance between the 2 wheels
     *  \param v : maximal linear speed (m/s)
     *  \param a : maximal linear acceleration (m/s^2)
     *  \param av : maximal rotation speed (rad/s)
     *  \param aa : maximal rotation acceleration (rad/s^2)
     *  \param bd : braking distance from max speed (m)
     *  \param l : lenght size (m)
     *  \param w : width size (m)
     *  \param cmb : true if the wheelchair can move backward
     */
    DiffDrive2Sys(double v,double a,double av,double aa,double l,double w,bool cmb=false);

    /**
     *  \brief destructor
     */
    virtual ~DiffDrive2Sys();

    /************** Getters and setters ***************/

    /**
     *  \brief getter : return the length of the robot
     *  \return the length of the robot
     */
    inline double getLength() const { return this->length; }
  
    /**
     *  \brief setter : set the length of the robot
     *  \param l : the new length of the robot
     */
    inline void setLength(const double& l) { this->length = l; }

    /**
     *  \brief getter : return the width of the robot
     *  \return the width of the robot
     */
    inline double getWidth() const { return this->width; }
    /**
     *  \brief setter : set the width of the robot
     *  \param d : the new width of the robot
     */
    inline void setWidth(const double& w) { this->width = w; }
  
    /**
     *  \brief getter : return the maximal speed
     *  \return maximal speed
     */
    inline double getVMax() const { return this->max_speed; }

    /**
     *  \brief setter : set the maximal speed
     *  \param v : new maximal speed
     */
    inline void setVMax(const double& v) { this->max_speed = v; }

    /**
     *  \brief getter : return the maximal acceleration
     *  \return maximal acceleration
     */
    inline double getAMax() const { return this->max_acc; }

    /**
     *  \brief setter : set the maximal acceleration
     *  \param v : new maximal acceleration
     */
    inline void setAMax(const double& a) { this->max_acc = a; }

    /**
     *  \brief getter : return the maximal speed
     *  \return maximal speed
     */
    inline double getAVMax() const { return this->max_rotation_speed; }

    /**
     *  \brief setter : set the maximal angular speed
     *  \param v : new maximal angular speed
     */
    inline void setAVMax(const double& av) { this->max_rotation_speed = av; }

    /**
     *  \brief getter : return the maximal angular acceleration
     *  \return maximal angular acceleration
     */
    inline double getAAMax() const { return this->max_rotation_acc; }

    /**
     *  \brief setter : set the maximal angular acceleration
     *  \param v : new maximal angular acceleration
     */
    inline void setAAMax(const double& aa) { this->max_rotation_acc = aa; }

    /**
     *  \brief getter : return true if the wheelchair can move backward
     *  \return true if the wheelchair can move backward
     */
    inline bool canMoveBack() const { return this->canMoveBack_; }

    /**
     *  \brief getter : return true if the wheelchair can move backward
     *  \return true if the wheelchair can move backward
     */
    inline void setMoveBack(bool cmb) { this->canMoveBack_ = cmb; }

    /*************** Display functions ****************/

    /**
     *  \brief display the features of the differential drive system
     */
    virtual void display() const = 0;

    /**
     *  \brief convert the system into a string
     *  \return the resulting string
     */
    virtual const char* toString() const = 0;

    /**
     *  \brief draw the differential drive system in an openGL viewer
     */
    virtual void toGL() const = 0;
  };

};
#endif // DIFFDRIVE2SYS_HPP
