#ifndef SYSCOMPONENT_HPP
#define SYSCOMPONENT_HPP

#include "common/inc/Object.hpp"
#include "common/inc/ControlPoint.hpp"
//using namespace mpc;

#include <vector>

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class SysComponent ***************/
                                    /**************************************************/

  /** \class SysComponent
   *  \brief Component of an articulated system
   */
  class SysComponent : public Object
  {
  protected : 
    std::vector<ControlPoint2D> listCP_;                /**< list of the control points defined over the system component, 
							 *   the control points are defined wrt. the system component frame */

  public : 
    /********** Constructors and destructors **********/

    /**
     *  \brief default constructor
     */
    SysComponent();
    /**
     *  \brief copy constructor
     *  \param c : the control point to copy
     */
    SysComponent(const SysComponent& sc);
    /**
     *  \brief init constructor
     *  \param pos : position of the component in the workspace (W)
     *  \param bbox : bounding box defined overthe system
     */
    SysComponent(const PositionXYTh& pos,const BoundingShape& bshape);
    /**
     *  \brief destructor
     */
    virtual ~SysComponent();

    /************** Getters and setters ***************/
  
    /**
     *  \brief Position accessor
     *  \return the position of the object in its workspace (W)
     */
    virtual PositionXYTh& pos() const; //{ return *((PositionXYTh*)pos_); }

    /**
     *  \brief setter : set the current position of the system component
     *  \param p : the new position of the system component
     */
    virtual void setPos(const PositionXYTh& p); //{ this->pos_ = new PositionXYTh(p); }

    /**
     *  \brief number of control point getter
     *  \return the number of control points defined over the component
     */
    inline int getNbCP() const { return this->listCP_.size(); }
  
    /**
     *  \brief control point accessor
     *  \param i : index of the control point to return
     *  \return reference to the i-th control point
     */
    inline ControlPoint2D getCP(int i) const { return this->listCP_.at(i); }
  
    /**
     *  \brief add a new control point over the component
     *  \param cp : the new control point to add
     */
    inline void addCP(const ControlPoint2D& cp) { this->listCP_.push_back(cp); }  

    /**************** Motion functions ****************/
  
    /**
     *  \brief translate the system component
     *  \param p : vector of the translation
     */
    void translate(const PositionXY& p);
  
    /**
     *  \brief rotate the system component around p about an angle a
     *  \param p : center of the rotation
     *  \param a : angle of the rotation
     */
    void rotate(const PositionXY& p,const double& a);

    /*************** Display functions ****************/
  
    /**
     *  \brief display the system component's features
     */
    virtual void display() const;

    /**
     *  \brief draw the system component in an openGL viewer
     */
    virtual void toGL() const;
  };

};
#endif // SYSCOMPONENT_HPP


#ifndef SYSCOMPONENT3D_HPP
#define SYSCOMPONENT3D_HPP

namespace mpc
{
                                    /**************************************************/
                                    /************** Class SysComponent3D **************/
                                    /**************************************************/

  /** \class SysComponent3D
   *  \brief Component of an articulated system (in a 3D workspace)
   */
  class SysComponent3D : public SysComponent
  {
    // TODO
  };
  
};
#endif // SYSCOMPONENT3D_HPP
