#ifndef MPC_MANATRAJFOLLOW_HPP
#define MPC_MANATRAJFOLLOW_HPP

#include "trajFollow/inc/TrajFollow.hpp"
#include "system/inc/ManaSys.hpp"
#include "trajGen/inc/ManaTrajGen.hpp"

#include "motion/inc/PCTraj.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /************** Class ManaTrajFollow **************/
                                    /**************************************************/

  /** \class ManaTrajFollow
   *  \brief Trajectory following process for a Mana robotic system
   */
  class ManaTrajFollow : public TrajFollow
  {
    ManaStateT lastStateToReach_;                     /**< variable to stock the last state we tried to reach */

    ManaPCTraj* pctraj_;                              /**< polynomial control trajectory computed to join the followed trajectory */
    ManaPCTraj* old_pctraj_;                          /**< polynomial control trajectory computed at the previous step */
    bool trajConnected_;                              /**< true if the trajectory generator was able to find a trajectory that connects the followed trajectory */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Should not be used
     */
    ManaTrajFollow();
  
  public : 
    /**
     *  \brief init constructor
     *  \param traj : the trajectory to follow
     */
    ManaTrajFollow(ManaSys* sys,ManaSTraj* traj);
  
    /**
     *  \brief destructor
     */
    virtual ~ManaTrajFollow();
  
  
    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    inline virtual const ManaSys* sys() const { return ((ManaSys*)system_); }

    /**
     *  \brief return the followed state trajectory
     *  \return the followed state trajectory
     */
    inline virtual const ManaSTraj* traj() const { return ((ManaSTraj*)traj_); }

    /**
     *  \brief return the last state we tried to reach (with the trajectory generator)
     *  \return the last state we tried to reach
     */
    inline virtual const ManaStateT& lastStateToReach() const { return this->lastStateToReach_; }

    /**
     *  \brief return the last state-time of the followed trajectory
     *  \return the last state-time of the followed trajectory
     */
    ManaStateT goalStateT() const;

  protected : 
    /**
     *  \brief return the last state-time saved
     *  \return the last state-time saved
     */
    inline const ManaStateT* curRegST() const { return ((ManaStateT*)(this->curRegST_)); }

    /**
     *  \brief return the trajectory generator used
     *  \return the trajectory generator used
     */
    inline virtual ManaTrajGen* trajgen() const { return ((ManaTrajGen*)trajgen_); }


    /*************** Localization update **************/
  
  public : 
    /**
     *  \brief selection of the state of the trajectory to reach
     */
    virtual void selectStateToReach(const ManaStateT& curS,ManaStateT& goalS);

  protected : 
    /**
     *  \brief Update the localization of the robot
     *  \param curS : current state-time updated
     *  \return 0 if the trajectory generation successed, -1 else
     */
    virtual int updateLoc(const ManaStateT& curS);


    /***************** Control request ****************/

  protected :   
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,ManaControl& ctrl);

    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,double dt,ManaControl& ctrl);

  public : 
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param curS : current state-time updated
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     *  \return 0 if the trajectory generation succeeded, -1 else
     *
     *  Synchronous localization update and control request
     */
    int getControl(const ManaStateT& curS,double t,double dt,ManaControl& ctrl);

    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param curS : current state-time updated
     *  \param ctrl : resulting control
     *  \return 0 if the trajectory generation succeeded, -1 else
     *
     *  Synchronous localization update and control request
     */
    int getControl(const ManaState& curS,ManaControl& ctrl);

    /**
     *  \brief return a control trajectory to apply on the robot
     *  \param curS : current state-time updated
     *  \return resulting control trajectory
     *
     *  Synchronous localization update and control request
     */
    ManaPCTraj* getCTraj(const ManaState& curS);

    /**************** Trajectory update ***************/

    /*
     *  \brief update the trajectory wrt. the maximal allowed speed
     *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
     */
    virtual void updateTrajVelProfile(float coefMaxVel);

    /**************** Utility functions ***************/

    /*
     *  \brief compute the minimal breaking distance of the robot wrt. its current state
     *  \return the minimal breaking distance of the robot
     */
    virtual float minBreakDist() const;

    /*
     *  \brief compute the minimal breaking time needed by the robot wrt. its current state
     *  \return the minimal breaking time needed by the robot
     */
    virtual float minBreakTime() const;

    /*
     *  \brief return true if the goal has been reached
     *  \param curStateT : the current state-time of the robot
     *  \param goalStateT : the goal state-time to reach
     *  \return true if the goal has been reached
     */
    bool goalReached(const ManaStateT& curStateT, const ManaStateT& goalStateT) const;

    /**************** Display functions ***************/
  
  public : 
    /**
     *  \brief display the trajectory computed by the trajectory generator
     */
    virtual void displayTrajGen();
  };

};
#endif // MPC_MANATRAJFOLLOW_HPP
