#ifndef MPC_WCTRAJFOLLOW_HPP
#define MPC_WCTRAJFOLLOW_HPP

#include "trajFollow/inc/TrajFollow.hpp"
#include "system/inc/WheelChairSys.hpp"
#include "trajGen/inc/WCTrajGen.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /*************** Class WLTrajFollow ***************/
                                    /**************************************************/

  /** \class WCTrajFollow
   *  \brief Trajectory following process for a wheelchair system
   */
  class WCTrajFollow : public TrajFollow
  {
    WCStateT lastStateToReach_;                     /**< variable to stock the last state we tried to reach */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     *  Should not be used
     */
    WCTrajFollow();
  
  public : 
    /**
     *  \brief init constructor
     *  \param traj : the trajectory to follow
     */
    WCTrajFollow(WheelChairSys* sys,WCSTraj* traj);
  
    /**
     *  \brief destructor
     */
    virtual ~WCTrajFollow();
  
  
    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the system constraints
     *  \return the system constraints
     */
    inline virtual const WheelChairSys* sys() const { return ((WheelChairSys*)system_); }

    /**
     *  \brief return the followed state trajectory
     *  \return the followed state trajectory
     */
    inline virtual const WCSTraj* traj() const { return ((WCSTraj*)traj_); }

    /**
     *  \brief return the last state we tried to reach (with the trajectory generator)
     *  \return the last state we tried to reach
     */
    inline virtual const WCStateT& lastStateToReach() const { return this->lastStateToReach_; }

  protected : 
    /**
     *  \brief return the last state-time saved
     *  \return the last state-time saved
     */
    inline const WCStateT* curRegST() const { return ((WCStateT*)(this->curRegST_)); }

    /**
     *  \brief return the trajectory generator used
     *  \return the trajectory generator used
     */
    inline virtual WCTrajGen* trajgen() const { return ((WCTrajGen*)trajgen_); }


    /*************** Localization update **************/
  
  public : 
    /**
     *  \brief selection of the state of the trajectory to reach
     */
    virtual void selectStateToReach(const WCStateT& curS,WCStateT& goalS);

  protected : 
    /**
     *  \brief Update the localization of the robot
     *  \param curS : current state-time updated
     *  \return 0 if the trajectory generation successed, -1 else
     */
    virtual int updateLoc(const WCStateT& curS);


    /***************** Control request ****************/

  protected :   
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,WCControl& ctrl);

    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     */
    virtual void getControl(double t,double dt,WCControl& ctrl);

  public : 
    /**
     *  \brief Request for a new control to apply (at a specified time)
     *  \param curS : current state-time updated
     *  \param t : time at which is evaluated the control
     *  \param dt : duration of the application of the control
     *  \param ctrl : resulting control
     *  \return 0 if the trajectory generation successed, -1 else
     *
     *  Synchronous localization update and control request
     */
    int getControl(const WCStateT& curS,double t,double dt,WCControl& ctrl);


    /**************** Trajectory update ***************/

    /*
     *  \brief update the trajectory wrt. the maximal allowed speed
     *  \param coefMaxVel : coefficient limiting the maximal allowed speed (\in [0;1])
     */
    virtual void updateTrajVelProfile(float coefMaxVel);


    /**************** Utility functions ***************/

    /*
     *  \brief compute the minimal breaking distance of the robot wrt. its current state
     *  \return the minimal breaking distance of the robot
     */
    virtual float minBreakDist() const;

    /*
     *  \brief compute the minimal breaking time needed by the robot wrt. its current state
     *  \return the minimal breaking time needed by the robot
     */
    virtual float minBreakTime() const;

    /*
     *  \brief get back the last time of the trajectory
     *  \return the last time of the trajectory
     */
    double lastTimeTraj();


    /**************** Display functions ***************/
  
  public : 
    /**
     *  \brief display the trajectory computed by the trajectory generator
     */
    virtual void displayTrajGen();
  };

};
#endif // MPC_WCTRAJFOLLOW_HPP
