#ifndef MPC_CLTRAJGEN_HPP
#define MPC_CLTRAJGEN_HPP

#include "trajGen/inc/TrajGen.hpp"
#include "math/inc/MultiItvFunction.hpp"
#include "math/inc/nrtypes.h"

#include "environment/inc/Map.hpp"

#include "system/inc/CarLikeSys.hpp"

#include "trajGen/inc/CLInitialGuess.hpp"

namespace mpc
{
  class CLInitialGuess;

                                    /**************************************************/
                                    /***************** Class CLTrajGen ****************/
                                    /**************************************************/

  /** \class CLTrajGen
   *  \brief Trajectory generator TIJI for a car-like system
   */
  class CLTrajGen : public TrajGen
  {
  protected:  
    double maxAcc;                                     /**< maximal acceleration */
    double maxSpd;                                     /**< maximal linear velocity */
    double maxSt;                                      /**< maximal steering angle */
    double maxStSpd;                                   /**< maximal steering speed */

    bool USE_VEL_CST;                                  /**< true if the linear velocity constraints are used */
    bool USE_ACC_CST;                                  /**< true if the linear acceleration constraints are used */
    bool USE_AVEL_CST;                                 /**< true if the angular velocity constraints are used */
    bool USE_AACC_CST;                                 /**< true if the angular acceleration constraints are used */

    int lastItvAccBdd_;                                // last interval of the acceleration, linear
    int lastItvVelBdd_;                                // and angular velocities profiles bounded
    int lastItvAVelBdd_;                               // -1 = minB,  0 = not bounded,  1 = maxB
    int lastItvAAccBdd_;

    CLState CLweights_;                                /**< values used to weight the cost function */

    vector<vector<double>* > regX;
    vector<vector<double>* > regY;
    vector<vector<double>* > regTH;
    vector<vector<double>* > regV;
    vector<vector<double>* > regAV;
    vector<vector<double>* > regT;

    /*************** Getters and Setters **************/

  public :   
    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    virtual const CarLikeSys* sys() const { return ((CarLikeSys*)sys_); }

    /** 
     *  \brief getter : return true if linear velocity constraints are used
     *  \return true if linear velocity constraints are used
     */
    bool getUseVelCst();

    /**
     *  \brief setter : set true if linear velocity constraints are used
     *  \param u : set it to true to use the linear velocity constraints
     */
    void setUseVelCst(bool u);

    /** 
     *  \brief getter : return true if linear acceleration constraints are used
     *  \return true if linear acceleration constraints are used
     */
    bool getUseAccCst();

    /**
     *  \brief setter : set true if linear acceleration constraints are used
     *  \param u : set it to true to use the linear acceleration constraints
     */
    void setUseAccCst(bool u);

    /** 
     *  \brief getter : return true if angular velocity constraints are used
     *  \return true if angular velocity constraints are used
     */
    bool getUseAVelCst();

    /**
     *  \brief setter : set true if angular velocity constraints are used
     *  \param u : set it to true to use the angular velocity constraints
     */
    void setUseAVelCst(bool u);

    /** 
     *  \brief getter : return true if angular acceleration constraints are used
     *  \return true if angular acceleration constraints are used
     */
    bool getUseAAccCst();

    /**
     *  \brief setter : set true if angular acceleration constraints are used
     *  \param u : set it to true to use the angular acceleration constraints
     */
    void setUseAAccCst(bool u);

    /**
     *  \brief fix or free the final position
     *  \param fixed : true if the final position must be fixed
     */
    void fixFinalXY(bool fixed);

    /**
     *  \brief fix or free the final heading
     *  \param fixed : true if the final heading must be fixed
     */
    void fixFinalTH(bool fixed);

    /**
     *  \brief fix or free the final velocity
     *  \param fixed : true if the final velocity must be fixed
     */
    void fixFinalV(bool fixed);

    /**
     *  \brief fix or free the final steering angle
     *  \param fixed : true if the final steering angle must be fixed
     */
    void fixFinalPH(bool fixed);

    /**
     *  \brief set the car-like state weights for the cost function
     *  \param wXY : position weight
     *  \param wTH : heading weight
     *  \param wV  : speed weight
     *  \param wPH : steering weight
     */
    void setCLweights(double wXY,double wTH,double wV,double wPH);
  
  protected : 
    MultiItvFunction* fctAcc_;                         /**< linear acceleration profile */
    MultiItvFunction* fctVel_;                         /**< linear velocity profile */
    MultiItvFunction* fctAVel_;                        /**< angular velocity profile */
    MultiItvFunction* fctAAcc_;                        /**< angular acceleration profile */
    MultiItvFunction* fctTH_;                          /**< heading profile */
    Polynomial* fctAccWithoutBounds_;                  /**< non bounded linear acceleration profile */
    Polynomial* fctVelWithoutBounds_;                  /**< non bounded linear velocity profile */
    Polynomial* fctAVelWithoutBounds_;                 /**< non bounded angular velocity profile */
    Polynomial* fctAAccWithoutBounds_;                 /**< non bounded angular acceleration profile */
    Polynomial* fctTHWithoutBounds_;                   /**< non bounded heading profile */

    bool xyFixed_;                                     /**< true if the final position xy must be fixed */
    bool thFixed_;                                     /**< true if the final heading th must be fixed */
    bool vFixed_;                                      /**< true if the final velocity v  must be fixed */
    bool phFixed_;                                     /**< true if the final steering angle ph must be fixed */

    int* matStateInd_;                                 /**< indexes of the state parameters in the matrix */

    /******** Partial derivative computations *********/

    /** 
     *  \brief compute the partial derivatives dS/dP
     */
    void derivSwrtP();

    double itgdXdB(const double t) const;
    double itgdXdC(const double t) const;
    double itgdXdD(const double t) const;
    double itgdXdB2(const double t) const;
    double itgdXdC2(const double t) const;
    double itgdXdD2(const double t) const;
    double itgdYdB(const double t) const;
    double itgdYdC(const double t) const;
    double itgdYdD(const double t) const;
    double itgdYdB2(const double t) const;
    double itgdYdC2(const double t) const;
    double itgdYdD2(const double t) const;

    double TH(const double t) const;
    double V(const double t) const;
    double AV(const double t) const;
    double A(const double t) const;
    double AA(const double t) const;
    double J(const double t) const;
    double AJ(const double t) const;

    double itgd2X_dBdB(const double t) const;
    double itgd2X_dBdC(const double t) const;
    double itgd2X_dBdD(const double t) const;
    double itgd2X_dBdB2(const double t) const;
    double itgd2X_dBdC2(const double t) const;
    double itgd2X_dBdD2(const double t) const;

    double itgd2X_dCdC(const double t) const;
    double itgd2X_dCdD(const double t) const;
    double itgd2X_dCdB2(const double t) const;
    double itgd2X_dCdC2(const double t) const;
    double itgd2X_dCdD2(const double t) const;

    double itgd2X_dDdD(const double t) const;
    double itgd2X_dDdB2(const double t) const;
    double itgd2X_dDdC2(const double t) const;
    double itgd2X_dDdD2(const double t) const;

    //   double itgd2X_dB2dB2(const double t) const;
    //   double itgd2X_dB2dC2(const double t) const;
    //   double itgd2X_dB2dD2(const double t) const;

    //   double itgd2X_dC2dC2(const double t) const;
    //   double itgd2X_dC2dD2(const double t) const;

    //   double itgd2X_dD2dD2(const double t) const;

    double itgd2Y_dBdB(const double t) const;
    double itgd2Y_dBdC(const double t) const;
    double itgd2Y_dBdD(const double t) const;
    double itgd2Y_dBdB2(const double t) const;
    double itgd2Y_dBdC2(const double t) const;
    double itgd2Y_dBdD2(const double t) const;

    double itgd2Y_dCdC(const double t) const;
    double itgd2Y_dCdD(const double t) const;
    double itgd2Y_dCdB2(const double t) const;
    double itgd2Y_dCdC2(const double t) const;
    double itgd2Y_dCdD2(const double t) const;

    double itgd2Y_dDdD(const double t) const;
    double itgd2Y_dDdB2(const double t) const;
    double itgd2Y_dDdC2(const double t) const;
    double itgd2Y_dDdD2(const double t) const;

    //   double itgd2Y_dB2dB2(const double t) const;
    //   double itgd2Y_dB2dC2(const double t) const;
    //   double itgd2Y_dB2dD2(const double t) const;

    //   double itgd2Y_dC2dC2(const double t) const;
    //   double itgd2Y_dC2dD2(const double t) const;

    //   double itgd2Y_dD2dD2(const double t) const;



    //   double dXdB(const double t);
    //   double dXdC(const double t);
    //   double dXdD(const double t);
    //   double dXdB2(const double t);
    //   double dXdC2(const double t);
    //   double dXdD2(const double t);
    //   double dYdB(const double t);
    //   double dYdC(const double t);
    //   double dYdD(const double t);
    //   double dYdB2(const double t);
    //   double dYdC2(const double t);
    //   double dYdD2(const double t);

    /********* Optimization related routines **********/

    /**
     *  \brief compute an initial guess of parameters to perform maneuvers
     */
    virtual void ComputeManeuverIG();

    /**
     *  \brief compute a default initial guess of parameters
     */
    virtual void ComputeDfltIG();

    /**
     *  \brief select an initial guess of parameters from an IG database
     *  \return 0 if a set of parameters has been found, -1 else
     */
    virtual int ComputeIGfromDB();

    //   /**
    //    *  \brief set the initial guess with a vector of parameters
    //    *  \param parameters : the vector of parameters
    //    */
    //   void ComputeIG(Vec_DP** parameters);

    /**
     *  \brief compute the error between the state reached and the state we want to reach
     */
    void ComputeG();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters
     */
    void ComputeJacobianS();

    /**
     *  \brief compute the second order partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeHessianS();

    /**
     *  \brief compute the residual between the goal state and the stae reached
     *  \return the residual between the goal state and the stae reached
     */
    virtual double ComputeResidual() const;
  
    /**************** Quadrature tools ****************/
  
    double qgaus(double (CLTrajGen::*func)(const double) const, const double a, const double b) const;
  
    /********* Simulation protected functions *********/

    /**
     *  \brief compute the linear acceleration profile
     */
    void computeFctAcc();

    /**
     *  \brief compute the linear velocity profile
     */
    void computeFctVel();

    /**
     *  \brief compute the angular acceleration profile
     */
    void computeFctAAcc();

    /**
     *  \brief compute the angular velocity profile
     */
    void computeFctAVel();

    /**
     *  \brief compute the heading profile
     */
    void computeFctTH();

    /**
     *  \brief compute the angular velocity, angular acceleration and heading profiles
     */
    void computeAngularFct();

    /**
     *  \brief compute all profiles
     */
    void computeAllFct();

  
  public:  
    double L_;                                         // length of the robot

    double finalErrX_;                                 /**< final error over X found */
    double finalErrY_;                                 /**< final error over Y found */
    double finalErrTH_;                                /**< final error over TH found */

    vector<double> *posX;                              /**< vector containing a sampling of the position X */
    vector<double> *posY;                              /**< vector containing a sampling of the position Y */
    vector<double> *timeT;                             /**< vector containing a sampling of the time T */
    vector<double> *velV;                              /**< vector containing a sampling of the linear velocity V */
    vector<double> *steerS;                            /**< vector containing a sampling of the steering angle S */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */
    CLTrajGen();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the constraints of the car-like system
     */
    CLTrajGen(const CarLikeSys* sys);

    /**
     *  \brief destructor
     */
    virtual ~CLTrajGen();
  
    /************ Initial Guess definition ************/

  public : 
    /**
     *  \brief set the initial guess of parameters
     *  \param b,c,d : coefficients of the polynomial angular acceleration profile
     *  \param b,c,d : coefficients of the polynomial linear acceleration profile
     *  \param t : final time to compute the trajectory (time of the final state)
     */
    void setInitialGuess(double b,double c,double d,double b2,double c2,double d2,double t);

    //   /**
    //    *  \brief set the initial and final state-times
    //    *  \param x0 : initial state-time
    //    *  \param xf : final state-time
    //    *
    //    *  x0 = (X0,Y0,TH0,OM0,V0,T0)
    //    *  xf = (Xf,Yf,THf,OMf,Vf,Tf)
    //    */
    //   void setStateT(Vec_DP& x0,Vec_DP& xf);

    /**
     *  \brief set the system parameters
     *  \param constraints : constraints defined over the system
     */
    virtual void setSysParams(const System* constraints);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialGuess(const double* parameters);

    /**
     *  \brief set the initial state
     *  \param x0 : the initial state-time
     *
     *  x0 = (X0,Y0,TH0,OM0,V0,T0)
     */
    virtual void setInitialConfig(const double* x0);


    /************* Trajectory Generation **************/

  protected : 
    /**
     *  \brief init a vector of the constraints defined over the system
     *  \param sys : the robotic system features
     *  \param vcstt : the resulting vector of constraints
     */
    virtual void initCstt(System* sys,Vec_DP& vcstt);

    /**
     *  \brief init the fixed state parameters
     *  \return the size of the vector of free parameters
     */
    virtual int initFixedSParam();

    //   /**
    //    *  \brief generate a trajectory by constraints optimiation
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    *  \param parameters : initial guess of parameters
    //    *
    //    *  x0 = (X0,Y0,TH0,OM0,V0,T0)
    //    *  xf = (Xf,Yf,THf,OMf,Vf,Tf)
    //    *  constraints = (maxCurv,maxAcc,maxVel,robotLength,maxSteer,maxSteerSpd)
    //    */
    //   int ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters);

  public : 
    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sf : final state
     *  \param parameters : initial guess of parameters
     *  \return 0 if a solution has been found, -1 else
     */
    int ComputeTrajectory(const CLStateT& s0,const CLStateT& sf,double** parameters = NULL);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sf : final state
     *  \param parameters : initial guess of parameters
     *  \return 0 if a solution has been found, -1 else
     */
    int ComputeTrajectory(const CLStateT& s0,const CLStateT& sf,Vec_DP** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sys : constraints defined over the system
     *  \param param : initial guess of parameters
     */
    virtual void ComputeFromIG(const CLStateT& s0,const double* param);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sys : constraints defined over the system
     *  \param param : initial guess of parameters
     */
    inline virtual void ComputeFromIG(const StateTime& s0,const double* param) 
    { this->ComputeFromIG(*((CLStateT*)(&s0)),param); }
  
    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param statet : initial state-time
     *  \param sys : constraints defined over the robotic system
     *  \param ctrlt : control-time applied on the system
     */
    virtual void ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt);

    /****** Trajectory Generation (2nd version) ******/

  protected : 
    /**
     *  \brief init the cot function weights
     */
    virtual void initWeights();

  public : 
    /**
     *  \brief generate a trajectory between two states
     *  \param s0 : initial state-time
     *  \param sf : initial state-time
     *  \param p : initial guess of parameters
     */
    int genTraj(const CLStateT& s0,const CLStateT& sf,const double* p=NULL);

    /********** State Parameters Computation **********/  

  protected : 
    double funcItgX(const double t) const;
    double funcItgY(const double t) const;

  public : 
    double funcX(const double t) const;
    double funcY(const double t) const;    
    double funcTheta(const double t) const;
    double funcAVel(const double t) const;
    double funcVel(const double t) const;
    double funcAAcc(const double t) const;
    double funcAcc(const double t) const;
    double funcSteer(const double t) const;
    double funcSteerSpd(const double t) const;

  protected : 
    double funcItgXwb(const double t) const;
    double funcItgYwb(const double t) const;
    double funcXwb(const double t) const;
    double funcYwb(const double t) const;    
    double funcThetawb(const double t) const;
    double funcAVelwb(const double t) const;
    double funcAAccwb(const double t) const;
    double funcVelwb(const double t) const;
    double funcAccwb(const double t) const;
    double funcSteerwb(const double t) const;
    double funcSteerSpdwb(const double t) const;

    double B2a,C2a,D2a;
    double B2v,C2v,D2v;
    double Baa,Caa,Daa;
    double Bav,Cav,Dav;

    /*************** Forward Solution *****************/

  public : 
    /**
     *  \brief compute the state reached by the current set of parameters
     */
    void ForwardSolution();
    
    /**************** Output Functions ****************/

  public : 
    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (x,y,th,ph,v,t)
     */
    virtual void getStateTReachedAt(double* rstatet,double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    CLStateT getCLStateTReachedAt(double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (x,y,th,ph,v,t)
     */
    virtual void getFinalStateTReached(double* rstatet) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    CLStateT getFinalCLStateTReached() const;

    /**
     *  \brief save a sampling of the trajectory profile
     *  \param nbp : number of sampling points
     */
    virtual void SaveTrajectory(int nbp); 

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     */  
    virtual void traceOn(Map *map,int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     *  \param initPos : initial position
     */  
    virtual void traceOn(Map *map,int nbp,const PositionXYTh& initPos);

    /**
     *  \brief return a sample trajectory from the parameterized one
     *  \param ts : the sampling time step
     *  \return a sample trajectory from the parameterized one
     *
     *  A new instance of sample trajectory is returned here
     */
    virtual CLSTraj* sampledTraj(double ts) const;

    /*************** Utility Functions ****************/

  protected : 
    /**
     *  \brief convert the initial and final input state-times into tables of double
     *  \param initST : the input initial state-time
     *  \param goalST : the input goal state-time
     *  \param initTD : the resulting initial table
     *  \param goalTD : the resulting goal table
     *
     *  Tables must be allocated first
     */
    virtual void InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const;

  public : 
    /**
     *  \brief display linear acceleration and velocity, angular velocity, steering angle, heading and position profiles
     */
    void displayProfiles();

    /**
     *  \brief display linear acceleration and velocity, angular velocity, steering angle, heading and position profiles
     */
    void displayProfiles2();

    /**
     *  \brief display linear acceleration and velocity, angular velocity and steering angle profiles without bounds
     */
    void displayProfilesWB();

    /*
     *  \brief display all bounded and non bounded profiles
     */
    void displayAllProfiles();

    /**
     *  \brief display all functions
     */
    virtual void displayAllFct();

    /**
     *  \brief display in an openGL viewer
     *  \param tmin : time at which the trajectory begins
     */
    void toGL(double tmin);

    /**
     *  \brief compute a sample of the trajectory in the workspace (X*Y*T)
     *  \param tmin : time at which the trajectory begins
     *  \return a vector of three sampled vectors (position X, position Y, time T)
     */
    vector<vector<double> > sampleXYT(double tmin);
  };

  // double itg2dXdB(double t, void* param);
  // double itg2dXdC(double t, void* param);
  // double itg2dXdD(double t, void* param);
  // double itg2dXdB2(double t, void* param);
  // double itg2dXdC2(double t, void* param);
  // double itg2dXdD2(double t, void* param);
  // double itg2dYdB(double t, void* param);
  // double itg2dYdC(double t, void* param);
  // double itg2dYdD(double t, void* param);
  // double itg2dYdB2(double t, void* param);
  // double itg2dYdC2(double t, void* param);
  // double itg2dYdD2(double t, void* param);

};
#endif // MPC_CLTRAJGEN_HPP
