#ifndef MPC_SSTRAJGEN_HPP
#define MPC_SSTRAJGEN_HPP

#include "trajGen/inc/TrajGen.hpp"
#include "math/inc/MultiItvFunction.hpp"
#include "system/inc/SpaceShipSys.hpp"

namespace mpc
{
                                    /**************************************************/
                                    /***************** Class SSTrajGen ****************/
                                    /**************************************************/

  /** \class SSTrajGen
   *  \brief Trajectory generator TIJI for a space ship system
   */
  class SSTrajGen : public TrajGen
  {
  protected :
    double maxAcc;                                     /**< maximal linear acceleration */
    double maxVel;                                     /**< maximal linear velocity */
    double maxAAcc;                                    /**< maximal angular acceleration */
    double maxAVel;                                    /**< maximal angular velocity */

    MultiItvFunction* fctAcc_;                         /**< linear acceleration profile */
    MultiItvFunction* fctAAcc_;                        /**< angular acceleration profile */
    MultiItvFunction* fctAVel_;                        /**< angular velocity profile */
    MultiItvFunction* fctTH_;                          /**< heading profile */
    //   MultiItvFunction* fctVelX_;                        /**< linear velocity on X profile */
    //   MultiItvFunction* fctVelY_;                        /**< linear velocity on Y profile */

    Polynomial* fctAccWithoutBounds_;                  /**< non bounded linear acceleration profile */
    Polynomial* fctAAccWithoutBounds_;                 /**< non bounded angular acceleration profile */
    Polynomial* fctAVelWithoutBounds_;                 /**< non bounded angular velocity profile */
    Polynomial* fctTHWithoutBounds_;                   /**< non bounded heading profile */

    int lastItvAccBdd_;                                // last interval of the acceleration, linear
    int lastItvAVelBdd_;                               // -1 = minB,  0 = not bounded,  1 = maxB
    int lastItvAAccBdd_;

  public : 
    double finalErrX_;                                 /**< final error over X found */
    double finalErrY_;                                 /**< final error over Y found */
    double finalErrTH_;                                /**< final error over TH found */

    vector<double> *posX;                              /**< vector containing a sampling of the position X */
    vector<double> *posY;                              /**< vector containing a sampling of the position Y */
    vector<double> *headTH;                            /**< vector containing a sampling of the heading TH */
    vector<double> *velVX;                             /**< vector containing a sampling of the linear velocity VX */
    vector<double> *velVY;                             /**< vector containing a sampling of the linear velocity VY */
    vector<double> *avelOM;                            /**< vector containing a sampling of the angular velocity OM */
    vector<double> *timeT;                             /**< vector containing a sampling of the time T */

    bool xyFixed_;                                     /**< true if the final position xy must be fixed */
    bool thFixed_;                                     /**< true if the final heading th must be fixed */
    bool vFixed_;                                      /**< true if the final linear velocity v  must be fixed */
    bool omFixed_;                                     /**< true if the final angular velocity om must be fixed */

    int* matStateInd_;                                 /**< indexes of the state parameters in the matrix */

    SSState SSweights_;                                /**< values used to weight the cost function */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     */  
    SSTrajGen();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the constraints of the spaceship system
     */
    SSTrajGen(const SpaceShipSys* sys);

    /**
     *  \brief destructor
     */  
    virtual ~SSTrajGen();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    virtual const SpaceShipSys* sys() const { return ((SpaceShipSys*)sys_); }

    /**
     *  \brief fix or free the final position
     *  \param fixed : true if the final position must be fixed
     */
    void fixFinalXY(bool fixed);

    /**
     *  \brief fix or free the final heading
     *  \param fixed : true if the final heading must be fixed
     */
    void fixFinalTH(bool fixed);

    /**
     *  \brief fix or free the final linear velocity
     *  \param fixed : true if the final linear velocity must be fixed
     */
    void fixFinalV(bool fixed);

    /**
     *  \brief fix or free the final angular velocity
     *  \param fixed : true if the final angular velocity must be fixed
     */
    void fixFinalOM(bool fixed);

    /********* Optimization related routines **********/

  protected : 
    /**
     *  \brief compute a default initial guess of parameters
     */
    virtual void ComputeDfltIG();

    /**
     *  \brief select an initial guess of parameters from an IG database
     *  \return 0 if a set of parameters has been found, -1 else
     */
    virtual int ComputeIGfromDB();

    //   /**
    //    *  \brief set the initial guess with a vector of parameters
    //    *  \param parameters : the vector of parameters
    //    */
    //   virtual void ComputeIG(Vec_DP** parameters);

    /**
     *  \brief compute the error between the state reached and the state we want to reach
     */
    virtual void ComputeG();

    //   /**
    //    *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
    //    */
    //   void ComputeJacobianG_FD();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeJacobianS();

    /**
     *  \brief compute the second order partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeHessianS();

    long double dVXdB(const long double t) const;
    long double dVXdC(const long double t) const;
    long double dVXdD(const long double t) const;
    long double dVXdB2(const long double t) const;
    long double dVXdC2(const long double t) const;
    long double dVXdD2(const long double t) const;
    long double dVYdB(const long double t) const;
    long double dVYdC(const long double t) const;
    long double dVYdD(const long double t) const;
    long double dVYdB2(const long double t) const;
    long double dVYdC2(const long double t) const;
    long double dVYdD2(const long double t) const;

    long double itgdVXdB(const long double t) const;
    long double itgdVXdC(const long double t) const;
    long double itgdVXdD(const long double t) const;
    long double itgdVXdB2(const long double t) const;
    long double itgdVXdC2(const long double t) const;
    long double itgdVXdD2(const long double t) const;
    long double itgdVYdB(const long double t) const;
    long double itgdVYdC(const long double t) const;
    long double itgdVYdD(const long double t) const;
    long double itgdVYdB2(const long double t) const;
    long double itgdVYdC2(const long double t) const;
    long double itgdVYdD2(const long double t) const;

    /**************** Quadrature tools ****************/
  
  protected : 
    /**
     *  \brief Quadrature tools...
     */
    long double qgaus(long double (SSTrajGen::*func)(const long double) const, const long double a, const long double b) const;

    /*********** State profiles computations **********/

  protected : 
    /**
     *  \brief compute the linear acceleration profile
     */
    void computeFctAcc();

    /**
     *  \brief compute the angular acceleration profile
     */
    void computeFctAAcc();

    /**
     *  \brief compute the angular velocity profile
     */
    void computeFctAVel();

    /**
     *  \brief compute the heading profile
     */
    void computeFctTH();

    //   /**
    //    *  \brief compute the linear velocity on X profile
    //    */
    //   void computeFctVelX();

    //   /**
    //    *  \brief compute the linear velocity on Y profile
    //    */
    //   void computeFctVelY();

    /**
     *  \brief compute all profiles
     */
    void computeAllFct();

  
    /********** State Parameters Computation **********/  

  public : 
    /**
     *  \brief get the value of the linear acceleration profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear acceleration profile at time t
     */
    double funcAcc(const double t) const;

    /**
     *  \brief get the value of the angular acceleration profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular acceleration profile at time t
     */
    double funcAAcc(const double t) const;

  protected : 
    /**
     *  \brief integration function for the velocity VX profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    long double funcItgVelX(const long double t) const;

    /**
     *  \brief integration function for the velocity VY profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    long double funcItgVelY(const long double t) const;

  public : 
    /**
     *  \brief get the value of the linear velocity on X profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear velocity on X profile at time t
     */
    long double funcVelX(const long double t) const;

    /**
     *  \brief get the value of the linear velocity on Y profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear velocity on Y profile at time t
     */
    long double funcVelY(const long double t) const;

    /**
     *  \brief get the value of the angular velocity profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular velocity profile at time t
     */
    double funcAVel(const double t) const;

    /**
     *  \brief get the value of the heading profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the heading profile at time t
     */
    double funcTheta(const double t) const;

    /**
     *  \brief get the value of the position X profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position X profile at time t
     */
    double funcX(const double t) const;

    /**
     *  \brief get the value of the position Y profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position Y profile at time t
     */
    double funcY(const double t) const;

    /*************** Forward Solution *****************/

  protected :   
    /**
     *  \brief compute the state reached by the current set of parameters
     */
    virtual void ForwardSolution();

    /***************** Initialization *****************/

  protected : 
    //   /**
    //    *  \brief set the initial state, final state and constraints vectors
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    */
    //   virtual void setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints);

    /**
     *  \brief set the system parameters
     *  \param constraints : constraints defined over the system
     */
    virtual void setSysParams(const System* constraints);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialGuess(const double* parameters);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     *
     *  x0 = (X0,Y0,TH0,VX0,VY0,OM0,T0)
     */
    virtual void setInitialConfig(const double* x0);

    /************* Trajectory Generation **************/

  protected : 
    /**
     *  \brief init a vector of the constraints defined over the system
     *  \param sys : the robotic system features
     *  \param vcstt : the resulting vector of constraints
     */
    virtual void initCstt(System* sys,Vec_DP& vcstt);

    /**
     *  \brief init the fixed state parameters
     *  \return the size of the vector of free parameters
     */
    virtual int initFixedSParam();

  public : 
    //   /**
    //    *  \brief generate a trajectory by constraints optimiation
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    *  \param parameters : initial guess of parameters
    //    *
    //    *  x0 = (X0,Y0,TH0,VX0,VY0,OM0,T0)
    //    *  xf = (Xf,Yf,THf,VXf,VYf,OMf,Tf)
    //    *  constraints = (maxAAcc,maxAcc,maxAVel,maxVel)
    //    */
    //   virtual int ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sys : constraints defined over the system
     *  \param param : initial guess of parameters
     */
    inline virtual void ComputeFromIG(const StateTime& s0,const double* param) 
    { this->ComputeFromIG(*((SSStateT*)(&s0)),param); }

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param statet : initial state-time
     *  \param ctrlt : control-time applied on the system
     */
    virtual void ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt);

    /****** Trajectory Generation (2nd version) ******/

  protected : 
    /**
     *  \brief init the cot function weights
     */
    virtual void initWeights();

  public : 
    /**
     *  \brief generate a trajectory between two states
     *  \param s0 : initial state-time
     *  \param sf : initial state-time
     *  \param p : initial guess of parameters
     */
    int genTraj(const SSStateT& s0,const SSStateT& sf,const double* p=NULL);
  
    /**************** Output Functions ****************/

  public : 
    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,vx,vy,om,t)
     */
    virtual void getStateTReachedAt(double* rstatet,double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    SSStateT getSSStateTReachedAt(double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,vx,vy,om,t)
     */
    virtual void getFinalStateTReached(double* rstatet) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    SSStateT getFinalSSStateTReached() const;

    /**
     *  \brief save a sampling of the trajectory profile
     *  \param nbp : number of sampling points
     */
    virtual void SaveTrajectory(int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     */
    virtual void traceOn(Map *map,int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     *  \param initPos : initial position
     */  
    virtual void traceOn(Map *map,int nbp,const PositionXYTh& initPos);

    /**
     *  \brief return a sample trajectory from the parameterized one
     *  \param ts : the sampling time step
     *  \return a sample trajectory from the parameterized one
     *
     *  A new instance of sample trajectory is returned here
     */
    virtual SSSTraj* sampledTraj(double ts) const;

    /*************** Utility Functions ****************/

  protected : 
    /**
     *  \brief convert the initial and final input state-times into tables of double
     *  \param initST : the input initial state-time
     *  \param goalST : the input goal state-time
     *  \param initTD : the resulting initial table
     *  \param goalTD : the resulting goal table
     *
     *  Tables must be allocated first
     */
    virtual void InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const;

  public : 
    /**
     *  \brief display linear and angular velocity, heading and position profiles
     */
    void displayProfiles();

    /**
     *  \brief display all functions
     */
    virtual void displayAllFct();
  };

};
#endif // MPC_SSTRAJGEN_HPP

