#ifndef MPC_TRAJGEN_HPP
#define MPC_TRAJGEN_HPP

#include "system/inc/System.hpp"
#include "common/inc/StateTime.hpp"
#include "common/inc/ControlTime.hpp"

#include "math/inc/nrtypes.h"
#include "math/inc/Matrix.hpp"
//#include "Utils.hpp"
#include "environment/inc/Map.hpp"

#include "motion/inc/STrajectory.hpp"
#include "motion/inc/PSTraj.hpp"

#include "trajGen/inc/InitialGuess.hpp"

namespace mpc
{
  class InitialGuess;

  /**
   * \enum timeCstt_t
   * \brief Enumeration of the different time constraint status
   *
   * TIME_FIXED : the time is fixed
   * TIME_FREE  : the time is free
   * TIME_CSTD  : the time is constrained between two values
   */
  typedef enum {
    TIME_FIXED     =  0,
    TIME_FREE      =  1,
    TIME_CSTD      =  2
  } timeCstt_t;

  /**
   * \enum costFct_t
   * \brief Enumeration of the different time constraint status
   *
   * SIMPLE_CF  : simple cost function
   * SQUARE_CF  : square cost function
   */
  typedef enum {
    OLD_CF        = -1,
    SIMPLE_CF     =  0,
    SQUARE_CF     =  1
  } costFct_t;

                                    /**************************************************/
                                    /****************** Class TrajGen *****************/
                                    /**************************************************/

  /** \class TrajGen
   *  \brief Trajectory generator TIJI
   */
  class TrajGen
  {
  protected : 
    const System* sys_;                         /**< robotic system for which the trajectory is generated */

    InitialGuess* ig_;                          /**< initial guess database */

    PCTraj* ctraj_;                             /**< Control trajectory */
    PSTraj* straj_;                             /**< State trajectory */

    Vec_DP vecState_x0;                         /**< vector containing the initial state of the system */
    Vec_DP vecState_xf;                         /**< vector containing the final state of the system */
    Vec_DP vecConstraints_g;                    /**< vector of constraints defined over the system */
  
    Mat_DP matJacobianG;                        /**< matrix containing the first order partial derivatives 
						 *   of the state wrt. parameters computed during the trajectory 
						 *   generation process */
    Mat_DP* matInvJacobianG;                    /**< Inverse matrix of the first order partial derivatives */
  
    Vec_DP vecCorrParameters_p;                 /**< Vector of the correction applied over parameters during the 
						 *   trajectory generation process */

    timeCstt_t tcStatus_;                       /**< time constraint status (fixed, free or constrained between two values) */
    double minTime_;                            /**< the minimal time allowed if the time is constrained */
    double maxTime_;                            /**< the maximal time allowed if the time is constrained */

    //   static bool timeFixed_;                     /**< true if the time is fixed */
    static bool variableCoef_;                  /**< true if the coefficient applied to the correction is variable */

    static int DBG_;                            /**< Debugging level */
    static int maxNbStep_;                      /**< Maximal number of steps allowed the generate the trajectory */
    static double maxResidual_;                 /**< Maximal residual allowed to stop the trajectory 
						 *   generation process */
 
    static const double E = 0.000000001;        /**< small value for computation of the derivatives by finite differences */
    static const double UnSurE = 1000000000.0;  /**< great value for computation of the derivatives by finite differences */

    bool FDpartDeriv_;                          /**< true if the partial derivatives are computed by finite differences */

    double residual_;                           /**< current residual */
    double oldResidual_;                        /**< residual of the previous step */
    double COEF_COR_;                           /**< initial coefficient applied to the correction */
    double coefCor_;                            /**< coefficient applied to the correction over the set of parameters */
    int nbDeacreaseStep_;                       /**< number of steps during which the residual is decreasing */
    int nbIncreaseStep_;                        /**< number of steps during which the residual is increasing */

    Vec_DP vecBestParam_;                       /**< vector to store the best parameters */
    double bestResidual_;                       /**< best residual obtained during the trajectory generation process */
    int CNT_;                                   /**< number of the current step */
    int bestCNT_;                               /**< number of the step at which we found the better residual */

    ///////// variables used for the second version of the trajectory generation process
    double* weightsCF_;                         /**< values used to weight the state parameters for the computation of the cost function */
    //   Vec_DP JacobianJ_;                          /**< jacobian of the cost function wrt parameters p */
    //   Vec_DP invJacobianJ_;                       /**< inverse vector of the jacobian of the cost function wrt parameters p */
    //   Mat_DP HessianJ_;                           /**< hessian of the cost function wrt parameters p */
    //   Mat_DP invHessianJ_;                        /**< inverse matrix of the hessian of the cost function wrt parameters p */

    //   mpc::Matrix2D<double> JacobianS_;           /**< jacobian of the state vector wrt parameters p */    (temporary not used, replaced by matJacobianG)
    //   mpc::Matrix3D<double> HessianS_;            /**< hessian of the state vector wrt parameters p */

    Vec_DP vecJacobianS_;                          /**< jacobian of the state vector given as a single vector of size dimS * dimP */
    Mat_DP matHessianS_;                           /**< hessian of the state vector given as a matrix of size dimP * (dimS * dimP) */
    Mat_DP invHessianS_;                           /**< inverse of the hessian of the state vector given as a matrix of size (dimS * dimP) * dimP */

    costFct_t CFtype_;                          /**< cost function type (simple or square) */

  public : 
    int xLength;                                /**< State-time length (x, y, th, ..., t) */ 
    int pLength;                                /**< Parameters length (B1, B2, C1, ...) */ 
    int gLength;                                /**< Constraints length (x(t)=x_f, y(t)=y_f, ...) */
    int dLength;                                /**< Dynamic Constraints length (maxK, maxAcc, ...) */   

    Vec_DP vecState_xsim;                       /**< vector containing the current state reached */
    Vec_DP* vecParameters_p;                    /**< vector of parameters */

    double finalResidual_;                      /**< final residual found */
    double finalNbSteps_;                       /**< number of steps required to generate the trajectory */

    /********** Constructors and destructors **********/

  protected : 
    /**
     *  \brief default constructor
     */  
    TrajGen();

    /**
     *  \brief init constructor
     *  \param sys : the constraints of the robotic system
     */  
    TrajGen(const System* sys);

  public : 
    /**
     *  \brief destructor
     */  
    virtual ~TrajGen();

    /*************** Getters and Setters **************/

  protected : 
    /**
     *  \brief set d(S_i) / d(p_j) 
     *  \param i : index of the state vector in the hessian matrix
     *  \param j : index of the parameter vector in the hessian matrix
     *  \param val : the new value
     */
    void setJacobianS(int i,int j,double val);

    /**
     *  \brief return a value of the jacobian of the state vector
     *  \param i : index of the state vector in the jacobian matrix
     *  \param j : index of the parameter vector in the jacobian matrix
     *  \return d(S_i) / d(p_j) 
     */
    double getJacobianS(int i,int j);

    /**
     *  \brief set d^2(S_i) / (d(p_j) d(p_k)) 
     *  \param i : index of the state vector in the hessian matrix
     *  \param j,k : indexes of parameters vectors in the hessian matrix
     *  \param val : the new value
     */
    void setHessianS(int i,int j,int k,double val);

    /**
     *  \brief return a value of the hessian of the state vector
     *  \param i : index of the state vector in the hessian matrix
     *  \param j,k : indexes of parameters vectors in the hessian matrix
     *  \return d^2(S_i) / (d(p_j) d(p_k))
     */
    double getHessianS(int i,int j,int k);

    /**
     *  \brief return a value of the inverse hessian of the state vector
     *  \param i : index of the state vector in the inverse hessian matrix
     *  \param j,k : indexes of parameters vectors in the inverse hessian matrix
     *  \return (d^2(S_i) / (d(p_j) d(p_k)))^{-1}
     */
    double getInvHessianS(int i,int j,int k);

  public : 
    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    virtual const System* sys() const = 0;

    /**
     *  \brief getter : return the maximum number of steps
     *  \return maximum number of steps allowed to generate the trajectory
     */
    static int getMaxNbStep();

    /**
     *  \brief setter : set the maximum number of steps
     *  \param n : maximum number of steps allowed to generate the trajectory
     */
    static void setMaxNbStep(const int n);

    /**
     *  \brief getter : return the maximal residual allowed
     *  \return maximal residual allowed to stop the trajectory generation process
     */
    static double getMaxResidual();

    /**
     *  \brief setter : set the maximal residual allowed
     *  \param r : maximal residual allowed to stop the trajectory generation process
     */
    static void setMaxResidual(const double r);

    /**
     *  \brief getter : return the debugging level
     *  \return debugging level
     */
    static int getDBG();

    /**
     *  \brief setter : set the debugging level
     *  \param dbg : debugging level
     */
    static void setDBG(const int dbg);

    //   /**
    //    *  \brief getter : return true if the time is fixed
    //    *  \return true if the time is fixed
    //    */
    //   inline static bool timeFixed() { return TrajGen::timeFixed_; }

    //   /**
    //    *  \brief setter : set the fixed time mode
    //    *  \param tf : true if the time is fixed
    //    */
    //   inline static void setTimeFixed(bool tf) { TrajGen::timeFixed_ = tf; }

    /**
     *  \brief getter : return true if the coefficient applied the the correction is variable
     *  \return true if the coefficient applied the the correction is variable
     */
    inline static bool variableCoef() { return TrajGen::variableCoef_; }

    /**
     *  \brief setter : set the debugging level
     *  \param dbg : debugging level
     */
    inline static void setVariableCoef(bool vc) { TrajGen::variableCoef_ = vc; }

    /**
     *  \brief getter : return true if the partial derivatives are computed by finite differences
     *  \return true if the partial derivatives are computed by finite differences
     */
    inline bool FDpartDeriv() const { return FDpartDeriv_; }

    /**
     *  \brief setter : set the partial derivatives by finite differences mode
     *  \param fd : true if the partial derivatives are computed by finite differences
     */
    inline void setFDpartDeriv(bool fd) { FDpartDeriv_ = fd; }

    /**
     *  \brief get back the final residual
     *  \return final residual
     */
    inline double finalResidual() const { return this->finalResidual_; }

    /**
     *  \brief get back the final residual
     *  \return final residual
     */
    inline double finalNbSteps() const { return this->finalNbSteps_; }

    /**
     *  \brief return the time constraint status
     *  \return the time constraint status
     */
    inline timeCstt_t tcStatus() const { return this->tcStatus_; }

    /**
     *  \brief set the time constraint status
     *  \param the time constraint status
     */
    void setTcStatus(timeCstt_t tcS,double minT,double maxT);

    /**
     *  \brief return the number of free parameters
     *  \return the number of free parameters
     */
    int pLengthFree() const;

    /**
     *  \brief return the final parameters
     *  \param params : the resulting final parameters
     *
     *  params must be allocated first
     */
    void finalParams(double* params) const;

    /**
     *  \brief set the initial guess database to use
     *  \param ig : a pointer to the initial guess database to use
     */
    inline void setIGdb(InitialGuess* ig) { this->ig_ = ig; }

    /********* Optimization related routines **********/

  protected : 
    /**
     *  \brief compute a default initial guess of parameters
     */
    virtual void ComputeDfltIG() = 0;

    /**
     *  \brief select an initial guess of parameters from an IG database
     *  \return 0 if a set of parameters has been found, -1 else
     */
    virtual int ComputeIGfromDB() = 0;

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void ComputeIG(Vec_DP** parameters);

    /**
     *  \brief compute the error between the state reached and the state we want to reach
     */
    virtual void ComputeG() = 0;

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeJacobianS() = 0;

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
     */
    virtual void ComputeJacobianG_FD();

    /**
     *  \brief invert the partial derivatives matrix
     */
    void ComputeInvJacobianS();

    /**
     *  \brief set the vector representing the jacobian of S
     */
    void setVecJacobianS();

    /**
     *  \brief compute the second order partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeHessianS() = 0;

    /**
     *  \brief invert the hessian matrix
     */
    virtual void ComputeInvHessianS();

    /**
     *  \brief compute the correction to apply to the parameters
     */
    virtual void ComputeCorrection();

    /**
     *  \brief apply the correction to the set of parameters
     */
    virtual void ApplyCorrection();

    /**
     *  \brief compute the residual between the goal state and the stae reached
     *  \return the residual between the goal state and the stae reached
     */
    virtual double ComputeResidual() const;

    /**
     *  \brief save the current vector of parameters into the vector of best parameters
     */
    void SaveBestParams();

    /**
     *  \brief load the vector of best parameters into the current vector of parameters
     */
    void LoadBestParams();


    /**************** Quadrature tools ****************/

  protected : 
    /**
     *  \brief Quadrature tools...
     */  
    void gauleg(const double x1, const double x2, double x[], double w[], int n) const;

    /**
     *  \brief Quadrature tools...
     */  
    void gauleg2(const long double x1, const long double x2, long double x[], long double w[], int n) const;

    /********* Simulation protected functions *********/
  
  protected : 
    /**
     *  \brief compute all profiles
     */
    virtual void computeAllFct() = 0;
  
    /*************** Forward Solution *****************/
  
  protected : 
    /**
     *  \brief compute the state reached by the current set of parameters
     */
    virtual void ForwardSolution() = 0;

    /***************** Initialization *****************/

  protected : 
    //   /**
    //    *  \brief set the initial state, final state and constraints vectors
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    */
    //   virtual void setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints) = 0;

    /**
     *  \brief set the initial and final states
     *  \param x0 : initial state
     *  \param xf : final state
     */
    virtual void setStateT(Vec_DP& x0,Vec_DP& xf);

    /**
     *  \brief set the initial and final states
     *  \param x0 : initial state
     *  \param xf : final state
     */
    virtual void setStateT(const double* x0,const double* xf);

    /**
     *  \brief set the system parameters
     *  \param constraints : constraints defined over the system
     */
    virtual void setSysParams(const System* constraints) = 0;

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialGuess(const double* parameters) = 0;

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialConfig(const double* x0) = 0;

    /************* Trajectory Generation **************/

  protected : 
    /**
     *  \brief initialize the trajectory generation process
     */
    virtual void initTrajGen();

//     /**
//      *  \brief init a vector of the constraints defined over the system
//      *  \param sys : the robotic system features
//      *  \param vcstt : the resulting vector of constraints
//      */
//     virtual void initCstt(System* sys,Vec_DP& vcstt) = 0;

    /**
     *  \brief init the fixed state parameters
     *  \return the size of the vector of free parameters
     */
    virtual int initFixedSParam() = 0;

  protected : 
    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param x0 : initial state
     *  \param xf : final state
     *  \param parameters : initial guess of parameters (NULL if no initial guess is defined)
     *  \return 0 if a solution has been found, -1 else
     */
    virtual int ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP** parameters);

  public : 
    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sf : final state
     *  \param parameters : initial guess of parameters (NULL if no initial guess is defined)
     *  \return 0 if a solution has been found, -1 else
     */
    virtual int ComputeTrajectory(const StateTime& s0,const StateTime& sf,double** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param x0 : initial state
     *  \param constraints : constraints defined over the system
     *  \param parameters : initial guess of parameters
     */
    virtual void ComputeFromIG(const double* x0,const double* parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param param : initial guess of parameters
     */
    virtual void ComputeFromIG(const StateTime& s0,const double* param) = 0;

//     /**
//      *  \brief generate a trajectory by constraints optimiation
//      *  \param statet : initial state-time
//      *  \param ctrlt : control-time applied on the system
//      */
//     virtual void ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt) = 0;


    /****** Trajectory Generation (2nd version) ******/

  protected : 
    /**
     *  \brief compute the cost function
     *  \return the current value of the cost function
     */
    virtual double costFunction() const;

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void ComputeIG(const double* p);

    //   /**
    //    *  \brief compute the jacobian of the cost function wrt parameters p
    //    */
    //   virtual void computeJacobianJ_1();

    //   /**
    //    *  \brief compute the jacobian of the cost function wrt parameters p
    //    */
    //   virtual void computeJacobianJ_2();

    //   /**
    //    *  \brief compute the hessian of the cost function wrt parameters p
    //    */
    //   virtual void computeHessianJ_1();

    //   /**
    //    *  \brief compute the hessian of the cost function wrt parameters p
    //    */
    //   virtual void computeHessianJ_2();

    //   /**
    //    *  \brief compute the jacobian of the cost function wrt parameters p
    //    */
    //   void computeJacobianJ();

    //   /**
    //    *  \brief compute the hessian of the cost function wrt parameters p
    //    */
    //   void computeHessianJ();

    //   /**
    //    *  \brief compute the inverse vector of the jacobian of the cost function wrt parameters p
    //    */
    //   void computeInvJacobianJ();

    //   /**
    //    *  \brief compute the inverse matrix of the hessian of the cost function wrt parameters p
    //    */
    //   void computeInvHessianJ();

    /**
     *  \brief compute the correction to apply over the set of parameters
     */
    void computeNewCorrection();

    /**
     *  \brief apply the correction to the set of parameters
     */
    virtual void applyNewCorrection();

    /**
     *  \brief initialize the trajectory generation process (for the new trajectory generation process)
     */
    virtual void initNewTrajGen();

    /**
     *  \brief init the cot function weights
     */
    virtual void initWeights() = 0;

  protected : 
    /**
     *  \brief generate a trajectory between two states
     *  \param x0 : initial state-time
     *  \param xf : initial state-time
     *  \param p : initial guess of parameters
     */
    int genTraj(const double* x0,const double* xf,const double* p=NULL);

  public : 
    /**
     *  \brief generate a trajectory between two states
     *  \param s0 : initial state-time
     *  \param sf : initial state-time
     *  \param p : initial guess of parameters
     */
    int genTraj(const StateTime& s0,const StateTime& sf,const double* p=NULL);

    /**************** Ouput Functions ****************/

  public : 
    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     */
    virtual void getStateTReachedAt(double* rstatet,double t) const = 0;

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     */
    virtual void getFinalStateTReached(double* rstatet) const = 0;

//     /**
//      *  \brief save a sampling of the trajectory profile
//      *  \param nbp : number of sampling points
//      */
//     virtual void SaveTrajectory(int nbp) = 0;

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     */  
    virtual void traceOn(Map *map,int nbp) = 0;

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     *  \param initPos : initial position
     */  
    virtual void traceOn(Map *map,int nbp,const PositionXYTh& initPos) = 0;

    /**
     *  \brief return a sample trajectory from the parameterized one
     *  \param ts : the sampling time step
     *  \return a sample trajectory from the parameterized one
     *
     *  A new instance of sample trajectory is returned here
     */
    virtual STraj* sampledTraj(double ts) const = 0;

    /*************** Utility Functions ****************/

  protected : 
    /**
     *  \brief convert the initial and final input state-times into tables of double
     *  \param initST : the input initial state-time
     *  \param goalST : the input goal state-time
     *  \param initTD : the resulting initial table
     *  \param goalTD : the resulting goal table
     *
     *  Tables must be allocated first
     */
    virtual void InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const = 0;

  public : 
    /**
     *  \brief display the different profiles defining the motion of the system
     */
    virtual void displayProfiles() = 0;

    /**
     *  \brief display the jacobian matrix
     */
    void displayJacobian();

    /**
     *  \brief display the inverse of the jacobian matrix
     */
    void displayInvJacobian();

    /**
     *  \brief display the set of parameters
     */
    virtual void displayParameters();

    /**
     *  \brief display the initial state
     */
    virtual void displayInitState();

    /**
     *  \brief display the goal state
     */
    virtual void displayGoalState();

    /**
     *  \brief display the state reached at the current step
     */
    virtual void displayStateReached();

    /**
     *  \brief display the correction computed
     */
    virtual void displayCorrection();

    /**
     *  \brief display the error between the goal state and the reached state
     */
    virtual void displayError();

    /**
     *  \brief display all functions
     */
    virtual void displayAllFct() = 0;

    //   /**
    //    *  \brief display the jacobian of the cost function
    //    */
    //   void displayJacobianJ();

    //   /**
    //    *  \brief display the inverse of the jacobian of the cost function
    //    */
    //   void displayInvJacobianJ();

    //   /**
    //    *  \brief display the hessian of the cost function
    //    */
    //   void displayHessianJ();

    //   /**
    //    *  \brief display the inverse of the hessian of the cost function
    //    */
    //   void displayInvHessianJ();

    /**
     *  \brief display the hessian of the state vector
     */
    void displayHessianS();

    /**
     *  \brief display the inverse of the hessian of the state vector
     */
    void displayInvHessianS();
  };

};
#endif // MPC_TRAJGEN_HPP

