#ifndef MPC_WCTRAJGEN_HPP
#define MPC_WCTRAJGEN_HPP

#include "trajGen/inc/TrajGen.hpp"
#include "math/inc/MultiItvFunction.hpp"
#include "system/inc/WheelChairSys.hpp"

#include "trajGen/inc/WCInitialGuess.hpp"

namespace mpc
{
class WCInitialGuess;

                                    /**************************************************/
                                    /***************** Class WCTrajGen ****************/
                                    /**************************************************/

  /** \class WCTrajGen
   *  \brief Trajectory generator TIJI for a wheelchair system
   */
  class WCTrajGen : public TrajGen
  {
  public :
    double maxAcc;                                     /**< maximal linear acceleration */
    double maxVel;                                     /**< maximal linear velocity */
    double maxAAcc;                                    /**< maximal angular acceleration */
    double maxAVel;                                    /**< maximal angular velocity */

    bool USE_VEL_CST;                                  /**< true if the linear velocity constraints are used */
    bool USE_ACC_CST;                                  /**< true if the linear acceleration constraints are used */
    bool USE_AVEL_CST;                                 /**< true if the angular velocity constraints are used */
    bool USE_AACC_CST;                                 /**< true if the angular acceleration constraints are used */

    //   bool canMoveBack_;                                 /**< true if the wheelchair can move backward */

    MultiItvFunction* fctAcc_;                         /**< linear acceleration profile */
    MultiItvFunction* fctVel_;                         /**< linear velocity profile */
    MultiItvFunction* fctAVel_;                        /**< angular velocity profile */
    MultiItvFunction* fctAAcc_;                        /**< angular acceleration profile */
    MultiItvFunction* fctTH_;                          /**< heading profile */

    Polynomial* fctAccWB_;                             /**< non bounded linear acceleration profile */
    Polynomial* fctVelWB_;                             /**< non bounded linear velocity profile */
    Polynomial* fctAVelWB_;                            /**< non bounded angular velocity profile */
    Polynomial* fctAAccWB_;                            /**< non bounded angular acceleration profile */
    Polynomial* fctTHWB_;                              /**< non bounded heading profile */

  public : 
    double finalErrX_;                                 /**< final error over X found */
    double finalErrY_;                                 /**< final error over Y found */
    double finalErrTH_;                                /**< final error over TH found */

    vector<double> *posX;                              /**< vector containing a sampling of the position X */
    vector<double> *posY;                              /**< vector containing a sampling of the position Y */
    vector<double> *headTH;                            /**< vector containing a sampling of the heading TH */
    vector<double> *velV;                              /**< vector containing a sampling of the linear velocity V */
    vector<double> *avelOM;                            /**< vector containing a sampling of the angular velocity OM */
    vector<double> *timeT;                             /**< vector containing a sampling of the time T */

    bool xyFixed_;                                     /**< true if the final position xy must be fixed */
    bool thFixed_;                                     /**< true if the final heading th must be fixed */
    bool vFixed_;                                      /**< true if the final linear velocity v  must be fixed */
    bool omFixed_;                                     /**< true if the final angular velocity om must be fixed */

    int* matStateInd_;                                 /**< indexes of the state parameters in the matrix */

    WCState WCweights_;                                /**< values used to weight the cost function */

    /********** Constructors and destructors **********/
  
  protected : 
    /**
     *  \brief default constructor
     */  
    WCTrajGen();

  public : 
    /**
     *  \brief init constructor
     *  \param sys : the constraints of the wheelchair system
     */
    WCTrajGen(const WheelChairSys* sys);

    /**
     *  \brief destructor
     */  
    virtual ~WCTrajGen();

    /*************** Getters and Setters **************/

  public : 
    /**
     *  \brief return the robotic system constraints
     *  \return the robotic system constraints
     */
    virtual const WheelChairSys* sys() const { return ((WheelChairSys*)sys_); }

    /** 
     *  \brief getter : return true if linear velocity constraints are used
     *  \return true if linear velocity constraints are used
     */
    bool getUseVelCst();

    /**
     *  \brief setter : set true if linear velocity constraints are used
     *  \param u : set it to true to use the linear velocity constraints
     */
    void setUseVelCst(bool u);

    /** 
     *  \brief getter : return true if linear acceleration constraints are used
     *  \return true if linear acceleration constraints are used
     */
    bool getUseAccCst();

    /**
     *  \brief setter : set true if linear acceleration constraints are used
     *  \param u : set it to true to use the linear acceleration constraints
     */
    void setUseAccCst(bool u);

    /** 
     *  \brief getter : return true if angular velocity constraints are used
     *  \return true if angular velocity constraints are used
     */
    bool getUseAVelCst();

    /**
     *  \brief setter : set true if angular velocity constraints are used
     *  \param u : set it to true to use the angular velocity constraints
     */
    void setUseAVelCst(bool u);

    /** 
     *  \brief getter : return true if angular acceleration constraints are used
     *  \return true if angular acceleration constraints are used
     */
    bool getUseAAccCst();

    /**
     *  \brief setter : set true if angular acceleration constraints are used
     *  \param u : set it to true to use the angular acceleration constraints
     */
    void setUseAAccCst(bool u);

    //   /**
    //    *  \brief set the backward motion mode
    //    *  \param cmb : true if the wheelchair can move backward
    //    */
    //   inline void setMoveBackward(bool cmb) { this->canMoveBack_ = cmb; }

    /**
     *  \brief fix or free the final position
     *  \param fixed : true if the final position must be fixed
     */
    void fixFinalXY(bool fixed);

    /**
     *  \brief fix or free the final heading
     *  \param fixed : true if the final heading must be fixed
     */
    void fixFinalTH(bool fixed);

    /**
     *  \brief fix or free the final linear velocity
     *  \param fixed : true if the final linear velocity must be fixed
     */
    void fixFinalV(bool fixed);

    /**
     *  \brief fix or free the final angular velocity
     *  \param fixed : true if the final angular velocity must be fixed
     */
    void fixFinalOM(bool fixed);

    /********* Optimization related routines **********/

  protected : 
    /**
     *  \brief compute a default initial guess of parameters
     */
    virtual void ComputeDfltIG();

    /**
     *  \brief select an initial guess of parameters from an IG database
     *  \return 0 if a set of parameters has been found, -1 else
     */
    virtual int ComputeIGfromDB();

    //   /**
    //    *  \brief set the initial guess with a vector of parameters
    //    *  \param parameters : the vector of parameters
    //    */
    //   virtual void ComputeIG(Vec_DP** parameters);

    /**
     *  \brief compute the error between the state reached and the state we want to reach
     */
    virtual void ComputeG();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters by finite differences
     */
    void ComputeJacobianG_FD();

    /**
     *  \brief compute the partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeJacobianS();

    /**
     *  \brief compute the second order partial derivatives of the state wrt. the parameters
     */
    virtual void ComputeHessianS();

    double fct1(const double t) const;
    double fct2(const double t) const;

    double itgdXdB(const double t) const;
    double itgdXdC(const double t) const;
    double itgdXdD(const double t) const;
    double itgdXdB2(const double t) const;
    double itgdXdC2(const double t) const;
    double itgdXdD2(const double t) const;
    double itgdYdB(const double t) const;
    double itgdYdC(const double t) const;
    double itgdYdD(const double t) const;
    double itgdYdB2(const double t) const;
    double itgdYdC2(const double t) const;
    double itgdYdD2(const double t) const;

    /**
     *  \brief compute the residual between the goal state and the stae reached
     *  \return the residual between the goal state and the stae reached
     */
    virtual double ComputeResidual() const;

    //   /**
    //    *  \brief apply the correction to the set of parameters
    //    */
    //   virtual void ApplyCorrection();

    /**************** Quadrature tools ****************/

  protected :   
    /**
     *  \brief Quadrature tools...
     */
    double qgaus(double (WCTrajGen::*func)(const double) const, const double a, const double b) const;

    /*********** State profiles computations **********/

  protected : 
    /**
     *  \brief compute the linear acceleration profile
     */
    void computeFctAcc();

    /**
     *  \brief compute the linear velocity profile
     */
    void computeFctVel();

    /**
     *  \brief compute the angular acceleration profile
     */
    void computeFctAAcc();

    /**
     *  \brief compute the angular velocity profile
     */
    void computeFctAVel();

    /**
     *  \brief compute the heading profile
     */
    void computeFctTH();

    /**
     *  \brief compute all profiles
     */
    void computeAllFct();

  
    /********** State Parameters Computation **********/  

  public : 
    /**
     *  \brief get the value of the linear acceleration profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear acceleration profile of the robot at time t
     */
    double funcAcc(const double t) const;

    /**
     *  \brief get the value of the angular acceleration profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular acceleration profile of the robot at time t
     */
    double funcAAcc(const double t) const;

    /**
     *  \brief get the value of the linear velocity profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the linear velocity profile of the robot at time t
     */
    double funcVel(const double t) const;

    /**
     *  \brief get the value of the angular velocity profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the angular velocity profile of the robot at time t
     */
    double funcAVel(const double t) const;

    /**
     *  \brief get the value of the heading profile of the robot at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the heading profile of the robot at time t
     */
    double funcTheta(const double t) const;

  protected : 
    /**
     *  \brief integration function for the position X profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    double funcItgX(const double t) const;

    /**
     *  \brief integration function for the position Y profile computation
     *  \param t : the time at which the function is evaluated
     *  \return the value at time t of the integration function
     */  
    double funcItgY(const double t) const;

  public : 
    /**
     *  \brief get the value of the position X profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position X profile at time t
     */
    double funcX(const double t) const;

    /**
     *  \brief get the value of the position Y profile at time t
     *  \param t : the time at which the function is evaluated
     *  \return value of the position Y profile at time t
     */
    double funcY(const double t) const;

    /*************** Forward Solution *****************/

  protected :   
    /**
     *  \brief compute the state reached by the current set of parameters
     */
    virtual void ForwardSolution();

    /***************** Initialization *****************/

  protected : 
    /**
     *  \brief set the initial guess of parameters
     *  \param b,c,d : coefficients of the polynomial right wheel acceleration profile
     *  \param b2,c2,d2 : coefficients of the polynomial left wheel acceleration profile
     *  \param t : final time to compute the trajectory (time of the final state)
     */
    void setInitialGuess(double b1,double c1,double d1,double b2,double c2,double d2,double t);

    //   /**
    //    *  \brief set the initial state, final state and constraints vectors
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    *
    //    *  constraints[0] : max linear acceleration
    //    *  constraints[1] : max angular acceleration
    //    *  constraints[2] : max linear velocity
    //    *  constraints[3] : max angular velocity
    //    */
    //   virtual void setInitialParam(Vec_DP& x0,Vec_DP& xf,Vec_DP& constraints);

    /**
     *  \brief set the system parameters
     *  \param constraints : constraints defined over the system
     */
    virtual void setSysParams(const System* constraints);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     */
    virtual void setInitialGuess(const double* parameters);

    /**
     *  \brief set the initial guess with a vector of parameters
     *  \param parameters : the vector of parameters
     *
     *  x0 = (X0,Y0,TH0,V0,OM0,T0)
     */
    virtual void setInitialConfig(const double* x0);

    /************* Trajectory Generation **************/

  protected : 
    /**
     *  \brief init a vector of the constraints defined over the system
     *  \param sys : the robotic system features
     *  \param vcstt : the resulting vector of constraints
     */
    virtual void initCstt(System* sys,Vec_DP& vcstt);

    /**
     *  \brief init the fixed state parameters
     *  \return the size of the vector of free parameters
     */
    virtual int initFixedSParam();

    //   /**
    //    *  \brief generate a trajectory by constraints optimiation
    //    *  \param x0 : initial state
    //    *  \param xf : final state
    //    *  \param constraints : constraints defined over the system
    //    *  \param parameters : initial guess of parameters
    //    *
    //    *  x0 = (X0,Y0,TH0,V0,OM0,T0)
    //    *  xf = (Xf,Yf,THf,Vf,OMf,Tf)
    //    *  constraints = (maxAcc,maxAAcc,maxVel,maxAVel)
    //    */
    //   virtual int ComputeTrajectory(Vec_DP x0,Vec_DP xf,Vec_DP constraints,Vec_DP** parameters);

  public : 
    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sf : final state
     *  \param parameters : initial guess of parameters
     *  \return 0 if a solution has been found, -1 else
     */
    int ComputeTrajectory(const WCStateT& s0,const WCStateT& sf,double** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sf : final state
     *  \param parameters : initial guess of parameters
     *  \return 0 if a solution has been found, -1 else
     */
    int ComputeTrajectory(const WCStateT& s0,const WCStateT& sf,Vec_DP** parameters);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param param : initial guess of parameters
     */
    virtual void ComputeFromIG(const WCStateT& s0,const double* param);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param s0 : initial state
     *  \param sys : constraints defined over the system
     *  \param param : initial guess of parameters
     */
    inline virtual void ComputeFromIG(const StateTime& s0,const double* param) 
    { this->ComputeFromIG(*((WCStateT*)(&s0)),param); }

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param statet : initial state-time
     *  \param ctrlt : control-time applied on the system
     */
    virtual void ComputeFromCtrl(const StateTime& statet,const ControlTime& ctrlt);

    /**
     *  \brief generate a trajectory by constraints optimiation
     *  \param statet : initial state-time
     *  \param ctrlt : control-time applied on the system
     */
    virtual void ComputeFromCtrl(const WCStateT& statet,const WCControlT& ctrlt);


    /****** Trajectory Generation (2nd version) ******/

  protected : 
    /**
     *  \brief init the cot function weights
     */
    virtual void initWeights();

  public : 
    /**
     *  \brief generate a trajectory between two states
     *  \param s0 : initial state-time
     *  \param sf : initial state-time
     *  \param p : initial guess of parameters
     *  \return 0 if a solution has been found, -1 else
     */
    int genTraj(const WCStateT& s0,const WCStateT& sf,const double* p=NULL);

    /**************** Output Functions ****************/

  public : 
    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,omr,oml,t)
     */
    virtual void getStateTReachedAt(double* rstatet,double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    WCStateT getWCStateTReachedAt(double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    ManaStateT getManaStateTReachedAt(double t) const;

    /**
     *  \brief get back the control reached at time t
     *  \param t : time at which we evaluate the reached control
     *  \result the reached control
     */
    WCControl getWCControlReachedAt(double t) const;

    /**
     *  \brief get back the control reached at time t
     *  \param t : time at which we evaluate the reached control
     *  \result the reached control
     */
    ManaControl getManaControlReachedAt(double t) const;

    /**
     *  \brief get back the state reached at time t
     *  \param rstatet : the resulting reached state
     *  \param t : time at which we evaluate the reached state
     *
     *  The resulting reached state must be already allocated
     *  Reached state-time : (w,y,th,omr,oml,t)
     */
    virtual void getFinalStateTReached(double* rstatet) const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    WCStateT getFinalWCStateTReached() const;

    /**
     *  \brief get back the state reached at time t
     *  \param t : time at which we evaluate the reached state
     *  \result the reached state
     */
    ManaStateT getFinalManaStateTReached() const;

    /**
     *  \brief save a sampling of the trajectory profile
     *  \param nbp : number of sampling points
     */
    virtual void SaveTrajectory(int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     */
    virtual void traceOn(Map *map,int nbp);

    /**
     *  \brief trace the generated trajectory on a map
     *  \param map : the map on which to trace the path
     *  \param nbp : number of sampling points
     *  \param initPos : initial position
     */  
    virtual void traceOn(Map *map,int nbp,const PositionXYTh& initPos);

    /**
     *  \brief return a sample trajectory from the parameterized one
     *  \param ts : the sampling time step
     *  \return a sample trajectory from the parameterized one
     *
     *  A new instance of sample trajectory is returned here
     */
    virtual WCSTraj* sampledTraj(double ts) const;

    /**
     *  \brief sample the X and Y coordinates of the trajectory into the input tables
     *  \param sizeTabs : size of the input tables
     *  \param tabX : resulting table containing the X coordinates
     *  \param tabY : resulting table containing the Y coordinates
     *  \return 0 if the sampling succeed, -1 else
     */
    int sampleXY(int sizeTabs,double** tabX, double** tabY) const;

    /*************** Utility Functions ****************/

  protected : 
    /**
     *  \brief convert the initial and final input state-times into tables of double
     *  \param initST : the input initial state-time
     *  \param goalST : the input goal state-time
     *  \param initTD : the resulting initial table
     *  \param goalTD : the resulting goal table
     *
     *  Tables must be allocated first
     */
    virtual void InitGoalST2Tab(const StateTime& initST,const StateTime& goalST,double* initTD,double* goalTD) const;

  public : 
    /**
     *  \brief display both wheels velocity, linear and angular velocity, heading and position profiles
     */
    void displayProfiles();

    /**
     *  \brief display all functions
     */
    virtual void displayAllFct();
  };

};
#endif // MPC_WCTRAJGEN_HPP

